/***************************************************************************
                         mmx_imdft.c  -  description
                             -------------------
   MMX version of the AC-3 IMDFT process. Suitable for all processors 
   (PMMX, P2, and later) that support the MMX instruction set.
                             -------------------
    begin                : Tue Nov 29 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm6 = 0xFFFFFFFF00000000
                      %%mm7 = 0x00000000FFFFFFFF */
   /*  op_code    dest        src */
# define STAGE_OUTPUT_2R2I(inc, r0, op0, r1, op1, off) \
 __asm movq       mm0,        [r0] \
 __asm movq       mm1,        [r0+off] \
 __asm movq       mm2,        mm0 \
 __asm movq       mm3,        mm1 \
 __asm movq       mm5,        [r1+off] \
 __asm pand       mm2,        mm7 \
 __asm movq       mm4,        mm5 \
 __asm psrlq      mm0,        32 \
 __asm pand       mm1,        mm6 \
 __asm pand       mm3,        mm7 \
 __asm psllq      mm4,        32 \
 __asm por        mm1,        mm2 \
 __asm movq       mm2,        mm3 \
 __asm pand       mm5,        mm6 \
 __asm paddw      mm3,        mm1 \
 __asm psubw      mm1,        mm2 \
 __asm movq       mm2,        [r1] \
 __asm paddw      mm3,        mm4 \
 __asm psubw      mm1,        mm4 \
 __asm movq       [r0],       mm3 \
 __asm movq       mm3,        mm2 \
 __asm movq       [r0+off],   mm1 \
 __asm pand       mm3,        mm7 \
 __asm pand       mm2,        mm6 \
 __asm por        mm5,        mm3 \
 __asm movq       mm4,        mm2 \
 __asm paddw      mm2,        mm5 \
 __asm psubw      mm5,        mm4 \
 __asm psubw      mm2,        mm0 \
 __asm paddw      mm5,        mm0 \
 __asm movq       [r1+off],   mm2 \
 __asm op0        r0,         (inc*2) \
 __asm movq       [r1],       mm5 \
 __asm op1        r1,         (inc*2)

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm6 = 0xFFFFFFFF00000000
                      %%mm7 = 0x00000000FFFFFFFF */
   /*  op_code    dest        src */
# define STAGE_OUTPUT___2I(inc, r0, op0, r1, op1, off) \
 __asm movq       mm2,        [r0] \
 __asm movq       mm1,        [r0+off] \
 __asm movq       mm0,        mm2 \
 __asm pand       mm2,        mm7 \
 __asm movq       mm3,        mm1 \
 __asm pand       mm1,        mm6 \
 __asm movq       mm4,        [r1] \
 __asm pand       mm3,        mm7 \
 __asm movq       mm5,        [r1+off] \
 __asm por        mm1,        mm2 \
 __asm paddw      mm0,        mm3 \
 __asm movq       mm2,        mm5 \
 __asm psubw      mm1,        mm3 \
 __asm movq       mm3,        mm4 \
 __asm pand       mm2,        mm6 \
 __asm pand       mm4,        mm7 \
 __asm movq       [r0],       mm0 \
 __asm por        mm4,        mm2 \
 __asm movq       [r0+off],   mm1 \
 __asm pand       mm3,        mm6 \
 __asm psubw      mm4,        mm3 \
 __asm paddw      mm5,        mm3 \
 __asm movq       [r1],       mm4 \
 __asm op0        r0,         (inc*2) \
 __asm movq       [r1+off],   mm5 \
 __asm op1        r1,         (inc*2)

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm6 = 0xFFFFFFFF00000000
                      %%mm7 = 0x00000000FFFFFFFF */
   /*  op_code    dest        src */
# define STAGE_OUTPUT_2R__(inc, r0, op0, r1, op1, off) \
 __asm movq       mm0,        [r0] \
 __asm movq       mm3,        [r0+off] \
 __asm movq       mm2,        mm0 \
 __asm movq       mm1,        mm3 \
 __asm movq       mm5,        [r1+off] \
 __asm psrlq      mm2,        32 \
 __asm movq       mm4,        mm5 \
 __asm pand       mm0,        mm7 \
 __asm psllq      mm4,        32 \
 __asm pand       mm3,        mm6 \
 __asm pand       mm5,        mm6 \
 __asm por        mm0,        mm3 \
 __asm movq       mm3,        [r1] \
 __asm paddw      mm0,        mm4 \
 __asm psubw      mm1,        mm4 \
 __asm movq       [r0],       mm0 \
 __asm movq       mm4,        mm3 \
 __asm psubw      mm5,        mm2 \
 __asm movq       [r0+off],   mm1 \
 __asm pand       mm4,        mm7 \
 __asm paddw      mm3,        mm2 \
 __asm paddw      mm5,        mm4 \
 __asm movq       [r1],       mm3 \
 __asm op0        r0,         (inc*2) \
 __asm movq       [r1+off],   mm5 \
 __asm op1        r1,         (inc*2)

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                 (edi)%4    = 0x5A815A81
                      %%mm6 = 0xFFFFFFFF00000000
                      %%mm7 = 0x00000000FFFFFFFF */
   /*  op_code    dest        src */
# define SR_BF_k_eq_N8_2R(inc, r0, op0, r1, op1, off) \
 __asm movq       mm0,        [r0] \
 __asm movq       mm1,        [r1+off] \
 __asm movq       mm4,        mm0 \
 __asm movq       mm5,        mm1 \
 __asm movq       mm2,        [r0+off] \
 __asm psllq      mm5,        32 \
 __asm movq       mm3,        [r1] \
 __asm psrlq      mm4,        32 \
 __asm paddw      mm0,        mm5 \
 __asm movq       mm5,        mm2 \
 __asm psubw      mm1,        mm4 \
 __asm psrlq      mm5,        32 \
 __asm movq       mm4,        mm3 \
 __asm paddw      mm3,        mm5 \
 __asm psllq      mm4,        32 \
 __asm movq       mm5,        mm0 \
 __asm psubw      mm2,        mm4 \
 __asm pand       mm5,        mm6 \
 __asm movq       mm4,        mm2 \
 __asm psubw      mm2,        mm5 \
 __asm pand       mm4,        mm6 \
 __asm movq       mm5,        mm1 \
 __asm paddw      mm0,        mm4 \
 __asm pand       mm5,        mm7 \
 __asm movq       mm4,        mm3 \
 __asm psubw      mm3,        mm5 \
 __asm pand       mm4,        mm7 \
 __asm movd       mm5,        edi \
 __asm paddw      mm1,        mm4 \
 __asm movq       mm4,        mm3 \
 __asm pmulhw     mm3,        mm5 \
 __asm pand       mm4,        mm6 \
 __asm psllw      mm3,        1 \
 __asm por        mm3,        mm4 \
 __asm movq       mm4,        mm1 \
 __asm pmulhw     mm1,        mm5 \
 __asm pand       mm4,        mm6 \
 __asm psllw      mm1,        1 \
 __asm por        mm1,        mm4 \
 __asm psllq      mm5,        32 \
 __asm movq       mm4,        mm0 \
 __asm pmulhw     mm0,        mm5 \
 __asm pand       mm4,        mm7 \
 __asm psllw      mm0,        1 \
 __asm por        mm0,        mm4 \
 __asm movq       mm4,        mm2 \
 __asm pmulhw     mm2,        mm5 \
 __asm pand       mm4,        mm7 \
 __asm psrlw      mm5,        14 \
 __asm psllw      mm2,        1 \
 __asm por        mm0,        mm5 \
 __asm por        mm2,        mm4 \
 __asm movq       mm4,        mm5 \
 __asm por        mm2,        mm5 \
 __asm psrlq      mm4,        32 \
 __asm movq       [r0],       mm0 \
 __asm por        mm3,        mm4 \
 __asm movq       [r0+off],   mm2 \
 __asm por        mm1,        mm4 \
 __asm movq       [r1],       mm3 \
 __asm op0        r0,         (inc*2) \
 __asm movq       [r1+off],   mm1 \
 __asm op1        r1,         (inc*2)

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                 (ecx)%2    = IDFT_CONST
                      %%mm6 = 0xFFFFFFFF00000000
                      %%mm7 = 0x00000000FFFFFFFF */
   /*  op_code    dest        src */
# define SR_BF_2R(inc, r0, op0, r1, op1, off, h1, h2) \
 __asm movq       mm0,        [r0] \
 __asm movq       mm1,        [r1+off] \
 __asm psrlq      mm0,        32 \
 __asm movq       mm4,        [ecx] \
 __asm punpcklwd  mm1,        mm0 \
 __asm movq       mm5,        mm4 \
 __asm punpck##h1##dq mm4,        mm4 \
 __asm movq       mm0,        mm1 \
 __asm punpck##h2##dq mm5,        mm5 \
 __asm movq       mm2,        [r0+off] \
 __asm pmaddwd    mm0,        mm4 \
 __asm movq       mm3,        [r1] \
 __asm pmaddwd    mm1,        mm5 \
 __asm movq       mm5,        [ecx+8] \
 __asm psrlq      mm2,        32 \
 __asm movq       mm4,        mm5 \
 __asm punpck##h2##dq mm5,        mm5 \
 __asm punpcklwd  mm3,        mm2 \
 __asm punpck##h1##dq mm4,        mm4 \
 __asm movq       mm2,        mm3 \
 __asm pmaddwd    mm3,        mm5 \
 __asm movq       mm5,        mm7 \
 __asm pmaddwd    mm2,        mm4 \
 __asm punpcklwd  mm5,        mm6 \
 __asm psrlw      mm5,        2 \
 __asm paddd      mm2,        mm5 \
 __asm paddd      mm3,        mm5 \
 __asm movq       mm4,        mm2 \
 __asm psubd      mm2,        mm0 \
 __asm paddd      mm0,        mm4 \
 __asm psrad      mm2,        15 \
 __asm movq       mm5,        mm3 \
 __asm psrad      mm0,        15 \
 __asm psubd      mm3,        mm1 \
 __asm paddd      mm1,        mm5 \
 __asm psrad      mm3,        15 \
 __asm pxor       mm4,        mm4 \
 __asm psrad      mm1,        15 \
 __asm pxor       mm5,        mm5 \
 __asm packssdw   mm3,        mm4 \
 __asm packssdw   mm5,        mm2 \
 __asm movq       mm2,        [r0] \
 __asm packssdw   mm1,        mm4 \
 __asm pand       mm2,        mm7 \
 __asm packssdw   mm4,        mm0 \
 __asm movq       mm0,        [r0+off] \
 __asm por        mm2,        mm5 \
 __asm movq       mm5,        [r1] \
 __asm pand       mm0,        mm7 \
 __asm movq       [r0],       mm2 \
 __asm pand       mm5,        mm6 \
 __asm movq       mm2,        [r1+off] \
 __asm por        mm0,        mm4 \
 __asm por        mm5,        mm1 \
 __asm movq       [r0+off],   mm0 \
 __asm pand       mm2,        mm6 \
 __asm movq       [r1],       mm5 \
 __asm por        mm2,        mm3 \
 __asm op0        r0,         (inc*2) \
 __asm movq       [r1+off],   mm2 \
 __asm op1        r1,         (inc*2)

   /* registers used: %%mm6 = 0xFFFFFFFF00000000
                      %%mm7 = 0x00000000FFFFFFFF */
   /*  op_code    dest        src */
# define PRE_POST_SCALING_SETUP \
 __asm movq       mm3,        mm7 \
 __asm punpcklwd  mm3,        mm6 \
 __asm psrlw      mm3,        2

   /* registers used: (eax)%0 = data+? -> add(4)
                      (ebx)%1 = data+? -> sub(4)
                      (edx)%5 = MMX_256_CONST
                        %%mm3 = 0x00003FFF00003FFF
                        %%mm6 = 0xFFFFFFFF00000000
                        %%mm7 = 0x00000000FFFFFFFF */
   /*  op_code    dest        src */
# define PRE_POST_SCALING(op) \
 __asm movq       mm0,        [eax] \
 __asm movq       mm1,        [ebx] \
 __asm psllq      mm0,        32 \
 __asm movq       mm4,        [edx] \
 __asm punpckhwd  mm0,        mm1 \
 __asm movq       mm5,        mm4 \
 __asm punpckldq  mm4,        mm4 \
 __asm movq       mm1,        mm0 \
 __asm punpckhdq  mm5,        mm5 \
 __asm movq       mm2,        [eax] \
 __asm pmaddwd    mm0,        mm4 \
 __asm pand       mm2,        mm6 \
 __asm pmaddwd    mm1,        mm5 \
 __asm paddd      mm0,        mm3 \
 __asm p##op##d   mm1,        mm3 \
 __asm psrad      mm0,        15 \
 __asm movq       mm5,        [ebx] \
 __asm pxor       mm4,        mm4 \
 __asm psrad      mm1,        15 \
 __asm pand       mm5,        mm7 \
 __asm packssdw   mm0,        mm4 \
 __asm add        edx,        8 \
 __asm packssdw   mm4,        mm1 \
 __asm por        mm0,        mm2 \
 __asm p##op##d   mm5,        mm4 \
 __asm movq       [eax],      mm0 \
 __asm add        eax,        8 \
 __asm movq       [ebx],      mm5 \
 __asm sub        ebx,        8


void AC3_imdft256(AC3_COEFF *data)
{
   __asm {
   /* input list
   eax %0: "q" (data),
   ebx %1  "q" (data+252),
   ecx %2  "q" (MMX_IDFT_CONST+256),
   esi %3  "r" (64),
   edi %4  "r" (0x5A815A81),
   edx %5  "q" (MMX_256_CONST) */
   /*  op_code    dest        src */
       mov        eax,        dword ptr[data]
       mov        ecx,        offset MMX_IDFT_CONST
       mov        ebx,        eax
       mov        esi,        64
       add        ecx,        (256*2)
       mov        edi,        0x5A815A81
       add        ebx,        (252*2)
       mov        edx,        offset MMX_256_CONST

   /* INIT MASKS */
   /*  op_code    dest        src */
       pcmpeqd    mm7,        mm7
       pxor       mm6,        mm6
       movq       mm5,        mm7
       punpckldq  mm7,        mm6    /* mm7 = 0x00000000FFFFFFFF */
       punpckldq  mm6,        mm5    /* mm6 = 0xFFFFFFFF00000000 */

   /* Pre-IDFT Scaling:
      registers used: %0 = data
                      %1 = data+252
                      %3 = 64 */
   /*  op_code    dest        src */
      PRE_POST_SCALING_SETUP
      /* %0 = data      %1 = data+252 */
   PRE_LOOP_256:
      PRE_POST_SCALING(add)
       dec        esi
       jnz        PRE_LOOP_256
      /* %0 = data+256  %1 = data-4 */

   /* STAGE 1: 2pt-DFTs
      registers used: %0 = data
                      %1 = data+128
                      %3 = 32 */
   /*  op_code    dest        src */
       sub        eax,        (256*2)
       add        ebx,        (132*2)
       add        esi,        32
      /* %0 = data      %1 = data+128 */
   STAGE1_LOOP1_256:
      STAGE1
       dec        esi
       jnz        STAGE1_LOOP1_256
      /* %0 = data+128  %1 = data+256 */

   /* STAGE 2: 4pt-DFTs
      registers used: %0 = data
                      %1 = data+188
                      %3 = 4 */
   /*  op_code    dest        src */
       sub        eax,        (128*2)
       sub        ebx,        (68*2)
       add        esi,        4
      /* %0 = data      %1 = data+188 */
      STAGE_OUTPUT_2R2I(4, eax, add, ebx, sub, 128)
      STAGE_OUTPUT___2I(4, eax, add, ebx, sub, 128)
      STAGE_OUTPUT_2R2I(4, eax, add, ebx, sub, 128)
      STAGE_OUTPUT_2R2I(4, eax, add, ebx, sub, 128)
   STAGE2_LOOP1_256:
      STAGE_OUTPUT___2I(4, eax, add, ebx, sub, 128)
       dec        esi
       jnz        STAGE2_LOOP1_256
       add        esi,        4
   STAGE2_LOOP2_256:
      STAGE_OUTPUT_2R__(4, eax, add, ebx, sub, 128)
       dec        esi
       jnz        STAGE2_LOOP2_256
      STAGE_OUTPUT_2R2I(4, eax, add, ebx, sub, 128)
      STAGE_OUTPUT_2R2I(4, eax, add, ebx, sub, 128)
      STAGE_OUTPUT_2R__(4, eax, add, ebx, sub, 128)
      STAGE_OUTPUT_2R2I(4, eax, sub, ebx, add, 128)
      /* %0 = data+56   %1 = data+132 */

   /* STAGE 3: 8pt-DFTs
      registers used: %0 = data+56
                      %1 = data+132
                      %3 = 4
                      %4 = 0x5A815A81 */
   /*  op_code    dest        src */
       add        esi,        4
      /* 1, 3 */
      /* %0 = data+56   %1 = data+132 */
      SR_BF_k_eq_0_2I(12, eax, sub, ebx, add, 128)
   STAGE3_LOOP1_256:
      SR_BF_k_eq_0_2I( 4, eax, sub, ebx, add, 128)
       dec        esi
       jnz        STAGE3_LOOP1_256
      /* 7, 5 */
      /* %0 = data+28   %1 = data+160 */
       add        esi,        4
   STAGE3_LOOP2_256:
      SR_BF_k_eq_N8_2R(4, eax, sub, ebx, add, 128)
       dec        esi
       jnz        STAGE3_LOOP2_256
       sub        eax,        (8*2)
       add        ebx,        (8*2)
      SR_BF_k_eq_N8_2R(4, eax, sub, ebx, add, 128)
      /* %0 = data      %1 = data+188 */

       add        ebx,        (32*2)
       add        esi,        2
      /* %0 = data      %1 = data+220 */
   STAGE3_LOOP3_256:
      STAGE_OUTPUT___2I(  4, eax, add, ebx, sub, 64)
      STAGE_OUTPUT_2R2I(  4, eax, add, ebx, sub, 64)
      STAGE_OUTPUT___2I(  4, eax, add, ebx, sub, 64)
      STAGE_OUTPUT___2I(  4, eax, add, ebx, sub, 64)
      STAGE_OUTPUT_2R__(  4, eax, add, ebx, sub, 64)
      STAGE_OUTPUT_2R__(  4, eax, add, ebx, sub, 64)
      STAGE_OUTPUT_2R2I(  4, eax, add, ebx, sub, 64)
      STAGE_OUTPUT_2R__(100, eax, add, ebx, sub, 64)
       dec        esi
       jnz        STAGE3_LOOP3_256
      /* %0 = data+256  %1 = data-36 */

   /* STAGE 4: 16pt-DFTs
      registers used: %0 = data
                      %1 = data+220
                      %2 = IDFT_CONST
                      %3 = 2
                      %4 = 0x5A815A81 */
   /*  op_code    dest        src */
       sub        eax,        (256*2)
       add        ebx,        (256*2)
       add        esi,        2
      /* 15, 13 */
      /* 11,  9 */
      /* %0 = data      %1 = data+220 */
   STAGE4_LOOP1_256:
      SR_BF_2R( 8, eax, add, ebx, sub, 64, h, l)
      SR_BF_2R( 4, eax, add, ebx, sub, 64, h, l)
      SR_BF_2R(52, eax, add, ebx, sub, 64, h, l)
       add        ecx,        (8*2)
       dec        esi
       jnz        STAGE4_LOOP1_256
      /* %0 = data+128  %1 = data+92 */
       add        eax,        (12*2)
       sub        ebx,        (12*2)
      /* 7, 5 */
      /* %0 = data+140  %1 = data+80 */
      SR_BF_k_eq_N8_2R( 4, eax, sub, ebx, add, 64)
      SR_BF_k_eq_N8_2R( 8, eax, sub, ebx, add, 64)
      SR_BF_k_eq_N8_2R(76, eax, add, ebx, sub, 64)
      /* 3, 1 */
      /* %0 = data+204  %1 = data+16 */
      SR_BF_k_eq_0_2I( 4, ebx, add, eax, sub, 64)
      SR_BF_k_eq_0_2I( 8, ebx, add, eax, sub, 64)
      SR_BF_k_eq_0_2I(28, ebx, sub, eax, add, 64)
      /* %0 = data+220  %1 = data */

       add        eax,        (16*2)
       add        esi,        4
      /* %0 = data+236  %1 = data */
   STAGE4_LOOP2_256:
      STAGE_OUTPUT_2R2I( 4, ebx, add, eax, sub, 32)
      STAGE_OUTPUT___2I( 4, ebx, add, eax, sub, 32)
      STAGE_OUTPUT_2R__( 4, ebx, add, eax, sub, 32)
      STAGE_OUTPUT_2R2I(52, ebx, add, eax, sub, 32)
       dec        esi
       jnz        STAGE4_LOOP2_256
      /* %0 = data-20   %1 = data+256 */

   /* STAGE 5: 32pt-DFTs
      registers used: %0 = data+4
                      %1 = data+232
                      %2 = IDFT_CONST+16
                      %3 = 6
                      %4 = 0x5A815A81 */
   /*  op_code    dest        src */
       add        eax,        (24*2)
       sub        ebx,        (24*2)
       add        esi,        6
      /* 31, 29
          :   :
         11,  9 */
      /* %0 = data+4    %1 = data+232 */
   STAGE5_LOOP1_256:
      SR_BF_2R(32, eax, add, ebx, sub, 32, l, h)
       add        ecx,        (8*2)
       dec        esi
       jnz        STAGE5_LOOP1_256
      /* 7, 5 */
      /* %0 = data+196  %1 = data+40 */
      SR_BF_k_eq_N8_2R(32, eax, add, ebx, sub, 32)
      /* 3, 1 */
      /* %0 = data+228  %1 = data+8 */
      SR_BF_k_eq_0_2I(8, ebx, sub, eax, add, 32)
      /* %0 = data+236  %1 = data+0 */

       add        eax,        (8*2)
       add        esi,        8
      /* %0 = data+244  %1 = data */
   STAGE5_LOOP2_256:
      STAGE_OUTPUT___2I( 4, ebx, add, eax, sub, 16)
      STAGE_OUTPUT_2R__(28, ebx, add, eax, sub, 16)
       dec        esi
       jnz        STAGE5_LOOP2_256
      /* %0 = data-12   %1 = data+256 */

   /* STAGE 6: 64pt-DFTs
      registers used: %0 = data+4
                      %1 = data+240
                      %2 = IDFT_CONST+64
                      %3 = 14
                      %4 = 0x5A815A81 */
   /*  op_code    dest        src */
       add        eax,        (16*2)
       sub        ebx,        (16*2)
       add        esi,        14
      /* 1, 3 */
      /* %0 = data+4    %1 = data+240 */
      SR_BF_k_eq_0_2I(16, eax, add, ebx, sub, 16)
      /* 5, 7 */
      /* %0 = data+20   %1 = data+224 */
      SR_BF_k_eq_N8_2R(16, ebx, sub, eax, add, 16)
      /*  9, 11
          :   :
         61, 63  */
      /* %0 = data+36   %1 = data+208 */
   STAGE6_LOOP1_256:
      SR_BF_2R(16, ebx, sub, eax, add, 16, h, l)
       add        ecx,        (8*2)
       dec        esi
       jnz        STAGE6_LOOP1_256
      /* %0 = data+260  %1 = data-16 */

       add        ebx,        (16*2)
       sub        eax,        (12*2)
       add        esi,        8
      /* %0 = data+248  %1 = data */
   STAGE6_LOOP2_256:
      STAGE_OUTPUT_2R2I(8, ebx, add, eax, sub, 8)
      STAGE_OUTPUT_2R2I(8, eax, sub, ebx, add, 8)
       dec        esi
       jnz        STAGE6_LOOP2_256
      /* %0 = data+120  %1 = data+128 */

   /* Undo Bit-Reverse Order
      registers used: %0 = data+4
                      %1 = data+128
                      %3 = 28
                      %5 = MMX_256_CONST */
   /*  op_code    dest        src */
       sub        eax,        (116*2)
       add        esi,        28
      /* %0 = data+4    %1 = data+128 */
   SWAP_LOOP_256:
       movq       mm0,        [eax]
       movq       mm1,        [ebx]
       movq       [ebx],      mm0
       mov        ecx,        [edx]
       movq       [eax],      mm1
       mov        edi,        [edx+4]
       add        eax,        ecx
       add        ebx,        edi
       add        edx,        8
       dec        esi
       jnz        SWAP_LOOP_256

   /* Post-IDFT Scaling:
      registers used: %0 = data
                      %1 = data+252
                      %3 = 64 */
   /*  op_code    dest        src */
       add        esi,        64
       sub        edx,        (368*2)
      PRE_POST_SCALING_SETUP
      /* %0 = data      %1 = data+252 */
   POST_LOOP_256:
      PRE_POST_SCALING(sub)
       dec        esi
       jnz        POST_LOOP_256
      /* %0 = data+256  %1 = data-4 */

       emms
   }
}
#undef STAGE1
#undef STAGE_OUTPUT_2R2I
#undef STAGE_OUTPUT___2I
#undef STAGE_OUTPUT_2R__
#undef SR_BF_k_eq_0_2I
#undef SR_BF_k_eq_N8_2R
#undef SR_BF_2R
#undef PRE_POST_SCALING_SETUP
#undef PRE_POST_SCALING

