/***************************************************************************
                         mmx_imdft.c  -  description
                             -------------------
   MMX version of the AC-3 IMDFT process. Suitable for all processors 
   (PMMX, P2, and later) that support the MMX instruction set.
                             -------------------
    begin                : Tue Nov 29 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
   
   /* registers used: (eax)%0 = data+? -> +4
                      (ebx)%1 = data+? -> -4
                        %%mm6 = 0xFFFF0000FFFF0000
                        %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code    dest        src */
# define STAGE1 \
 __asm movq       mm0,        [eax] \
 __asm movq       mm1,        [ebx] \
 __asm movq       mm2,        mm0 \
 __asm movq       mm3,        mm0 \
 __asm pand       mm2,        mm7 \
 __asm pand       mm3,        mm6 \
 __asm movq       mm4,        mm1 \
 __asm movq       mm5,        mm1 \
 __asm pand       mm4,        mm7 \
 __asm pand       mm5,        mm6 \
 __asm paddw      mm1,        mm2 \
 __asm psubw      mm0,        mm4 \
 __asm psubw      mm1,        mm3 \
 __asm paddw      mm0,        mm5 \
 __asm movq       [eax],      mm1 \
 __asm add        eax,        (4*2) \
 __asm movq       [ebx],      mm0 \
 __asm add        ebx,        (4*2)

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm6 = 0xFFFF0000FFFF0000
                      %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code    dest        src */
# define STAGE_OUTPUT_1R2I(inc, r0, op0, r1, op1, off, d1, d2, sc) \
 __asm movq       mm0,        [r0] \
 __asm movq       mm5,        mm6 \
 __asm movq       mm1,        [r0+off] \
 __asm ps##d1##lq mm5,        32 \
 __asm movq       mm2,        mm0 \
 __asm movq       mm3,        mm1 \
 __asm movq       mm4,        mm1 \
 __asm pand       mm2,        mm7 \
 __asm pand       mm3,        mm7 \
 __asm pand       mm4,        mm5 \
 __asm ps##d2##lq mm5,        32 \
 __asm por        mm4,        mm2 \
 __asm movq       mm2,        mm0 \
 __asm pand       mm0,        mm5 \
 __asm ps##d2##lq mm2,        sc \
 __asm pand       mm1,        mm5 \
 __asm paddw      mm0,        mm3 \
 __asm psubw      mm1,        mm3 \
 __asm paddw      mm0,        mm4 \
 __asm paddw      mm1,        mm4 \
 __asm movq       mm3,        mm7 \
 __asm movq       mm4,        [r1+off] \
 __asm psrlq      mm5,        16 \
 __asm pand       mm3,        mm4 \
 __asm pand       mm2,        mm7 \
 __asm ps##d1##lq mm3,        sc \
 __asm paddw      mm0,        mm3 \
 __asm psubw      mm1,        mm3 \
 __asm movq       [r0],       mm0 \
 __asm movq       mm3,        [r1] \
 __asm movq       [r0+off],   mm1 \
 __asm movq       mm0,        mm4 \
 __asm movq       mm1,        mm3 \
 __asm pand       mm0,        mm6 \
 __asm pand       mm1,        mm5 \
 __asm ps##d1##lq mm5,        32 \
 __asm por        mm0,        mm1 \
 __asm movq       mm1,        mm3 \
 __asm pand       mm3,        mm5 \
 __asm pand       mm4,        mm5 \
 __asm pand       mm1,        mm6 \
 __asm paddw      mm3,        mm0 \
 __asm paddw      mm4,        mm0 \
 __asm psubw      mm3,        mm1 \
 __asm paddw      mm4,        mm1 \
 __asm paddw      mm3,        mm2 \
 __asm psubw      mm4,        mm2 \
 __asm movq       [r1],       mm3 \
 __asm op0        r0,         (inc*2) \
 __asm movq       [r1+off],   mm4 \
 __asm op1        r1,         (inc*2)

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm6 = 0xFFFF0000FFFF0000
                      %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code    dest        src */
# define STAGE_OUTPUT___2I(inc, r0, op0, r1, op1, off) \
 __asm movq       mm0,        [r0] \
 __asm movq       mm1,        [r0+off] \
 __asm movq       mm2,        mm0 \
 __asm movq       mm3,        mm1 \
 __asm pand       mm2,        mm7 \
 __asm pand       mm3,        mm7 \
 __asm pand       mm1,        mm6 \
 __asm psubw      mm2,        mm3 \
 __asm paddw      mm0,        mm3 \
 __asm por        mm1,        mm2 \
 __asm movq       [r0],       mm0 \
 __asm movq       mm5,        [r1] \
 __asm movq       [r0+off],   mm1 \
 __asm movq       mm4,        [r1+off] \
 __asm movq       mm3,        mm5 \
 __asm movq       mm2,        mm4 \
 __asm pand       mm3,        mm6 \
 __asm pand       mm2,        mm6 \
 __asm pand       mm5,        mm7 \
 __asm psubw      mm2,        mm3 \
 __asm paddw      mm4,        mm3 \
 __asm por        mm5,        mm2 \
 __asm movq       [r1+off],   mm4 \
 __asm op0        r0,         (inc*2) \
 __asm movq       [r1],       mm5 \
 __asm op1        r1,         (inc*2)

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm6 = 0xFFFF0000FFFF0000
                      %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code    dest        src */
# define STAGE_OUTPUT_2R2I(inc, r0, op0, r1, op1, off) \
 __asm movq       mm0,        [r0] \
 __asm movq       mm1,        [r0+off] \
 __asm movq       mm2,        mm0 \
 __asm movq       mm4,        mm0 \
 __asm movq       mm3,        mm1 \
 __asm pand       mm0,        mm6 \
 __asm pand       mm2,        mm7 \
 __asm psllq      mm0,        16 \
 __asm pand       mm1,        mm6 \
 __asm psrlq      mm4,        48 \
 __asm por        mm1,        mm2 \
 __asm pand       mm3,        mm7 \
 __asm movq       mm5,        [r1+off] \
 __asm movq       mm2,        mm1 \
 __asm paddw      mm1,        mm3 \
 __asm por        mm0,        mm4 \
 __asm movq       mm4,        mm5 \
 __asm pand       mm5,        mm7 \
 __asm psubw      mm2,        mm3 \
 __asm movq       mm3,        mm5 \
 __asm psllq      mm5,        48 \
 __asm psrlq      mm3,        16 \
 __asm pand       mm4,        mm6 \
 __asm por        mm3,        mm5 \
 __asm movq       mm5,        [r1] \
 __asm paddw      mm1,        mm3 \
 __asm psubw      mm2,        mm3 \
 __asm movq       mm3,        mm5 \
 __asm movq       [r0],       mm1 \
 __asm pand       mm3,        mm7 \
 __asm pand       mm5,        mm6 \
 __asm por        mm4,        mm3 \
 __asm movq       [r0+off],   mm2 \
 __asm movq       mm3,        mm4 \
 __asm psubw      mm4,        mm5 \
 __asm paddw      mm3,        mm5 \
 __asm paddw      mm4,        mm0 \
 __asm psubw      mm3,        mm0 \
 __asm movq       [r1],       mm4 \
 __asm op0        r0,         (inc*2) \
 __asm movq       [r1+off],   mm3 \
 __asm op1        r1,         (inc*2)

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm6 = 0xFFFF0000FFFF0000
                      %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code    dest        src */
# define STAGE_OUTPUT_2R__(inc, r0, op0, r1, op1, off) \
 __asm movq       mm0,        [r0] \
 __asm movq       mm1,        [r0+off] \
 __asm movq       mm2,        mm0 \
 __asm pand       mm0,        mm7 \
 __asm movq       mm3,        mm1 \
 __asm pand       mm2,        mm6 \
 __asm pand       mm3,        mm6 \
 __asm movq       mm5,        mm2 \
 __asm psllq      mm2,        16 \
 __asm movq       mm4,        [r1+off] \
 __asm por        mm0,        mm3 \
 __asm movq       mm3,        mm4 \
 __asm psrlq      mm5,        48 \
 __asm pand       mm3,        mm7 \
 __asm por        mm2,        mm5 \
 __asm movq       mm5,        mm4 \
 __asm psrlq      mm3,        16 \
 __asm pand       mm4,        mm6 \
 __asm psllq      mm5,        48 \
 __asm psubw      mm4,        mm2 \
 __asm por        mm3,        mm5 \
 __asm movq       mm5,        [r1] \
 __asm paddw      mm0,        mm3 \
 __asm psubw      mm1,        mm3 \
 __asm movq       [r0],       mm0 \
 __asm movq       mm0,        mm5 \
 __asm movq       [r0+off],   mm1 \
 __asm pand       mm0,        mm7 \
 __asm paddw      mm5,        mm2 \
 __asm paddw      mm4,        mm0 \
 __asm movq       [r1],       mm5 \
 __asm op0        r0,         (inc*2) \
 __asm movq       [r1+off],   mm4 \
 __asm op1        r1,         (inc*2)

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm6 = 0xFFFF0000FFFF0000
                      %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code    dest        src */
# define STAGE_OUTPUT_2R1I(inc, r0, op0, r1, op1, off, d1, d2) \
 __asm movq       mm5,        mm7 \
 __asm movq       mm0,        [r0] \
 __asm ps##d2##lq mm5,        32 \
 __asm movq       mm2,        mm0 \
 __asm movq       mm3,        [r0+off] \
 __asm pand       mm0,        mm7 \
 __asm movq       mm1,        mm3 \
 __asm pand       mm3,        mm6 \
 __asm movq       mm4,        mm1 \
 __asm por        mm0,        mm3 \
 __asm pand       mm4,        mm5 \
 __asm movq       mm3,        mm0 \
 __asm paddw      mm0,        mm4 \
 __asm pand       mm3,        mm5 \
 __asm ps##d1##lq mm5,        32 \
 __asm psubw      mm3,        mm4 \
 __asm pcmpeqw    mm5,        mm7 \
 __asm pand       mm2,        mm6 \
 __asm pand       mm1,        mm5 \
 __asm por        mm1,        mm3 \
 __asm movq       mm5,        mm2 \
 __asm movq       mm4,        [r1+off] \
 __asm psllq      mm2,        16 \
 __asm movq       mm3,        mm4 \
 __asm psrlq      mm5,        48 \
 __asm pand       mm3,        mm7 \
 __asm por        mm2,        mm5 \
 __asm movq       mm5,        mm3 \
 __asm psrlq      mm3,        16 \
 __asm psllq      mm5,        48 \
 __asm por        mm3,        mm5 \
 __asm movq       mm5,        [r1] \
 __asm paddw      mm0,        mm3 \
 __asm psubw      mm1,        mm3 \
 __asm movq       [r0],       mm0 \
 __asm movq       mm3,        mm5 \
 __asm movq       mm0,        mm6 \
 __asm pand       mm3,        mm7 \
 __asm movq       [r0+off],   mm1 \
 __asm paddw      mm5,        mm2 \
 __asm psubw      mm3,        mm2 \
 __asm movq       mm2,        mm6 \
 __asm pand       mm4,        mm6 \
 __asm ps##d1##lq mm0,        32 \
 __asm por        mm4,        mm3 \
 __asm movq       mm1,        mm5 \
 __asm ps##d2##lq mm2,        32 \
 __asm movq       mm3,        mm4 \
 __asm pcmpeqw    mm2,        mm6 \
 __asm pand       mm1,        mm0 \
 __asm pand       mm3,        mm0 \
 __asm paddw      mm4,        mm1 \
 __asm psubw      mm3,        mm1 \
 __asm pand       mm5,        mm2 \
 __asm movq       [r1+off],   mm4 \
 __asm por        mm5,        mm3 \
 __asm op0        r0,         (inc*2) \
 __asm movq       [r1],       mm5 \
 __asm op1        r1,         (inc*2)

   /* registers used: %r0   = data+? -> op0(inc)
                      %%mm4 = 0xFFFFFFFFFFFF0000 or 0x0000FFFFFFFFFFFF
                      %%mm5 = 0x000000000000FFFF or 0xFFFF000000000000 */
   /*  op_code    dest        src */
# define STAGE6_OUTPUT_01(inc, r0, op0, dat0, dat1) \
 __asm movq       mm0,        dat0 \
 __asm movq       mm1,        dat1 \
 __asm movq       mm2,        mm0 \
 __asm movq       mm3,        mm1 \
 __asm pand       mm1,        mm4 \
 __asm pand       mm3,        mm5 \
 __asm pand       mm2,        mm5 \
 __asm paddw      mm0,        mm3 \
 __asm psubw      mm2,        mm3 \
 __asm movq       dat0,       mm0 \
 __asm por        mm1,        mm2 \
 __asm movq       dat1,       mm1 \
 __asm op0        r0,         (inc*2)

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm4 = 0x000000000000FFFF
                      %%mm5 = 0xFFFF000000000000 */
   /*  op_code    dest        src */
# define STAGE6_OUTPUT_23(inc, r0, op0, r1, op1) \
 __asm movq       mm0,        [r1] \
 __asm movq       mm1,        [r0+8] \
 __asm psrlq      mm0,        48 \
 __asm movq       mm3,        [r1+8] \
 __asm psllq      mm1,        48 \
 __asm movq       mm2,        [r0] \
 __asm psubw      mm3,        mm1 \
 __asm paddw      mm2,        mm0 \
 __asm paddw      mm1,        [r1+8] \
 __asm movq       [r1+8],     mm3 \
 __asm pcmpgtw    mm3,        mm3 \
 __asm psubw      mm0,        [r0] \
 __asm pcmpeqw    mm3,        mm5 \
 __asm movq       [r0],       mm2 \
 __asm pcmpgtw    mm2,        mm2 \
 __asm pand       mm3,        [r1] \
 __asm pcmpeqw    mm2,        mm4 \
 __asm pand       mm1,        mm5 \
 __asm pand       mm0,        mm4 \
 __asm pand       mm2,        [r0+8] \
 __asm por        mm3,        mm1 \
 __asm psubw      mm2,        mm0 \
 __asm movq       [r1],       mm3 \
 __asm op1        r1,         (inc*2) \
 __asm movq       [r0+8],     mm2 \
 __asm op0        r0,         (inc*2)
   
   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc) */
   /*  op_code    dest        src */
# define STAGE7_OUTPUT(inc, r0, op0, r1, op1) \
 __asm movq       mm2,        [r0] \
 __asm pcmpgtd    mm3,        mm3 \
 __asm movq       mm1,        [r1] \
 __asm movq       mm0,        mm2 \
 __asm punpckhwd  mm2,        mm1 \
 __asm psubw      mm3,        mm0 \
 __asm punpcklwd  mm0,        mm1 \
 __asm movq       mm4,        mm2 \
 __asm punpcklwd  mm3,        mm1 \
 __asm psllq      mm0,        32 \
 __asm psllq      mm2,        32 \
 __asm punpckhwd  mm0,        mm4 \
 __asm psrlq      mm3,        32 \
 __asm movq       mm1,        mm0 \
 __asm punpckldq  mm0,        mm0 \
 __asm punpckhdq  mm1,        mm1 \
 __asm psubw      mm0,        mm2 \
 __asm psrlq      mm2,        32 \
 __asm psubw      mm1,        mm3 \
 __asm psllq      mm3,        32 \
 __asm paddw      mm0,        mm2 \
 __asm paddw      mm1,        mm3 \
 __asm movq       [r0],       mm0 \
 __asm op0        r0,         (inc*2) \
 __asm movq       [r1],       mm1 \
 __asm op1        r1,         (inc*2)
   
   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm6 = 0xFFFF0000FFFF0000
                      %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code    dest        src */
# define SR_BF_k_eq_0_1I(inc, r0, op0, r1, op1, off, d1, d2) \
 __asm movq       mm5,        mm7 \
 __asm pxor       mm4,        mm4 \
 __asm ps##d1##lq mm5,        32 \
 __asm movq       mm0,        [r0] \
 __asm pcmpeqw    mm4,        mm5 \
 __asm movq       mm2,        mm0 \
 __asm movq       mm3,        [r0+off] \
 __asm pand       mm2,        mm5 \
 __asm movq       mm1,        mm3 \
 __asm pand       mm3,        mm5 \
 __asm pand       mm1,        mm4 \
 __asm movq       mm5,        mm6 \
 __asm psubw      mm2,        mm3 \
 __asm ps##d2##lq mm5,        32 \
 __asm pxor       mm4,        mm4 \
 __asm paddw      mm0,        mm3 \
 __asm movq       mm3,        [r1] \
 __asm por        mm1,        mm2 \
 __asm movq       [r0],       mm0 \
 __asm movq       mm2,        [r1+off] \
 __asm movq       mm0,        mm3 \
 __asm movq       [r0+off],   mm1 \
 __asm pcmpeqw    mm4,        mm5 \
 __asm movq       mm1,        mm2 \
 __asm pand       mm3,        mm5 \
 __asm pand       mm2,        mm5 \
 __asm pand       mm0,        mm4 \
 __asm paddw      mm1,        mm3 \
 __asm psubw      mm2,        mm3 \
 __asm movq       [r1+off],   mm1 \
 __asm por        mm0,        mm2 \
 __asm op0        r0,         (inc*2) \
 __asm movq       [r1],       mm0 \
 __asm op1        r1,         (inc*2)

   /* registers used: %r0    = data+? -> op0(inc)
                      %r1    = data+? -> op1(inc)
                      %%mm6 = 0xFFFF0000FFFF0000
                      %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code    dest        src */
# define SR_BF_k_eq_0_2I(inc, r0, op0, r1, op1, off) \
 __asm movq       mm2,        [r0] \
 __asm movq       mm3,        [r0+off] \
 __asm movq       mm0,        mm2 \
 __asm movq       mm1,        mm3 \
 __asm pand       mm2,        mm7 \
 __asm pand       mm3,        mm7 \
 __asm movq       mm4,        [r1] \
 __asm pand       mm1,        mm6 \
 __asm psubw      mm2,        mm3 \
 __asm movq       mm5,        [r1+off] \
 __asm paddw      mm0,        mm3 \
 __asm por        mm1,        mm2 \
 __asm movq       mm3,        mm4 \
 __asm movq       [r0],       mm0 \
 __asm movq       mm2,        mm5 \
 __asm pand       mm3,        mm6 \
 __asm movq       [r0+off],   mm1 \
 __asm pand       mm2,        mm6 \
 __asm pand       mm4,        mm7 \
 __asm psubw      mm2,        mm3 \
 __asm paddw      mm5,        mm3 \
 __asm por        mm4,        mm2 \
 __asm movq       [r1+off],   mm5 \
 __asm op0        r0,         (inc*2) \
 __asm movq       [r1],       mm4 \
 __asm op1        r1,         (inc*2)

   /* registers used: (edi)%4 = 0x5A815A81 */
   /*  op_code    dest        src */
# define SR_BF_k_eq_N8_SETUP \
 __asm movd       mm2,        edi \
 __asm pxor       mm3,        mm3 \
 __asm punpcklwd  mm3,        mm2

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm3 = 0x5A8100005A810000
                      %%mm6 = 0xFFFF0000FFFF0000
                      %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code    dest        src */
# define SR_BF_k_eq_N8_1R(inc, r0, op0, r1, op1, off, d1, d2, sc1, sc2) \
 __asm movq       mm1,        [r1+off] \
 __asm ps##d2##lq mm3,        32 \
 __asm movq       mm0,        [r0] \
 __asm pand       mm1,        mm7 \
 __asm pand       mm0,        mm6 \
 __asm ps##d2##lq mm1,        sc1 \
 __asm movq       mm2,        mm0 \
 __asm movq       mm4,        [r0+off] \
 __asm paddw      mm0,        mm1 \
 __asm movq       mm5,        [r1] \
 __asm psubw      mm1,        mm2 \
 __asm pand       mm5,        mm7 \
 __asm pand       mm4,        mm6 \
 __asm ps##d2##lq mm5,        sc1 \
 __asm movq       mm2,        mm4 \
 __asm psubw      mm4,        mm5 \
 __asm paddw      mm5,        mm2 \
 __asm movq       mm2,        mm0 \
 __asm paddw      mm0,        mm4 \
 __asm psubw      mm4,        mm2 \
 __asm pmulhw     mm0,        mm3 \
 __asm movq       mm2,        mm1 \
 __asm pmulhw     mm4,        mm3 \
 __asm paddw      mm1,        mm5 \
 __asm psllw      mm0,        1 \
 __asm psubw      mm5,        mm2 \
 __asm psllw      mm4,        1 \
 __asm movq       mm2,        mm3 \
 __asm pmulhw     mm1,        mm3 \
 __asm psrlw      mm2,        14 \
 __asm pmulhw     mm5,        mm3 \
 __asm movq       mm3,        mm2 \
 __asm ps##d1##lq mm2,        sc1 \
 __asm por        mm0,        mm3 \
 __asm ps##d1##lq mm1,        sc2 \
 __asm por        mm4,        mm3 \
 __asm ps##d1##lq mm5,        sc2 \
 __asm por        mm1,        mm2 \
 __asm por        mm5,        mm2 \
 __asm pxor       mm2,        mm2 \
 __asm pand       mm1,        mm7 \
 __asm pcmpeqw    mm2,        mm3 \
 __asm movq       mm3,        [r0] \
 __asm pand       mm5,        mm7 \
 __asm pand       mm3,        mm2 \
 __asm pand       mm2,        [r0+off] \
 __asm por        mm0,        mm3 \
 __asm movq       mm3,        mm7 \
 __asm por        mm4,        mm2 \
 __asm ps##d1##lq mm3,        32 \
 __asm pxor       mm2,        mm2 \
 __asm movq       [r0],       mm0 \
 __asm pcmpeqw    mm3,        mm2 \
 __asm movq       mm0,        [r1] \
 __asm movq       [r0+off],   mm4 \
 __asm pand       mm0,        mm3 \
 __asm pand       mm3,        [r1+off] \
 __asm por        mm5,        mm0 \
 __asm por        mm1,        mm3 \
 __asm movq       [r1],       mm5 \
 __asm op0        r0,         (inc*2) \
 __asm movq       [r1+off],   mm1 \
 __asm op1        r1,         (inc*2)

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm3 = 0x5A8100005A810000
                      %%mm6 = 0xFFFF0000FFFF0000
                      %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code    dest        src */
# define SR_BF_k_eq_N8_2R(inc, r0, op0, r1, op1, off) \
 __asm movq       mm1,        [r1+off] \
 __asm movq       mm0,        [r0] \
 __asm movq       mm2,        mm1 \
 __asm pand       mm1,        mm7 \
 __asm psllq      mm2,        48 \
 __asm movq       mm5,        [r1] \
 __asm psrlq      mm1,        16 \
 __asm pand       mm0,        mm6 \
 __asm por        mm1,        mm2 \
 __asm movq       mm2,        mm0 \
 __asm paddw      mm0,        mm1 \
 __asm psubw      mm1,        mm2 \
 __asm movq       mm2,        mm5 \
 __asm pand       mm5,        mm7 \
 __asm psllq      mm2,        48 \
 __asm movq       mm4,        [r0+off] \
 __asm psrlq      mm5,        16 \
 __asm pand       mm4,        mm6 \
 __asm por        mm5,        mm2 \
 __asm movq       mm2,        mm4 \
 __asm psubw      mm4,        mm5 \
 __asm paddw      mm5,        mm2 \
 __asm movq       mm2,        mm0 \
 __asm paddw      mm0,        mm4 \
 __asm psubw      mm4,        mm2 \
 __asm pmulhw     mm0,        mm3 \
 __asm movq       mm2,        mm1 \
 __asm pmulhw     mm4,        mm3 \
 __asm paddw      mm1,        mm5 \
 __asm psubw      mm5,        mm2 \
 __asm pmulhw     mm1,        mm3 \
 __asm movq       mm2,        mm3 \
 __asm psllw      mm0,        1 \
 __asm psrlw      mm2,        14 \
 __asm psllw      mm4,        1 \
 __asm por        mm0,        mm2 \
 __asm pmulhw     mm5,        mm3 \
 __asm por        mm4,        mm2 \
 __asm psllw      mm1,        1 \
 __asm psllw      mm5,        1 \
 __asm por        mm1,        mm2 \
 __asm por        mm5,        mm2 \
 __asm movq       mm2,        [r0] \
 __asm pand       mm2,        mm7 \
 __asm por        mm0,        mm2 \
 __asm movq       mm2,        [r0+off] \
 __asm movq       [r0],       mm0 \
 __asm pand       mm2,        mm7 \
 __asm movq       mm0,        mm5 \
 __asm psllq      mm5,        16 \
 __asm por        mm4,        mm2 \
 __asm psrlq      mm0,        48 \
 __asm movq       [r0+off],   mm4 \
 __asm por        mm5,        mm0 \
 __asm movq       mm4,        mm1 \
 __asm psllq      mm1,        16 \
 __asm movq       mm0,        [r1] \
 __asm psrlq      mm4,        48 \
 __asm pand       mm0,        mm6 \
 __asm movq       mm2,        [r1+off] \
 __asm por        mm1,        mm4 \
 __asm pand       mm2,        mm6 \
 __asm por        mm5,        mm0 \
 __asm por        mm1,        mm2 \
 __asm movq       [r1],       mm5 \
 __asm op0        r0,         (inc*2) \
 __asm movq       [r1+off],   mm1 \
 __asm op1        r1,         (inc*2)

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                 (ecx)%2    = IDFT_CONST
                      %%mm6 = 0xFFFF0000FFFF0000
                      %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code    dest        src */
# define SR_BF_1R(inc, r0, op0, r1, op1, off, h1, h2, d1, d2) \
 __asm movq       mm0,        [r0] \
 __asm movq       mm2,        [r1+off] \
 __asm pand       mm0,        mm6 \
 __asm movq       mm1,        [r0+off] \
 __asm pand       mm2,        mm7 \
 __asm movq       mm3,        [r1] \
 __asm punpck##h1##dq mm0,        mm0 \
 __asm pand       mm1,        mm6 \
 __asm punpck##h2##dq mm2,        mm2 \
 __asm pand       mm3,        mm7 \
 __asm punpck##h1##dq mm1,        mm1 \
 __asm movq       mm4,        [ecx] \
 __asm punpck##h2##dq mm3,        mm3 \
 __asm por        mm2,        mm0 \
 __asm movq       mm5,        [ecx+8] \
 __asm pmaddwd    mm2,        mm4 \
 __asm por        mm1,        mm3 \
 __asm movq       mm4,        mm6 \
 __asm pmaddwd    mm1,        mm5 \
 __asm movq       mm5,        mm7 \
 __asm movq       mm0,        mm1 \
 __asm psrld      mm5,        2 \
 __asm paddd      mm1,        mm2 \
 __asm psubd      mm0,        mm2 \
 __asm paddd      mm1,        mm5 \
 __asm paddd      mm0,        mm5 \
 __asm pslld      mm1,        1 \
 __asm pxor       mm3,        mm3 \
 __asm pslld      mm0,        1 \
 __asm movq       mm2,        [r0] \
 __asm ps##d1##lq mm4,        32 \
 __asm movq       mm5,        mm1 \
 __asm psrlq      mm1,        16 \
 __asm pcmpeqw    mm3,        mm4 \
 __asm pand       mm5,        mm4 \
 __asm pand       mm2,        mm3 \
 __asm pand       mm4,        mm0 \
 __asm pand       mm3,        [r0+off] \
 __asm psrlq      mm0,        16 \
 __asm por        mm2,        mm4 \
 __asm por        mm5,        mm3 \
 __asm movq       mm4,        mm7 \
 __asm movq       [r0],       mm2 \
 __asm ps##d2##lq mm4,        32 \
 __asm pxor       mm3,        mm3 \
 __asm movq       [r0+off],   mm5 \
 __asm pcmpeqw    mm3,        mm4 \
 __asm movq       mm2,        [r1+off] \
 __asm pand       mm0,        mm4 \
 __asm pand       mm1,        mm4 \
 __asm pand       mm2,        mm3 \
 __asm pand       mm3,        [r1] \
 __asm por        mm0,        mm2 \
 __asm por        mm1,        mm3 \
 __asm movq       [r1+off],   mm0 \
 __asm op0        r0,         (inc*2) \
 __asm movq       [r1],       mm1 \
 __asm op1        r1,         (inc*2)

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                 (ecx)%2    = IDFT_CONST
                      %%mm6 = 0xFFFF0000FFFF0000
                      %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code    dest        src */
# define SR_BF_2R(inc, r0, op0, r1, op1, off, h1, h2) \
 __asm movq       mm0,        [r0] \
 __asm movq       mm1,        [r1+off] \
 __asm pand       mm0,        mm6 \
 __asm pand       mm1,        mm7 \
 __asm movq       mm3,        [r1] \
 __asm movq       mm2,        mm1 \
 __asm pand       mm3,        mm7 \
 __asm psllq      mm1,        32 \
 __asm movq       mm5,        mm3 \
 __asm psrlq      mm2,        32 \
 __asm por        mm0,        mm1 \
 __asm psllq      mm3,        32 \
 __asm por        mm0,        mm2 \
 __asm movq       mm4,        [ecx] \
 __asm psrlq      mm5,        32 \
 __asm movq       mm2,        [r0+off] \
 __asm por        mm3,        mm5 \
 __asm movq       mm5,        mm4 \
 __asm punpck##h1##dq mm4,        mm4 \
 __asm pand       mm2,        mm6 \
 __asm punpck##h2##dq mm5,        mm5 \
 __asm movq       mm1,        mm0 \
 __asm pmaddwd    mm0,        mm4 \
 __asm movq       mm4,        [ecx+8] \
 __asm por        mm2,        mm3 \
 __asm pmaddwd    mm1,        mm5 \
 __asm movq       mm5,        mm4 \
 __asm punpck##h1##dq mm4,        mm4 \
 __asm movq       mm3,        mm2 \
 __asm pmaddwd    mm2,        mm4 \
 __asm punpck##h2##dq mm5,        mm5 \
 __asm movq       mm4,        mm7 \
 __asm pmaddwd    mm3,        mm5 \
 __asm psrlw      mm4,        2 \
 __asm paddd      mm3,        mm4 \
 __asm paddd      mm2,        mm4 \
 __asm movq       mm5,        mm3 \
 __asm psubd      mm3,        mm1 \
 __asm paddd      mm1,        mm5 \
 __asm movq       mm4,        mm2 \
 __asm movq       mm5,        mm1 \
 __asm psubd      mm2,        mm0 \
 __asm psllq      mm1,        17 \
 __asm paddd      mm0,        mm4 \
 __asm psrlq      mm5,        47 \
 __asm movq       mm4,        mm3 \
 __asm pslld      mm2,        1 \
 __asm por        mm1,        mm5 \
 __asm psllq      mm3,        17 \
 __asm movq       mm5,        [r0+off] \
 __asm pslld      mm0,        1 \
 __asm pand       mm2,        mm6 \
 __asm psrlq      mm4,        47 \
 __asm pand       mm0,        mm6 \
 __asm pand       mm5,        mm7 \
 __asm por        mm3,        mm4 \
 __asm movq       mm4,        [r0] \
 __asm por        mm0,        mm5 \
 __asm pand       mm4,        mm7 \
 __asm movq       [r0+off],   mm0 \
 __asm por        mm2,        mm4 \
 __asm movq       mm0,        [r1] \
 __asm pand       mm1,        mm7 \
 __asm pand       mm0,        mm6 \
 __asm movq       [r0],       mm2 \
 __asm movq       mm2,        [r1+off] \
 __asm pand       mm3,        mm7 \
 __asm pand       mm2,        mm6 \
 __asm por        mm1,        mm0 \
 __asm por        mm3,        mm2 \
 __asm movq       [r1],       mm1 \
 __asm op0        r0,         (inc*2) \
 __asm movq       [r1+off],   mm3 \
 __asm op1        r1,         (inc*2)

   /* registers used: %r0 = data+? -> add(%4)
                 (ecx)%2  = IDFT_CONST
                 (edi)%4  = -> (%2) */
   /*  op_code    dest        src */
# define SWAP8_1(r0) \
 __asm movq       mm0,        [r0] \
 __asm movq       mm1,        [r0+32] \
 __asm movq       mm2,        mm0 \
 __asm movq       mm4,        [r0+256] \
 __asm movq       mm3,        mm1 \
 __asm punpckldq  mm0,        mm4 \
 __asm movq       mm5,        [r0+288] \
 __asm movq       [r0],       mm0 \
 __asm punpckldq  mm1,        mm5 \
 __asm mov        edi,        [ecx] \
 __asm punpckhdq  mm2,        mm4 \
 __asm movq       [r0+32],    mm1 \
 __asm punpckhdq  mm3,        mm5 \
 __asm movq       [r0+256],   mm2 \
 __asm add        ecx,        4 \
 __asm movq       [r0+288],   mm3 \
 __asm add        r0,         edi

   /* registers used: %r0 = data+? -> sub(%4)
                      %r1 = data+? -> sub(%4)
                 (ecx)%2  = IDFT_CONST
                 (edi)%4  = -> (%2) */
   /*  op_code    dest        src */
# define SWAP8_2(r0, r1) \
 __asm movq       mm0,        [r1] \
 __asm movq       mm1,        [r0] \
 __asm movq       mm2,        mm0 \
 __asm movq       mm4,        [r1+256] \
 __asm movq       mm3,        mm1 \
 __asm punpckldq  mm0,        mm4 \
 __asm movq       mm5,        [r0+256] \
 __asm movq       [r0],       mm0 \
 __asm punpckldq  mm1,        mm5 \
 __asm mov        edi,        [ecx] \
 __asm punpckhdq  mm2,        mm4 \
 __asm movq       [r1],       mm1 \
 __asm punpckhdq  mm3,        mm5 \
 __asm movq       [r0+256],   mm2 \
 __asm sub        r0,         edi \
 __asm mov        edi,        [ecx+4] \
 __asm add        ecx,        8 \
 __asm movq       [r1+256],   mm3 \
 __asm sub        r1,         edi

   /* registers used: (eax)%0 = data+? -> add(inc)
                      (ebx)%1 = data+? -> sub(inc)
                      (ecx)%2 = IDFT_CONST
                         %mm7 = 0x0001000100010001 */
   /*  op_code    dest        src */
# define PRE_POST_SCALING(inc, m1, m2) \
 __asm movq       mm0,        [eax] \
 __asm movq       mm1,        [ebx] \
 __asm movq       mm4,        mm0 \
 __asm movq       mm2,        mm0 \
 __asm psllq      mm4,        32 \
 __asm movq       mm5,        mm1 \
 __asm punpckhwd  mm2,        mm4 \
 __asm movq       mm3,        mm1 \
 __asm psllq      mm5,        32 \
 __asm movq       mm4,        mm2 \
 __asm punpckhwd  mm3,        mm5 \
 __asm psllq      mm4,        32 \
 __asm movq       mm5,        mm3 \
 __asm punpckhwd  mm2,        mm4 \
 __asm psllq      mm5,        32 \
 __asm movq       mm4,        [ecx] \
 __asm punpckhwd  mm3,        mm5 \
 __asm movq       mm5,        [ecx+8] \
 __asm pmulhw     mm0,        mm4 \
 __asm movq       mm6,        mm4 \
 __asm pmulhw     mm3,        mm5 \
 __asm psllq      mm6,        32 \
 __asm paddw      mm0,        mm3 \
 __asm punpckhwd  mm4,        mm6 \
 __asm movq       mm3,        mm5 \
 __asm movq       mm6,        mm4 \
 __asm psllq      mm3,        32 \
 __asm psllq      mm6,        32 \
 __asm punpckhwd  mm5,        mm3 \
 __asm punpckhwd  mm4,        mm6 \
 __asm movq       mm3,        mm5 \
 __asm pmulhw     mm1,        mm4 \
 __asm psllq      mm3,        32 \
 __asm punpckhwd  mm5,        mm3 \
 __asm paddw      mm0,        mm7 \
 __asm pmulhw     mm2,        mm5 \
 __asm add        ecx,        16 \
 __asm psllw      mm0,        1 \
 __asm psubw      mm##m1,     mm##m2 \
 __asm movq       [eax],      mm0 \
 __asm psllw      mm##m1,     1 \
 __asm add        eax,        (inc*2) \
 __asm movq       [ebx],      mm##m1 \
 __asm sub        ebx,        (inc*2)


void AC3_imdft512(AC3_COEFF *data)
{
   __asm {
   /* input list
   eax %0: "q" (data),
   ebx %1  "q" (data+252),
   ecx %2  "q" (MMX_IDFT_CONST),
   esi %3  "r" (32),
   edi %4  "r" (0x5A815A81) */
   /*  op_code    dest        src */
       mov        eax,        dword ptr[data]
       mov        ecx,        offset MMX_IDFT_CONST
       mov        ebx,        eax
       mov        esi,        32
       add        ebx,        (252*2)
       mov        edi,        0x5A815A81

   /* Pre-IDFT Scaling:
      registers used: %0 = data
                      %1 = data+252
                      %3 = 32 */
   /*  op_code    dest        src */
       pcmpeqd    mm7,        mm7
       psrlw      mm7,        15          /* mm7 = 0x0001000100010001 */
      /* %0 = data      %1 = data+252 */
   PRE_LOOP:
      PRE_POST_SCALING(4, 1, 2)
       dec        esi
       jnz        PRE_LOOP
      /* %0 = data+128  %1 = data+124 */

   /* INIT MASKS */
   /*  op_code    dest        src */
       pcmpeqd    mm7,        mm7
       pxor       mm6,        mm6
       movq       mm5,        mm7
       punpcklwd  mm7,        mm6         /* mm7 = 0x0000FFFF0000FFFF */
       punpcklwd  mm6,        mm5         /* mm6 = 0xFFFF0000FFFF0000 */
   
   /* STAGE 1: 2pt-DFTs
      registers used: %0 = data
                      %1 = data+128 
                      %3 = 32 */
   /*  op_code    dest        src */
       sub        eax,        (128*2)
       add        ebx,        (4*2)
       add        esi,        32
      /* %0 = data      %1 = data+128 */
   STAGE1_LOOP:
      STAGE1
       dec        esi
       jnz        STAGE1_LOOP
      /* %0 = data+128  %1 = data+256 */

   /* STAGE 2: 4pt-DFTs
      registers used: %0 = data
                      %1 = data+188
                      %3 = 4 */
   /*  op_code    dest        src */
       sub        eax,        (128*2)
       sub        ebx,        (68*2)
       add        esi,        4
      /* %0 = data      %1 = data+188 */
      STAGE_OUTPUT_1R2I(4, eax, add, ebx, sub, 128, l, r, 48)
      STAGE_OUTPUT___2I(4, eax, add, ebx, sub, 128)
      STAGE_OUTPUT_2R2I(4, eax, add, ebx, sub, 128)
      STAGE_OUTPUT_2R2I(4, eax, add, ebx, sub, 128)
   STAGE2_LOOP1:
      STAGE_OUTPUT___2I(4, eax, add, ebx, sub, 128)
       dec        esi
       jnz        STAGE2_LOOP1
       add        esi,        4
   STAGE2_LOOP2:
      STAGE_OUTPUT_2R__(4, eax, add, ebx, sub, 128)
       dec        esi
       jnz        STAGE2_LOOP2
      STAGE_OUTPUT_2R2I(4, eax, add, ebx, sub, 128)
      STAGE_OUTPUT_2R2I(4, eax, add, ebx, sub, 128)
      STAGE_OUTPUT_2R__(4, eax, add, ebx, sub, 128)
      STAGE_OUTPUT_2R1I(0, eax, add, ebx, sub, 128, l, r)
      /* %0 = data+60   %1 = data+128 */

   /* STAGE 3: 8pt-DFTs
      registers used: %0 = data+60
                      %1 = data+128
                      %3 = 4
                      %4 = 0x5A815A81 */
   /*  op_code    dest        src */
       add        esi,        4
      /* 1, 3 */
      /* %0 = data+60   %1 = data+128 */
      SR_BF_k_eq_0_1I( 4, eax, sub, ebx, add, 128, l, r)
      SR_BF_k_eq_0_2I(12, eax, sub, ebx, add, 128)
   STAGE3_LOOP1:
      SR_BF_k_eq_0_2I( 4, eax, sub, ebx, add, 128)
       dec        esi
       jnz        STAGE3_LOOP1
      /* 7, 5 */
      /* %0 = data+28   %1 = data+160 */
       add        esi,        4
      SR_BF_k_eq_N8_SETUP
   STAGE3_LOOP2:
      SR_BF_k_eq_N8_2R(4, eax, sub, ebx, add, 128)
       dec        esi
       jnz        STAGE3_LOOP2
       sub        eax,        (8*2)
       add        ebx,        (8*2)
      SR_BF_k_eq_N8_2R(4, eax, sub, ebx, add, 128)
      SR_BF_k_eq_N8_1R(0, eax, sub, ebx, add, 128, l, r, 16, 17)
      /* %0 = data      %1 = data+188 */
      
       add        ebx,        (32*2)
       add        esi,        2
      /* %0 = data      %1 = data+220 */
   STAGE3_LOOP3:
      STAGE_OUTPUT_1R2I(  4, eax, add, ebx, sub, 64, r, l, 16)
      STAGE_OUTPUT_2R2I(  4, eax, add, ebx, sub, 64)
      STAGE_OUTPUT___2I(  4, eax, add, ebx, sub, 64)
      STAGE_OUTPUT___2I(  4, eax, add, ebx, sub, 64)
      STAGE_OUTPUT_2R__(  4, eax, add, ebx, sub, 64)
      STAGE_OUTPUT_2R__(  4, eax, add, ebx, sub, 64)
      STAGE_OUTPUT_2R2I(  4, eax, add, ebx, sub, 64)
      STAGE_OUTPUT_2R1I(100, eax, add, ebx, sub, 64, r, l)
       dec        esi
       jnz        STAGE3_LOOP3
      /* %0 = data+256  %1 = data-36 */

   /* STAGE 4: 16pt-DFTs
      registers used: %0 = data
                      %1 = data+220
                      %2 = IDFT_CONST
                      %3 = 4
                      %4 = 0x5A815A81 */
   /*  op_code    dest        src */
       sub        eax,        (256*2)
       add        ebx,        (256*2)
       add        esi,        4
      /* 15, 13 */
      /* %0 = data      %1 = data+220 */
      SR_BF_1R( 8, eax, add, ebx, sub, 64, h, l, l, r)
      SR_BF_2R( 4, eax, add, ebx, sub, 64, h, l)
      SR_BF_2R(52, eax, add, ebx, sub, 64, h, l)
       add        ecx,        (8*2)
      /* 11,  9 */
      /* %0 = data+64   %1 = data+156 */
      SR_BF_1R( 8, eax, add, ebx, sub, 64, h, l, l, r)
      SR_BF_2R( 4, eax, add, ebx, sub, 64, h, l)
      SR_BF_2R(64, eax, add, ebx, sub, 64, h, l)
      /* 7, 5 */
      /* %0 = data+140  %1 = data+80 */
      SR_BF_k_eq_N8_SETUP
      SR_BF_k_eq_N8_2R( 4, eax, sub, ebx, add, 64)
      SR_BF_k_eq_N8_2R( 8, eax, sub, ebx, add, 64)
      SR_BF_k_eq_N8_1R(76, eax, add, ebx, sub, 64, r, l, 48, 47)
      /* 3, 1 */
      /* %0 = data+204  %1 = data+16 */
      SR_BF_k_eq_0_2I( 4, ebx, add, eax, sub, 64)
      SR_BF_k_eq_0_2I( 8, ebx, add, eax, sub, 64)
      SR_BF_k_eq_0_1I(28, ebx, sub, eax, add, 64, r, l)
      /* %0 = data+220  %1 = data */
      
       add        eax,        (16*2)
      /* %0 = data+236  %1 = data */
   STAGE4_LOOP1:
      STAGE_OUTPUT_1R2I( 4, ebx, add, eax, sub, 32, l, r, 48)
      STAGE_OUTPUT___2I( 4, ebx, add, eax, sub, 32)
      STAGE_OUTPUT_2R__( 4, ebx, add, eax, sub, 32)
      STAGE_OUTPUT_2R1I(52, ebx, add, eax, sub, 32, l, r)
       dec        esi
       jnz        STAGE4_LOOP1
      /* %0 = data-20   %1 = data+256 */

   /* STAGE 5: 32pt-DFTs
      registers used: %0 = data
                      %1 = data+236
                      %2 = IDFT_CONST+16
                      %3 = 6
                      %4 = 0x5A815A81 */
   /*  op_code    dest        src */
       add        eax,        (20*2)
       sub        ebx,        (20*2)
       add        ecx,        (8*2)
       add        esi,        6
      /* 31, 29
          :   :
         11,  9 */
      /* %0 = data      %1 = data+236 */
   STAGE5_LOOP1:
      SR_BF_1R( 4, eax, add, ebx, sub, 32, l, h, r, l)
      SR_BF_2R(28, eax, add, ebx, sub, 32, l, h)
       add        ecx,        (8*2)
       dec        esi
       jnz        STAGE5_LOOP1
      /* %0 = data+192  %1 = data+44 */
      /* 7, 5 */
       add        eax,        (4*2)
       sub        ebx,        (4*2)
      /* %0 = data+196  %1 = data+40 */
      SR_BF_k_eq_N8_SETUP
      SR_BF_k_eq_N8_2R( 4, eax, sub, ebx, add, 32)
      SR_BF_k_eq_N8_1R(36, eax, add, ebx, sub, 32, l, r, 16, 17)
      /* 3, 1 */
      /* %0 = data+228  %1 = data+8 */
      SR_BF_k_eq_0_2I( 4, ebx, add, eax, sub, 32)
      SR_BF_k_eq_0_1I(12, ebx, sub, eax, add, 32, l, r)
      /* %0 = data+236  %1 = data+0 */

       add        eax,        (8*2)
       add        esi,        8
      /* %0 = data+244  %1 = data */
   STAGE5_LOOP2:
      STAGE_OUTPUT_1R2I( 4, ebx, add, eax, sub, 16, r, l, 16)
      STAGE_OUTPUT_2R1I(28, ebx, add, eax, sub, 16, r, l)
       dec        esi
       jnz        STAGE5_LOOP2
      /* %0 = data-12   %1 = data+256 */

   /* STAGE 6: 64pt-DFTs
      registers used: %0 = data+4
                      %1 = data+240
                      %2 = IDFT_CONST+64
                      %3 = 14
                      %4 = 0x5A815A81 */
   /*  op_code    dest        src */
       add        eax,        (16*2)
       sub        ebx,        (16*2)
       add        esi,        14
      /* 1, 3 */
      /* %0 = data+4    %1 = data+240 */
      SR_BF_k_eq_0_1I(16, eax, add, ebx, sub, 16, r, l)
      /* 5, 7 */
      /* %0 = data+20   %1 = data+224 */
      SR_BF_k_eq_N8_SETUP
      SR_BF_k_eq_N8_1R(16, ebx, sub, eax, add, 16, r, l, 48, 47)
      /*  9, 11
          :   :
         61, 63  */
      /* %0 = data+36   %1 = data+208 */
   STAGE6_LOOP1:
      SR_BF_1R(16, ebx, sub, eax, add, 16, h, l, l, r)
       add        ecx,        (8*2)
       dec        esi
       jnz        STAGE6_LOOP1
      /* %0 = data+260  %1 = data-16 */

       add        ebx,        (16*2)
       movq       mm5,        mm7
       pcmpgtd    mm4,        mm4
       psrlq      mm5,        32
       add        esi,        16
       pcmpeqw    mm4,        mm5
      /* %0 = data+260  %1 = data */
   STAGE6_LOOP2:
      STAGE6_OUTPUT_01(16, ebx, add, [ebx], [ebx+8])
       dec        esi
       jnz        STAGE6_LOOP2
      /* %0 = data+260  %1 = data+256 */
       psrlq      mm4,        16
       sub        eax,        (12*2)
       add        esi,        16
       psllq      mm5,        48
      /* %0 = data+248  %1 = data+256 */
   STAGE6_LOOP3:
      STAGE6_OUTPUT_01(16, eax, sub, [eax+8], [eax])
       dec        esi
       jnz        STAGE6_LOOP3
      /* %0 = data-8    %1 = data+256 */
       movq       mm4,        mm5
       add        eax,        (16*2)
       sub        ebx,        (16*2)
       psrlq      mm4,        48
       add        esi,        16
      /* %0 = data+8    %1 = data+240 */
   STAGE6_LOOP4:
      STAGE6_OUTPUT_23(16, eax, add, ebx, sub)
       dec        esi
       jnz        STAGE6_LOOP4
      /* %0 = data+264  %1 = data-16 */

   /* STAGE 7: 128pt-DFTs
      registers used: %0 = data+0
                      %1 = data+248
                      %2 = IDFT_CONST+176
                      %3 = 30
                      %4 = 0x5A815A81 */
   /*  op_code    dest        src */
       sub        eax,        (264*2)
       add        ebx,        (264*2)
       add        esi,        30
      /* 1, 3 */
      /* %0 = data+0    %1 = data+248 */
      SR_BF_k_eq_0_1I(8, eax, add, ebx, sub, 8, l, r)
      /* 5, 7 */
      /* %0 = data+8    %1 = data+240 */
      SR_BF_k_eq_N8_SETUP
      SR_BF_k_eq_N8_1R(8, ebx, sub, eax, add, 8, l, r, 16, 17)
      /*  9, 11
          :   :
        125,127  */
      /* %0 = data+16   %1 = data+232 */
   STAGE7_LOOP1:
      SR_BF_1R(8, ebx, sub, eax, add, 8, l, h, r, l)
       add        ecx,        (8*2)
       dec        esi
       jnz        STAGE7_LOOP1
      /* %0 = data+256  %1 = data-8 */

       add        ebx,        (8*2)
       sub        eax,        (4*2)
       add        esi,        16
      /* %0 = data+252  %1 = data */
   STAGE7_LOOP2:
      STAGE7_OUTPUT(4, ebx, add, eax, sub)
      STAGE7_OUTPUT(4, eax, sub, ebx, add)
       dec        esi
       jnz        STAGE7_LOOP2
      /* %0 = data+124  %1 = data+128 */

   /* Undo Bit-Reverse Order
      registers used: %0 = data+0
                      %1 = data+92
                      %2 = IDFT_CONST+416
                      %3 = 4 */
   /*  op_code    dest        src */
       sub        eax,        (124*2)
       sub        ebx,        (36*2)
       add        esi,        4
      /* %0 = data+0    %1 = data+92 */
   SWAP_LOOP1:
      SWAP8_1(eax)
       dec        esi
       jnz        SWAP_LOOP1
      /* %0 = data+116  %1 = data+92 */
       add        esi,        12
   SWAP_LOOP2:
      SWAP8_2(ebx, eax)
       dec        esi
       jnz        SWAP_LOOP2

   /* Post-IDFT Scaling:
      registers used: %0 = data
                      %1 = data+252
                      %3 = 32 */
   /*  op_code    dest        src */
       pcmpeqd    mm7,        mm7
       add        esi,        32
       psrlw      mm7,        15          /* mm7 = 0x0001000100010001 */
      /* %0 = data      %1 = data+252 */
   POST_LOOP:
      PRE_POST_SCALING(4, 2, 1)
       dec        esi
       jnz        POST_LOOP
      
       emms
   }
}

#undef STAGE_OUTPUT_1R2I
#undef STAGE_OUTPUT___2I
#undef STAGE_OUTPUT_2R2I
#undef STAGE_OUTPUT_2R__
#undef STAGE6_OUTPUT_01
#undef STAGE6_OUTPUT_23
#undef STAGE7_OUTPUT
#undef SR_BF_k_eq_0_1I
#undef SR_BF_k_eq_N8_SETUP
#undef SR_BF_k_eq_N8_1R
#undef SR_BF_k_eq_N8_2R
#undef SR_BF_1R
#undef SR_BF_2R
#undef SWAP8_1
#undef SWAP8_2
#undef PRE_POST_SCALING

