/***************************************************************************
                          Timer.h  -  description
                             -------------------
   These functions provide access to the high resolution clock.
                             -------------------
    begin                : Sat Jul 17 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __TIMER_H__
#define __TIMER_H__

#ifdef WIN32
# ifndef WIN32_LEAN_AND_MEAN
#  define WIN32_LEAN_AND_MEAN
# endif
# include <windows.h>

#ifdef __cplusplus
extern "C" {
#endif

/************************Start WIN32 Version************************/
typedef struct {
   long tv_sec,
        tv_nsec;
} TIME;


/* Get the resolution of this TIMER.
   Returns 0 if successful. */
int TIMER_res(TIME *res);

/* Get the current time.
   Returns 0 if successful. */
int TIMER_curtime(TIME *ct);

#define TIMER_sleep(milli_sec)   Sleep((unsigned int)(milli_sec))

#ifdef __cplusplus
};
#endif
/*************************End WIN32 Version*************************/
#else
/************************Start POSIX Version************************/
#include <time.h>
#include <sys/time.h>
#include <unistd.h>

typedef struct timespec TIME;

/* Even though I am using macros here, the prototypes above still
   apply. So be careful when using these (you have been warned). */
#define TIMER_res(res)     clock_getres(CLOCK_REALTIME, res)
#define TIMER_curtime(ct)  clock_gettime(CLOCK_REALTIME, ct)

#define TIMER_sleep(milli_sec)   usleep((unsigned int)(milli_sec)*1000)

/*************************End POSIX Version*************************/
#endif

/* The rest of this file is shared between all versions. */
#ifdef __cplusplus
extern "C" {
#endif

/* Calculates the difference between t2 and t1 (t1 = t2 - t1).
   Result is stored in t1.
   Returns 0 if successful. */
int TIMER_diff(TIME *t1, const TIME *t2);

/* Calculates the sum of t1 and t2 (t1 += t2).
   Result is stored in t1.
   Returns 0 if successful. */
int TIMER_add(TIME *t1, const TIME *t2);

/* Returns the value of t as a double value */
double TIMER_getdbl(TIME *t);

#ifdef __cplusplus
};
#endif

#endif

