/***************************************************************************
                          Timer_Win32.c  -  description
                             -------------------
   These functions provide access to the high resolution clock.
                             -------------------
    begin                : Sat Jul 17 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "../typedefs.h"
#include "../Timer.h"

// Private Globals
static BOOL init      = TRUE;
static BOOL perfTimer = FALSE;
static s64  TimerFreq;  // Timer frequency (Hz)


/* Get the resolution of this TIMER.
   Returns 0 if successful. */
int TIMER_res(TIME *res)
{
   if (init) {
      if ( (!QueryPerformanceFrequency((LARGE_INTEGER *)&TimerFreq))||
           (TimerFreq < 1000) ) {
         // No Performace Counter Available
         perfTimer  = FALSE;  /* Performance Counter is not available */
         TimerFreq  = 1000;   /* Timer Frequency is 1khz */
      }
      else
         perfTimer = TRUE;    /* Performance Counter is available */
      init = FALSE;
   }

   if (!res)
      return -1;

   res->tv_sec  = 0;
   res->tv_nsec = (long)((s64)1000000000 / TimerFreq);
   return 0;
}

/* Get the current time.
   Returns 0 if successful. */
int TIMER_curtime(TIME *ct)
{
   s64 time;

   if (init)
      TIMER_res(0);

   if (!ct)
      return -1;
      
   if (perfTimer)
      QueryPerformanceCounter((LARGE_INTEGER *)&time);
   else
      time = (s64)GetTickCount();

   ct->tv_sec  = (long)(time / TimerFreq);
   time       -= (s64)ct->tv_sec * TimerFreq;
   time       *= (s64)1000000000;
   ct->tv_nsec = (long)(time / TimerFreq);
   return 0;
}
