/***************************************************************************
                          PAT_module.h  -  description
                             -------------------
   PAT PSI module functions.
                             -------------------
    begin                : Sat Aug 14 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __PAT_MODULE_H__
#define __PAT_MODULE_H__

#include <stddef.h>
#include "common/typedefs.h"
#include "PMT_SECTION.h"
#include "MPEG2_SYS.h"

/* Use this macro to initialize a MPEG2_PSI_PES_MODULE structure. */
#define PAT_PSI_MODULE \
{\
   PAT_PSI_MODULE_stream_init, \
   PAT_PSI_MODULE_process_pkt, \
   {\
      PAT_PSI_MODULE_mem_needed, \
      PAT_PSI_MODULE_init, \
      PAT_PSI_MODULE_done, \
      (MPEG2_MODULE_set_offset)NULL, \
      (MPEG2_MODULE_discontinuity)NULL, \
      NULL \
   }\
}

#ifdef __cplusplus
extern "C" {
#endif

/*****************************************************************
   GENERAL MODULE FUNCTIONS
     These functions are used by all SYSTEM, PSI, and PES modules.
 *****************************************************************/
/* Returns the number of bytes of memory needed by the module. */
size_t PAT_PSI_MODULE_mem_needed();

/* Initializes the memory provided for the module.
      INPUT: void *     - Pointer to the memory provided for the module.
             SYS_DATA * - Pointer to system data. */
void   PAT_PSI_MODULE_init(void *mem, SYS_DATA *sys_data);

/* Clean up (DO NOT FREE) the memory provided for the module.
      INPUT: void * - Pointer to the memory provided for the module. */
void   PAT_PSI_MODULE_done(void *mem);


/**************************************************************************
   COMMON PSI MODULE FUNCTIONS
     These functions are used by SYSTEM modules to process the PSI_SECTIONs
       they produce.
 **************************************************************************/
/* PSI modules should always return 0. */
int  PAT_PSI_MODULE_stream_init(void *mem, PMT_SECTION *pmt, int idx);

/* Processes the packet handled by the module.
      INPUT: void * - Pointer to the memory provided for the module.
             void * - Pointer to the PSI_SECTION that needs handling. */
void PAT_PSI_MODULE_process_pkt(void *mem, void *psi);


/*************************************************************************
   PAT MODULE FUNCTIONS
     These functions are used by the MPEG2 TS decoder to get the PAT data.
       The PAT data is extracted from the PSI_SECTIONs sent by a system
       module.
 *************************************************************************/
/* Adds an entry to the PAT, using the given values.  This function is
     used when there is no PAT data in the stream.
      INPUT: void * - Pointer to the memory provided for the module.
             u16    - Program number of the new entry. Any value greater
                        than 0 is valid.
             u16    - Program Map PID of the new entry. Any unused user
                        defined PID value is valid. */
void PAT_PSI_MODULE_add_entry(void *mem, u16 program_number, u16 PMT_PID);

/* Returns the PID of the TS_PKTs that contains the PMT data of the
     specified program number.
      INPUT: void * - Pointer to the memory provided for the module.
             u16    - Program number to search for.
      RETURN: Program found     - Valid user defined PID value.
              Program not found - PID value that is >= TS_PKT_NUL_ID. */
u16  PAT_PSI_MODULE_get_pid(void *mem, u16 program_number);

/* Returns the program number at the specified location, in the
     access table.
      INPUT: void * - Pointer to the memory provided for the module.
             u16    - Table index to search for.
      RETURN: Program found     - Valid program number.
              Program not found - 0 */
u16  PAT_PSI_MODULE_get_program(void *mem, u16 idx);

#ifdef __cplusplus
};
#endif

#endif
