/***************************************************************************
                         bitstream.h  -  description
                             -------------------
   Macros to help parse data that is not byte aligned.
                             -------------------
    begin                : Fri Oct 14 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef _BITSTREAM_H_
#define _BITSTREAM_H_

/*
Before using the following macros, make sure you have the
following macros defined:
   DATA_BYTE         -the byte where data is read from
   BIT_SHIFT_BYTE    -the number of bits to shift to align DATA_BYTE
   DATA_PTR          -pointer to the data that follows DATA_BYTE
   END_DATA_PTR      -pointer to the byte following the last byte to be read
   PARSER_STATE      -stores the location where the parser should continue parsing
   STOP_PARSING      -what to do when no more data is available
*/

/* Moves to the next byte in the data stream. */
#define NEXT_BYTE(resume_pt) \
   if (++(DATA_PTR) >= END_DATA_PTR) { \
      PARSER_STATE = resume_pt; \
      STOP_PARSING; \
   }


# ifndef _BITSTREAM_TRACE_

/* Fills in the low bits of DATA_BYTE.  Must be called after the
   ALIGN* macros, and before DATA_BYTE is read. */
#define GET_LSB \
   if (BIT_SHIFT_BYTE < 8) \
      DATA_BYTE |= *(DATA_PTR) >> (8 - BIT_SHIFT_BYTE); \
   else \
      DATA_BYTE  = *(DATA_PTR)


/* Re-aligns DATA_BYTE by discarding the high bits. */
#define ALIGN_DATA_BYTE(resume_pt, discard_bits) \
   BIT_SHIFT_BYTE += discard_bits; \
   if (BIT_SHIFT_BYTE > 8) { \
      BIT_SHIFT_BYTE -= 8; \
      DATA_BYTE       = *(DATA_PTR) << BIT_SHIFT_BYTE; \
      NEXT_BYTE(resume_pt) \
   } \
   else \
      DATA_BYTE <<= discard_bits

/* Re-aligns DATA_BYTE by discarding the entire byte. */
#define ALIGN8_DATA_BYTE(resume_pt) \
   DATA_BYTE = *(DATA_PTR) << BIT_SHIFT_BYTE; \
   NEXT_BYTE(resume_pt)

/* Re-aligns DATA_BYTE. */
#define ALIGN0_DATA_BYTE(resume_pt) \
   if (BIT_SHIFT_BYTE > 8) { \
      BIT_SHIFT_BYTE -= 8; \
      DATA_BYTE       = *(DATA_PTR) << BIT_SHIFT_BYTE; \
      NEXT_BYTE(resume_pt) \
   }

# else

# ifndef BIT_TRACE_BEG_BLK
#  define BIT_TRACE_BEG_BLK  {
# endif
# ifndef BIT_TRACE_END_BLK
#  define BIT_TRACE_END_BLK  }
# endif

#define GET_LSB \
   if (BIT_SHIFT_BYTE < 8) \
      DATA_BYTE |= *(DATA_PTR) >> (8 - BIT_SHIFT_BYTE); \
   else \
      DATA_BYTE  = *(DATA_PTR); \
BIT_TRACE_BEG_BLK \
printf("GET  > 0x%02hX,%hu (%02hX %02hX)\n", \
       (u16)DATA_BYTE, \
       (u16)BIT_SHIFT_BYTE, \
       (u16)*(DATA_PTR), \
       (u16)*(DATA_PTR+1)); \
BIT_TRACE_END_BLK


#define ALIGN_DATA_BYTE(resume_pt, discard_bits) \
   BIT_SHIFT_BYTE += discard_bits; \
   if (BIT_SHIFT_BYTE > 8) { \
      BIT_SHIFT_BYTE -= 8; \
      DATA_BYTE       = *(DATA_PTR) << BIT_SHIFT_BYTE; \
      NEXT_BYTE(resume_pt) \
   } \
   else \
      DATA_BYTE <<= discard_bits; \
BIT_TRACE_BEG_BLK \
printf("ALIGN> 0x%02hX,%hu (%02hX %02hX, +%hu)\n", \
      (u16)DATA_BYTE, \
      (u16)BIT_SHIFT_BYTE, \
      (u16)*(DATA_PTR), \
      (u16)*(DATA_PTR+1), \
      (u16)(discard_bits)); \
BIT_TRACE_END_BLK

#define ALIGN8_DATA_BYTE(resume_pt) \
   DATA_BYTE = *(DATA_PTR) << BIT_SHIFT_BYTE; \
   NEXT_BYTE(resume_pt) \
BIT_TRACE_BEG_BLK \
printf("ALIGN8> 0x%02hX,%hu (%02hX %02hX)\n", \
       (u16)DATA_BYTE, \
       (u16)BIT_SHIFT_BYTE, \
       (u16)*(DATA_PTR), \
       (u16)*(DATA_PTR+1)); \
BIT_TRACE_END_BLK

#define ALIGN0_DATA_BYTE(resume_pt) \
   if (BIT_SHIFT_BYTE > 8) { \
      BIT_SHIFT_BYTE -= 8; \
      DATA_BYTE       = *(DATA_PTR) << BIT_SHIFT_BYTE; \
      NEXT_BYTE(resume_pt) \
   } \
BIT_TRACE_BEG_BLK \
printf("ALIGN0> 0x%02hX,%hu (%02hX %02hX)\n", \
      (u16)DATA_BYTE, \
      (u16)BIT_SHIFT_BYTE, \
      (u16)*(DATA_PTR), \
      (u16)*(DATA_PTR+1)); \
BIT_TRACE_END_BLK

# endif

#endif

