/***************************************************************************
                        detect_mpeg.h  -  description
                             -------------------
    begin                : Sat Feb 5 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __DETECT_MPEG_H__
#define __DETECT_MPEG_H__

#include "common/typedefs.h"
#ifndef _NO_AUDIO_
# include "ac3_audio/AC3_audio.h"
# include "mpeg_audio/MPEG_audio.h"
#endif


#define MPEG_TYPE_UNKNOWN  0x00
#define MPEG_TYPE_ES_V1    0x02
#define MPEG_TYPE_ES_A1    0x03
#define MPEG_TYPE_ES_V2    0x04
#define MPEG_TYPE_ES_A2    0x05
#define MPEG_TYPE_TS       0x08
#define MPEG_TYPE_PS       0x10
#define MPEG_TYPE_ES_AC3   0x20
#define MPEG_TYPE_ES_WMV   0x40

/* Use this macro to initialize a MPEG2_PSI_PES_MODULE structure. */
#define DETECT_MODULE \
{\
   NULL, \
   DETECT_MODULE_process_pkt, \
   {\
      NULL, \
      NULL, \
      NULL, \
      NULL, \
      NULL, \
      NULL \
   }\
}

/* This is the data structure used by the DETECT_MODULE */
typedef struct {
   u32              state, bytes_found;
   u8               stream_type; /* See PMT_SECTION.h for STREAM_TYPE_* values. */
#ifndef _NO_AUDIO_
   AC3_SYNC_FRAME   ac3;
   MPEG_AUDIO_FRAME audio;
#endif
} DETECT_MODULE_data;

#ifdef __cplusplus
extern "C" {
#endif

/* Tries to determine the type of mpeg data in the data stream.
      INPUT: data_stream - Points to the data to look through.
             data_len    - Number of bytes in data_stream.
      RETURN: One of the MPEG_TYPE_* macros. */
int detect_mpeg(const u8 *data_stream, u32 data_len);

/* Find the first start code in the given data stream */
const u8 *detect_start_code(const u8 *data_stream,
                            u32       data_len,
                            u32      *bytes_found);

                            
/**************************************************************************
   COMMON PSI and PES MODULE FUNCTIONS
     These functions are used by SYSTEM modules to process the PSI_SECTIONs
       or PES_PKTs they produce.
 **************************************************************************/
/* Processes the packet handled by the module.
      INPUT: void * - Pointer to the memory provided for the module.
             void * - Pointer to the PES_PKT that needs handling.
                        Set pes->stream_id to zero to let the PES
                        module know to expect the start of a new
                        PES_PKT. */
void DETECT_MODULE_process_pkt(void *mem, void *pes);

#ifdef __cplusplus
};
#endif

#endif

