/***************************************************************************
                          PIC_HEADER.h  -  description
                             -------------------
   Stuff needed for handling MPEG2 picture and gop headers.
                             -------------------
    begin                : Tue Sep 28 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __PIC_HEADER_H__
#define __PIC_HEADER_H__

#include "common/typedefs.h"


#define GOP_HEADER_DROP_FRAME_FLAG(ptr_to_PIC_HEADER) \
   (ptr_to_PIC_HEADER->type & 0x04)
#define GOP_HEADER_CLOSED_GOP_FLAG(ptr_to_PIC_HEADER) \
   (ptr_to_PIC_HEADER->type & 0x02)
#define GOP_HEADER_BROKEN_LINK_FLAG(ptr_to_PIC_HEADER) \
   (ptr_to_PIC_HEADER->type & 0x01)

#define PIC_HEADER_CODING_TYPE(ptr_to_PIC_HEADER) \
   (ptr_to_PIC_HEADER->type & 0x70)
#define CODING_TYPE_I_FRAME   0x10
#define CODING_TYPE_P_FRAME   0x20
#define CODING_TYPE_B_FRAME   0x30
#define CODING_TYPE_FORBIDDEN(ptr_to_PIC_HEADER) \
   (!(ptr_to_PIC_HEADER->type & 0x70))
#define CODING_TYPE_RESERVED(ptr_to_PIC_HEADER) \
   ((ptr_to_PIC_HEADER->type & 0x70) > 0x40)

#define PIC_HEADER_TOP_FIELD_FIRST(ptr_to_PIC_HEADER) \
   (ptr_to_PIC_HEADER->pic_struct & 0x80)
#define PIC_HEADER_STRUCTURE(ptr_to_PIC_HEADER) \
   (ptr_to_PIC_HEADER->pic_struct & 0x03)
#define STRUCTURE_RESERVED    0x00
#define STRUCTURE_TOP_FIELD   0x01
#define STRUCTURE_BOT_FIELD   0x02
#define STRUCTURE_FULL_FRAME  0x03

#define PIC_HEADER_PROGRESSIVE_FRAME(ptr_to_PIC_HEADER) \
   (ptr_to_PIC_HEADER->flags & 0x80)
#define PIC_HEADER_FRAME_PRED_DCT(ptr_to_PIC_HEADER) \
   (ptr_to_PIC_HEADER->flags & 0x40)
#define PIC_HEADER_CONCEALMENT_VECTORS(ptr_to_PIC_HEADER) \
   (ptr_to_PIC_HEADER->flags & 0x20)
#define PIC_HEADER_Q_SCALE_TYPE(ptr_to_PIC_HEADER) \
   (ptr_to_PIC_HEADER->flags & 0x10)
#define PIC_HEADER_INTRA_VLC_FORMAT(ptr_to_PIC_HEADER) \
   (ptr_to_PIC_HEADER->flags & 0x08)
#define PIC_HEADER_ALT_SCAN(ptr_to_PIC_HEADER) \
   (ptr_to_PIC_HEADER->flags & 0x04)
#define PIC_HEADER_REPEAT_FIRST_FIELD(ptr_to_PIC_HEADER) \
   (ptr_to_PIC_HEADER->flags & 0x02)
#define PIC_HEADER_CHROMA_420_TYPE(ptr_to_PIC_HEADER) \
   (ptr_to_PIC_HEADER->flags & 0x01)

#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
   u8 type,                 /* <'000'> forbidden
                               <'001'> I_FRAME (intra-coded)
                               <'010'> P_FRAME (predictive-coded)
                               <'011'> B_FRAME (bidirectionally-predictive-coded)
                               <'100'> shall not be used
                         <'101'-'111'> reserved */
/* GOP_HEADER FLAGS
      1   drop_frame_flag              Set when the frame rate is 29.97fps.
                                         When not set the frame rate is
                                         rounded to the nearest integer.
      1   closed_gop                   Set when the first set of B-frames,
                                         after the I-frame following this
                                         GOP_HEADER, have been encoded using
                                         only backward prediction (the
                                         B-frames can be decoded, if you start
                                         decoding from the following I-frame).
                                         When not set, the decoder must skip
                                         the B-frames, if it starts decoding
                                         from the following I-frame.
      1   broken_link                  Set when the first set of B-frames,
                                         after the I-frame following this
                                         GOP_HEADER, cannot be decoded
                                         because the forward prediction frame
                                         was removed from the sequence. After
                                         initial encoding, this bit is 0.
                                         The value can change during the
                                         editing process.
*/
      intra_dc_precision,           /* 8, 9, 10, or 11 bits */
      pic_struct,
/*    
      1   top_field_first
 *
 * If SEQ_EXT.progressive_sequence is NOT set:
 *    If picture_structure is FULL_FRAME:
 *       top_field_first indicates which field, from this frame, is
 *       displayed first. Set for top first, and cleared for bottom first.
 *    Otherwise:
 *       top_field_first must be 0. Output fields in proper display order.
 * Otherwise:
 *    If repeat_first_field is NOT set:
 *       top_field_first must be 0. Ouput frames in proper display order.
 *    Otherwise:
 *       top_field_first indicates how many times to display this frame. When
 *       set, this frame is displayed 3 times. When not set, this frame is
 *       displayed 2 times.
 *
      2   picture_structure            Indicates the type of picture
                                         (field/frame).
                             <'00'> reserved
                             <'01'> TOP_FIELD
                             <'10'> BOTTOM_FIELD
                             <'11'> FULL_FRAME
 *
 * When the TOP_FIELD and BOTTOM_FIELD of a frame is encoded into a sequence,
 * the following must be true:
 *    If the first field is an I_FRAME, the second can be an I_FRAME or
 *      P_FRAME. Otherwise, both fields must have equal
 *      PIC_HEADER.picture_coding_type values.
 *    Both fields cannot have equal picture_structure values.
 *    Both fields can have different values for the following:
 *          f_code[0][0], f_code[0][1], f_code[1][0], f_code[1][1],
 *          intra_dc_precision, concealment_motion_vectors, q_scale_type,
 *          intra_vlc_format, alternate_scan
*/
       flags,
/*
      1   progressive_frame            Set when the 2 fields of the frame are
                                         from the same time instant. When not
                                         set, the 2 fields are from different
                                         time instants (true interlaced
                                         sequence).
 *
 * If progressive_frame is NOT set:
 *    repeat_first_field must be 0. Display frame 2 times.
 * Otherwise:
 *    picture_structure must be FULL_FRAME, and frame_pred_frame_dct must be 1.
 *
      1   frame_pred_frame_dct         Set when only frame DCT and prediction
                                         are used (FULL_FRAME). For ATSC
                                         sequences, this value must be set
                                         whenever progressive_frame is set.
      1   concealment_motion_vectors   Set when motion vectors are coded in
                                         the intra macroblocks.
      1   q_scale_type                 Affects the inverse quantisation
                                         process.
      1   intra_vlc_format             Affects the decoding of transform
                                         coefficient data.
      1   alternate_scan               Affects the decoding of transform
                                         coefficient data.
      1   repeat_first_field           Only used for FULL_FRAMES (see
                                         top_field_first). Field pictures must
                                         have a cleared bit here.
      1   chroma_420_type              If SEQ_EXT.chroma_format is "4:2:0",
                                         this value must be equal to
                                         progressive_frame. Otherwise it is 0.
*/
       f_code[2][2];
/*
      4   f_code[0][0]                 forward horizontal
      4   f_code[0][1]                 forward vertical
      4   f_code[1][0]                 backward horizontal
      4   f_code[1][1]                 backward vertical
 *
 *        f_code[s][t]          <0x00> forbidden
 *        (unsigned byte)  <0x01-0x09> Valid values for P_FRAMEs(forward) and B_FRAMEs.
 *                         <0x0A-0x0E> reserved
 *                              <0x0F> No motion vector. Used for I_FRAMEs and P_FRAMEs(backward).
 *                         <0x10-0xFF> forbidden (these are limited to 4-bit values!)
 */
   u16 temporal_reference,      /* 10: Indicates the order in which this
                                         picture will be displayed relative
                                         to the others in the sequence. For
                                         interlaced fields, the 2 fields that
                                         belong to each frame will have equal
                                         temporal_reference values. */
/*
 * if SEQ_EXT.low_delay is NOT set, or this picture does not follow a
 * "big picture":
 *    The temporal_reference value of each frame (or 2 fields) is incremented
 *    by 1 modulo 1024, in the display order. I-frames that follow a
 *    GOP_HEADER have a temporal_reference value of 0.
 * else if SEQ_EXT.low_delay is set, and this picture immediately follows a
 * "big picture":
 *    The temporal_reference value of this frame (or 2 fields) is equal to
 *    the temporal_reference value of the "big picture" plus N+1 modulo 1024.
 *    Where N is the number of times the VBV buffer is re-examined for the
 *    "big picture". If this I-frame was preceded by a GOP_HEADER, then the
 *    temporal_reference value is N.
 */
       vbv_delay;               /* 16: Delay between the time the last byte of
                                         the PIC_START_CODE is received, and
                                         the time this picture begins to be
                                         decoded. Measured in periods of the
                                         90kHz (27MHz/300) system derived
                                         clock. If one picture, in the
                                         sequence, has a vbv_delay value of
                                         0xFFFF, then all pictures must have
                                         the same value.
                            <45000> Max ATSC value (0xFFFF is still valid) */
} PIC_HEADER;


/* Reads the gop header in data_stream.
      INPUT: data_stream - Points to the data stream holding the gop
                             header. This stream does not have to
                             contain the entire header. Repeated calls to
                             this function can be used to read the header
                             in parts.
             data_len    - Number of bytes in the data stream.
      OUTPUT: hdr     - Points to the PIC_HEADER that will hold the data.
              cur_pos - Used by MPEG2_VIDEO_MODULE.
              code    - Used by MPEG2_VIDEO_MODULE.
      RETURN: Number of bytes read. */
u32 GOP_HEADER_read(const u8         *data_stream,
                          u32         data_len,
                          PIC_HEADER *hdr,
                          u32        *cur_pos,
                          u16        *code);

/* Reads the picture header in data_stream.
      INPUT: data_stream - Points to the data stream holding the picture
                             header. This stream does not have to
                             contain the entire header. Repeated calls to
                             this function can be used to read the header
                             in parts.
             data_len    - Number of bytes in the data stream.
      OUTPUT: hdr     - Points to the PIC_HEADER that will hold the data.
              cur_pos - Used by MPEG2_VIDEO_MODULE.
              code    - Used by MPEG2_VIDEO_MODULE.
      RETURN: Number of bytes read. */
u32 PIC_HEADER_read(const u8         *data_stream,
                          u32         data_len,
                          PIC_HEADER *hdr,
                          u32        *cur_pos,
                          u16        *code);

/* Reads the picture coding extension in data_stream.
      INPUT: data_stream - Points to the data stream holding the picture
                             coding extension. This stream does not have to
                             contain the entire extension. Repeated calls to
                             this function can be used to read the extension
                             in parts.
             data_len    - Number of bytes in the data stream.
      OUTPUT: hdr     - Points to the PIC_HEADER that will hold the data.
              cur_pos - Used by MPEG2_VIDEO_MODULE.
              code    - Used by MPEG2_VIDEO_MODULE.
      RETURN: Number of bytes read. */
u32 PIC_CODING_EXT_read(const u8         *data_stream,
                              u32         data_len,
                              PIC_HEADER *hdr,
                              u32        *cur_pos,
                              u16        *code);
#ifdef __cplusplus
};
#endif

#endif
