/***************************************************************************
                         mmx_synth.c  -  description
                             -------------------
   MMX version of the MPEG synthesis filter. Suitable for all processors
   (PMMX, P2, and later) that support the MMX instruction set.
                             -------------------
    begin                : Fri May 6 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#define WIN_CONST(v) ((v+2) >> 2)
static const s16 window[8*4*4*4] =
{
/**** 0 ->  3****/
WIN_CONST(     0), WIN_CONST(   213), WIN_CONST(    -1), WIN_CONST(   218),
WIN_CONST(    -1), WIN_CONST(   222), WIN_CONST(    -1), WIN_CONST(   225),
WIN_CONST(   -31), WIN_CONST(  -519), WIN_CONST(   -29), WIN_CONST(  -459),
WIN_CONST(   -38), WIN_CONST(  -645), WIN_CONST(   -35), WIN_CONST(  -581),
   
WIN_CONST(  2037), WIN_CONST(  6574), WIN_CONST(  2000), WIN_CONST(  5959),
WIN_CONST(  1952), WIN_CONST(  5288), WIN_CONST(  1893), WIN_CONST(  4561),
WIN_CONST( -5517), WIN_CONST(-39336), WIN_CONST( -5153), WIN_CONST(-37489),
WIN_CONST( -6237), WIN_CONST(-43006), WIN_CONST( -5879), WIN_CONST(-41176),
 
WIN_CONST( 75038), WIN_CONST(  6574), WIN_CONST( 74992), WIN_CONST(  7134),
WIN_CONST( 74856), WIN_CONST(  7640), WIN_CONST( 74630), WIN_CONST(  8092),
WIN_CONST( 35640), WIN_CONST(  4788), WIN_CONST( 37489), WIN_CONST(  5153),
WIN_CONST( 31947), WIN_CONST(  4063), WIN_CONST( 33791), WIN_CONST(  4425),
 
WIN_CONST(  2037), WIN_CONST(   213), WIN_CONST(  2063), WIN_CONST(   208),
WIN_CONST(  2080), WIN_CONST(   202), WIN_CONST(  2087), WIN_CONST(   196),
WIN_CONST(   401), WIN_CONST(    26), WIN_CONST(   459), WIN_CONST(    29),
WIN_CONST(   294), WIN_CONST(    21), WIN_CONST(   347), WIN_CONST(    24),

/**** 4 ->  7****/
WIN_CONST(    -1), WIN_CONST(   227), WIN_CONST(    -1), WIN_CONST(   228),
WIN_CONST(    -1), WIN_CONST(   228), WIN_CONST(    -2), WIN_CONST(   227),
WIN_CONST(   -45), WIN_CONST(  -779), WIN_CONST(   -41), WIN_CONST(  -711),
WIN_CONST(   -53), WIN_CONST(  -919), WIN_CONST(   -49), WIN_CONST(  -848),
     
WIN_CONST(  1822), WIN_CONST(  3776), WIN_CONST(  1739), WIN_CONST(  2935),
WIN_CONST(  1644), WIN_CONST(  2037), WIN_CONST(  1535), WIN_CONST(  1082),
WIN_CONST( -6935), WIN_CONST(-46617), WIN_CONST( -6589), WIN_CONST(-44821),
WIN_CONST( -7597), WIN_CONST(-50137), WIN_CONST( -7271), WIN_CONST(-48390),
      
WIN_CONST( 74313), WIN_CONST(  8492), WIN_CONST( 73908), WIN_CONST(  8840),
WIN_CONST( 73415), WIN_CONST(  9139), WIN_CONST( 72835), WIN_CONST(  9389),
WIN_CONST( 28289), WIN_CONST(  3351), WIN_CONST( 30112), WIN_CONST(  3705),
WIN_CONST( 24694), WIN_CONST(  2663), WIN_CONST( 26482), WIN_CONST(  3004),
     
WIN_CONST(  2085), WIN_CONST(   190), WIN_CONST(  2075), WIN_CONST(   183),
WIN_CONST(  2057), WIN_CONST(   176), WIN_CONST(  2032), WIN_CONST(   169),
WIN_CONST(   197), WIN_CONST(    17), WIN_CONST(   244), WIN_CONST(    19),
WIN_CONST(   111), WIN_CONST(    14), WIN_CONST(   153), WIN_CONST(    16),

/**** 8 -> 11****/
WIN_CONST(    -2), WIN_CONST(   224), WIN_CONST(    -2), WIN_CONST(   221),
WIN_CONST(    -2), WIN_CONST(   215), WIN_CONST(    -3), WIN_CONST(   208),
WIN_CONST(   -63), WIN_CONST( -1064), WIN_CONST(   -58), WIN_CONST(  -991),
WIN_CONST(   -73), WIN_CONST( -1210), WIN_CONST(   -68), WIN_CONST( -1137),
     
WIN_CONST(  1414), WIN_CONST(    70), WIN_CONST(  1280), WIN_CONST(  -998),
WIN_CONST(  1131), WIN_CONST( -2122), WIN_CONST(   970), WIN_CONST( -3300),
WIN_CONST( -8209), WIN_CONST(-53534), WIN_CONST( -7910), WIN_CONST(-51853),
WIN_CONST( -8755), WIN_CONST(-56778), WIN_CONST( -8491), WIN_CONST(-55178),
      
WIN_CONST( 72169), WIN_CONST(  9592), WIN_CONST( 71420), WIN_CONST(  9750),
WIN_CONST( 70590), WIN_CONST(  9863), WIN_CONST( 69679), WIN_CONST(  9935),
WIN_CONST( 21189), WIN_CONST(  2006), WIN_CONST( 22929), WIN_CONST(  2330),
WIN_CONST( 17799), WIN_CONST(  1388), WIN_CONST( 19478), WIN_CONST(  1692),
       
WIN_CONST(  2001), WIN_CONST(   161), WIN_CONST(  1962), WIN_CONST(   154),
WIN_CONST(  1919), WIN_CONST(   147), WIN_CONST(  1870), WIN_CONST(   139),
WIN_CONST(    36), WIN_CONST(    11), WIN_CONST(    72), WIN_CONST(    13),
WIN_CONST(   -29), WIN_CONST(     9), WIN_CONST(     2), WIN_CONST(    10),
    
/****12 -> 15****/
WIN_CONST(    -3), WIN_CONST(   200), WIN_CONST(    -4), WIN_CONST(   189),
WIN_CONST(    -4), WIN_CONST(   177), WIN_CONST(    -5), WIN_CONST(   163),
WIN_CONST(   -85), WIN_CONST( -1356), WIN_CONST(   -79), WIN_CONST( -1283),
WIN_CONST(   -97), WIN_CONST( -1498), WIN_CONST(   -91), WIN_CONST( -1428),
      
WIN_CONST(   794), WIN_CONST( -4533), WIN_CONST(   605), WIN_CONST( -5818),
WIN_CONST(   402), WIN_CONST( -7154), WIN_CONST(   185), WIN_CONST( -8540),
WIN_CONST( -9219), WIN_CONST(-59838), WIN_CONST( -8998), WIN_CONST(-58333),
WIN_CONST( -9585), WIN_CONST(-62684), WIN_CONST( -9416), WIN_CONST(-61289),
       
WIN_CONST( 68692), WIN_CONST(  9966), WIN_CONST( 67629), WIN_CONST(  9959),
WIN_CONST( 66494), WIN_CONST(  9916), WIN_CONST( 65290), WIN_CONST(  9838),
WIN_CONST( 14548), WIN_CONST(   814), WIN_CONST( 16155), WIN_CONST(  1095),
WIN_CONST( 11455), WIN_CONST(   288), WIN_CONST( 12980), WIN_CONST(   545),
       
WIN_CONST(  1817), WIN_CONST(   132), WIN_CONST(  1759), WIN_CONST(   125),
WIN_CONST(  1698), WIN_CONST(   117), WIN_CONST(  1634), WIN_CONST(   111),
WIN_CONST(   -83), WIN_CONST(     7), WIN_CONST(   -57), WIN_CONST(     8),
WIN_CONST(  -127), WIN_CONST(     6), WIN_CONST(  -106), WIN_CONST(     7),

/****16 -> 19****/
WIN_CONST(   104), WIN_CONST(  1567), WIN_CONST(   111), WIN_CONST(  1634),
WIN_CONST(   117), WIN_CONST(  1698), WIN_CONST(   125), WIN_CONST(  1759),
WIN_CONST(     6), WIN_CONST(  -127),                0,                 0,
WIN_CONST(     7), WIN_CONST(   -83), WIN_CONST(     7), WIN_CONST(  -106),

WIN_CONST(  9727), WIN_CONST( 64019), WIN_CONST(  9838), WIN_CONST( 65290),
WIN_CONST(  9916), WIN_CONST( 66494), WIN_CONST(  9959), WIN_CONST( 67629),
WIN_CONST(   288), WIN_CONST( 11455),                0,                 0,
WIN_CONST(   814), WIN_CONST( 14548), WIN_CONST(   545), WIN_CONST( 12980),
 
WIN_CONST( -9975), WIN_CONST(   -45), WIN_CONST( -8540), WIN_CONST(   185),
WIN_CONST( -7154), WIN_CONST(   402), WIN_CONST( -5818), WIN_CONST(   605),
WIN_CONST(-62684), WIN_CONST( -9585),                0,                 0,
WIN_CONST(-59838), WIN_CONST( -9219), WIN_CONST(-61289), WIN_CONST( -9416),
 
WIN_CONST(   146), WIN_CONST(    -5), WIN_CONST(   163), WIN_CONST(    -5),
WIN_CONST(   177), WIN_CONST(    -4), WIN_CONST(   189), WIN_CONST(    -4),
WIN_CONST( -1498), WIN_CONST(   -97),                0,                 0,
WIN_CONST( -1356), WIN_CONST(   -85), WIN_CONST( -1428), WIN_CONST(   -91),

/****20 -> 23****/
WIN_CONST(   132), WIN_CONST(  1817), WIN_CONST(   139), WIN_CONST(  1870),
WIN_CONST(   147), WIN_CONST(  1919), WIN_CONST(   154), WIN_CONST(  1962),
WIN_CONST(     9), WIN_CONST(   -29), WIN_CONST(     8), WIN_CONST(   -57),
WIN_CONST(    11), WIN_CONST(    36), WIN_CONST(    10), WIN_CONST(     2),

WIN_CONST(  9966), WIN_CONST( 68692), WIN_CONST(  9935), WIN_CONST( 69679),
WIN_CONST(  9863), WIN_CONST( 70590), WIN_CONST(  9750), WIN_CONST( 71420),
WIN_CONST(  1388), WIN_CONST( 17799), WIN_CONST(  1095), WIN_CONST( 16155),
WIN_CONST(  2006), WIN_CONST( 21189), WIN_CONST(  1692), WIN_CONST( 19478),

WIN_CONST( -4533), WIN_CONST(   794), WIN_CONST( -3300), WIN_CONST(   970),
WIN_CONST( -2122), WIN_CONST(  1131), WIN_CONST(  -998), WIN_CONST(  1280),
WIN_CONST(-56778), WIN_CONST( -8755), WIN_CONST(-58333), WIN_CONST( -8998),
WIN_CONST(-53534), WIN_CONST( -8209), WIN_CONST(-55178), WIN_CONST( -8491),

WIN_CONST(   200), WIN_CONST(    -3), WIN_CONST(   208), WIN_CONST(    -3),
WIN_CONST(   215), WIN_CONST(    -2), WIN_CONST(   221), WIN_CONST(    -2),
WIN_CONST( -1210), WIN_CONST(   -73), WIN_CONST( -1283), WIN_CONST(   -79),
WIN_CONST( -1064), WIN_CONST(   -63), WIN_CONST( -1137), WIN_CONST(   -68),

/****24 -> 27****/
WIN_CONST(   161), WIN_CONST(  2001), WIN_CONST(   169), WIN_CONST(  2032),
WIN_CONST(   176), WIN_CONST(  2057), WIN_CONST(   183), WIN_CONST(  2075),
WIN_CONST(    14), WIN_CONST(   111), WIN_CONST(    13), WIN_CONST(    72),
WIN_CONST(    17), WIN_CONST(   197), WIN_CONST(    16), WIN_CONST(   153),

WIN_CONST(  9592), WIN_CONST( 72169), WIN_CONST(  9389), WIN_CONST( 72835),
WIN_CONST(  9139), WIN_CONST( 73415), WIN_CONST(  8840), WIN_CONST( 73908),
WIN_CONST(  2663), WIN_CONST( 24694), WIN_CONST(  2330), WIN_CONST( 22929),
WIN_CONST(  3351), WIN_CONST( 28289), WIN_CONST(  3004), WIN_CONST( 26482),

WIN_CONST(    70), WIN_CONST(  1414), WIN_CONST(  1082), WIN_CONST(  1535),
WIN_CONST(  2037), WIN_CONST(  1644), WIN_CONST(  2935), WIN_CONST(  1739),
WIN_CONST(-50137), WIN_CONST( -7597), WIN_CONST(-51853), WIN_CONST( -7910),
WIN_CONST(-46617), WIN_CONST( -6935), WIN_CONST(-48390), WIN_CONST( -7271),

WIN_CONST(   224), WIN_CONST(    -2), WIN_CONST(   227), WIN_CONST(    -2),
WIN_CONST(   228), WIN_CONST(    -1), WIN_CONST(   228), WIN_CONST(    -1),
WIN_CONST(  -919), WIN_CONST(   -53), WIN_CONST(  -991), WIN_CONST(   -58),
WIN_CONST(  -779), WIN_CONST(   -45), WIN_CONST(  -848), WIN_CONST(   -49),

/****28 -> 31****/
WIN_CONST(   190), WIN_CONST(  2085), WIN_CONST(   196), WIN_CONST(  2087),
WIN_CONST(   202), WIN_CONST(  2080), WIN_CONST(   208), WIN_CONST(  2063),
WIN_CONST(    21), WIN_CONST(   294), WIN_CONST(    19), WIN_CONST(   244),
WIN_CONST(    26), WIN_CONST(   401), WIN_CONST(    24), WIN_CONST(   347),

WIN_CONST(  8492), WIN_CONST( 74313), WIN_CONST(  8092), WIN_CONST( 74630),
WIN_CONST(  7640), WIN_CONST( 74856), WIN_CONST(  7134), WIN_CONST( 74992),
WIN_CONST(  4063), WIN_CONST( 31947), WIN_CONST(  3705), WIN_CONST( 30112),
WIN_CONST(  4788), WIN_CONST( 35640), WIN_CONST(  4425), WIN_CONST( 33791),

WIN_CONST(  3776), WIN_CONST(  1822), WIN_CONST(  4561), WIN_CONST(  1893),
WIN_CONST(  5288), WIN_CONST(  1952), WIN_CONST(  5959), WIN_CONST(  2000),
WIN_CONST(-43006), WIN_CONST( -6237), WIN_CONST(-44821), WIN_CONST( -6589),
WIN_CONST(-39336), WIN_CONST( -5517), WIN_CONST(-41176), WIN_CONST( -5879),

WIN_CONST(   227), WIN_CONST(    -1), WIN_CONST(   225), WIN_CONST(    -1),
WIN_CONST(   222), WIN_CONST(    -1), WIN_CONST(   218), WIN_CONST(    -1),
WIN_CONST(  -645), WIN_CONST(   -38), WIN_CONST(  -711), WIN_CONST(   -41),
WIN_CONST(  -519), WIN_CONST(   -31), WIN_CONST(  -581), WIN_CONST(   -35)
};

/* using UNROLL_LOOPS should give a small increase in performance */
/*#define UNROLL_LOOPS*/

/* The mpeg synthesis filter.
      INPUT: sample - Points to the current group of DCT coefficients.
                        This function assumes there are 15 other groups
                        prior to the current one.
      OUTPUT: audio - The 32 audio samples generated by this filter. */
void synth_filter(const AV1 *sample, s16 *audio)
{
#if defined(__GCC__)

    /* op_code       src         dest */
#define SUM_FWD(reg, op) \
      "movq (%"    #reg"), "  "%%mm2\n\t" \
      "movq -128(%"#reg"), "  "%%mm4\n\t" \
      "movq "     "%%mm2, "   "%%mm3\n\t" \
      "movq "     "(%2), "    "%%mm6\n\t" \
      "punpcklwd ""%%mm4, "   "%%mm2\n\t" \
      "movq "     "8(%2), "   "%%mm7\n\t" \
      "punpckhwd ""%%mm4, "   "%%mm3\n\t" \
      "subl "     "$(128*2), %"#reg"\n\t" \
      "pmaddwd "  "%%mm6, "   "%%mm2\n\t" \
      "addl "     "$(8*2), "  "%2\n\t" \
      "pmaddwd "  "%%mm7, "   "%%mm3\n\t" \
      "p"#op"d "  "%%mm2, "   "%%mm0\n\t" \
      "p"#op"d "  "%%mm3, "   "%%mm1\n\t"

    /* op_code       src         dest */
#define SUM_BCK(reg, op) \
      "movq (%"    #reg"), "  "%%mm2\n\t" \
      "movq -128(%"#reg"), "  "%%mm4\n\t" \
      "movq "     "%%mm2, "   "%%mm3\n\t" \
      "movq "     "(%2), "    "%%mm6\n\t" \
      "punpckhwd ""%%mm4, "   "%%mm2\n\t" \
      "movq "     "8(%2), "   "%%mm7\n\t" \
      "punpcklwd ""%%mm4, "   "%%mm3\n\t" \
      "subl "     "$(128*2), %"#reg"\n\t" \
      "pmaddwd "  "%%mm6, "   "%%mm2\n\t" \
      "addl "     "$(8*2), "  "%2\n\t" \
      "pmaddwd "  "%%mm7, "   "%%mm3\n\t" \
      "movq "     "%%mm2, "   "%%mm6\n\t" \
      "psrlq "    "$32, "     "%%mm2\n\t" \
      "movq "     "%%mm3, "   "%%mm7\n\t" \
      "psrlq "    "$32, "     "%%mm3\n\t" \
      "p"#op"d "  "%%mm2, "   "%%mm0\n\t" \
      "psllq "    "$32, "     "%%mm6\n\t" \
      "p"#op"d "  "%%mm3, "   "%%mm1\n\t" \
      "psllq "    "$32, "     "%%mm7\n\t" \
      "p"#op"d "  "%%mm6, "   "%%mm0\n\t" \
      "p"#op"d "  "%%mm7, "   "%%mm1\n\t"

    /* op_code       src         dest */
#define AUDIO_SAMPLE(inc0, inc1) \
      "paddd "    "%%mm5, "   "%%mm0\n\t" \
      "paddd "    "%%mm5, "   "%%mm1\n\t" \
      "psrad "    "%6, "      "%%mm0\n\t" \
      "addl "     "$(4*2), "  "%3\n\t" \
      "psrad "    "%6, "      "%%mm1\n\t" \
      "addl $"    #inc0", "   "%0\n\t" \
      "packssdw " "%%mm1, "   "%%mm0\n\t" \
      "addl $"    #inc1", "   "%1\n\t" \
      "movq "     "%%mm0, "   "(%3)\n\t"

   __asm__ __volatile__(
    /* registers used: %0   = sample ptr 1
                       %1   = sample ptr 2
                       %2   = window constants
                       %3   = audio sample buffer
                       %4   = outer loop counter
                       %5   = inner loop counter
                      %%mm0 = audio samples
                      %%mm1 = audio samples
                      %%mm5 = error correction bit */
    /* op_code     src         dest */
      "movd "     "%5, "      "%%mm5\n\t"
      "punpckldq ""%%mm5, "   "%%mm5\n\t"
      "SYNTH_LOOP_OUT_1:\n\t"

      "pxor "     "%%mm0, "   "%%mm0\n\t"
      "pxor "     "%%mm1, "   "%%mm1\n\t"
#ifndef UNROLL_LOOPS
      "movl "     "$4, "      "%5\n\t"
      "SYNTH_LOOP_IN_1:\n\t"
#else
      SUM_FWD(0, add)
      SUM_BCK(1, sub)
      SUM_FWD(0, add)
      SUM_BCK(1, sub)
      SUM_FWD(0, add)
      SUM_BCK(1, sub)
#endif
      SUM_FWD(0, add)
      SUM_BCK(1, sub)
#ifndef UNROLL_LOOPS
      "dec "                  "%5\n\t"
      "jnz "      "SYNTH_LOOP_IN_1\n\t"
#endif

      AUDIO_SAMPLE((516*2), (508*2))
      "dec "                  "%4\n\t"
      "jnz "      "SYNTH_LOOP_OUT_1\n\t"


      "addl "     "$(3*2), "  "%1\n\t"    /* %1 += 3 */
      "subl "     "$(3*2), "  "%0\n\t"    /* %0 -= 3 */
      "movl "     "$4, "      "%4\n\t"
      "SYNTH_LOOP_OUT_2:\n\t"

      "pxor "     "%%mm0, "   "%%mm0\n\t"
      "pxor "     "%%mm1, "   "%%mm1\n\t"
#ifndef UNROLL_LOOPS
      "movl "     "$4, "      "%5\n\t"
      "SYNTH_LOOP_IN_2:\n\t"
#else
      SUM_FWD(1, add)
      SUM_BCK(0, add)
      SUM_FWD(1, add)
      SUM_BCK(0, add)
      SUM_FWD(1, add)
      SUM_BCK(0, add)
#endif
      SUM_FWD(1, add)
      SUM_BCK(0, add)
#ifndef UNROLL_LOOPS
      "dec "                  "%5\n\t"
      "jnz "      "SYNTH_LOOP_IN_2\n\t"
#endif

      AUDIO_SAMPLE((508*2), (516*2))
      "dec "                  "%4\n\t"
      "jnz "      "SYNTH_LOOP_OUT_2\n\t"

   /* output list */
   :

   /* input list */
   : "q" (sample+16),
     "q" (sample-19),
     "q" (window),
     "q" (audio-4),
     "r" (4),
     "r" ((int)1 << (FP_SHIFT-2)),
     "I" (FP_SHIFT-1)
   );

#elif defined(__MSVC__)
        /* op_code   dest     src */
#define SUM_FWD(reg, op) \
     __asm movq      mm2,     [reg] \
     __asm movq      mm4,     [reg-128] \
     __asm movq      mm3,     mm2 \
     __asm movq      mm6,     [edx] \
     __asm punpcklwd mm2,     mm4 \
     __asm movq      mm7,     [edx+8] \
     __asm punpckhwd mm3,     mm4 \
     __asm sub       reg,     (128*2) \
     __asm pmaddwd   mm2,     mm6 \
     __asm add       edx,     (8*2) \
     __asm pmaddwd   mm3,     mm7 \
     __asm p##op##d  mm0,     mm2 \
     __asm p##op##d  mm1,     mm3

        /* op_code   dest     src */
#define SUM_BCK(reg, op) \
     __asm movq      mm2,     [reg] \
     __asm movq      mm4,     [reg-128] \
     __asm movq      mm3,     mm2 \
     __asm movq      mm6,     [edx] \
     __asm punpckhwd mm2,     mm4 \
     __asm movq      mm7,     [edx+8] \
     __asm punpcklwd mm3,     mm4 \
     __asm sub       reg,     (128*2) \
     __asm pmaddwd   mm2,     mm6 \
     __asm add       edx,     (8*2) \
     __asm pmaddwd   mm3,     mm7 \
     __asm movq      mm6,     mm2 \
     __asm psrlq     mm2,     32 \
     __asm movq      mm7,     mm3 \
     __asm psrlq     mm3,     32 \
     __asm p##op##d  mm0,     mm2 \
     __asm psllq     mm6,     32 \
     __asm p##op##d  mm1,     mm3 \
     __asm psllq     mm7,     32 \
     __asm p##op##d  mm0,     mm6 \
     __asm p##op##d  mm1,     mm7

        /* op_code   dest     src */
#define AUDIO_SAMPLE(inc0, inc1) \
     __asm paddd     mm0,     mm5 \
     __asm paddd     mm1,     mm5 \
     __asm psrad     mm0,     (FP_SHIFT-1) \
     __asm add       ecx,     (4*2) \
     __asm psrad     mm1,     (FP_SHIFT-1) \
     __asm add       edi,     inc0 \
     __asm packssdw  mm0,     mm1 \
     __asm add       esi,     inc1 \
     __asm movq      [ecx],   mm0

    /* registers used: edi = sample ptr 1
                       esi = sample ptr 2
                       edx = window constants
                       ecx = audio sample buffer
                       ebx = outer loop counter
                       eax = inner loop counter
                       mm0 = audio samples
                       mm1 = audio samples
                       mm5 = error correction bit */
    /* op_code    dest     src */
   __asm {
      mov         edi,     dword ptr[sample]
      mov         eax,     (1 << (FP_SHIFT-2))
      mov         esi,     edi
      mov         edx,     offset window
      add         edi,     (16*2)
      mov         ecx,     dword ptr[audio]
      sub         esi,     (19*2)
      mov         ebx,     4
      sub         ecx,     (4*2)

      movd        mm5,     eax
      punpckldq   mm5,     mm5
   SYNTH_LOOP_OUT_1:
      pxor        mm0,     mm0
      pxor        mm1,     mm1
#ifndef UNROLL_LOOPS
      mov         eax,     4
   SYNTH_LOOP_IN_1:
#else
      SUM_FWD(edi, add)
      SUM_BCK(esi, sub)
      SUM_FWD(edi, add)
      SUM_BCK(esi, sub)
      SUM_FWD(edi, add)
      SUM_BCK(esi, sub)
#endif
      SUM_FWD(edi, add)
      SUM_BCK(esi, sub)
#ifndef UNROLL_LOOPS
      dec         eax
      jnz         SYNTH_LOOP_IN_1
#endif

      AUDIO_SAMPLE((516*2), (508*2))
      dec         ebx
      jnz         SYNTH_LOOP_OUT_1


      add         esi,     (3*2)    /* %1 += 3 */
      sub         edi,     (3*2)    /* %0 -= 3 */
      mov         ebx,     4
   SYNTH_LOOP_OUT_2:

      pxor        mm0,     mm0
      pxor        mm1,     mm1
#ifndef UNROLL_LOOPS
      mov         eax,     4
   SYNTH_LOOP_IN_2:
#else
      SUM_FWD(esi, add)
      SUM_BCK(edi, add)
      SUM_FWD(esi, add)
      SUM_BCK(edi, add)
      SUM_FWD(esi, add)
      SUM_BCK(edi, add)
#endif
      SUM_FWD(esi, add)
      SUM_BCK(edi, add)
#ifndef UNROLL_LOOPS
      dec         eax
      jnz         SYNTH_LOOP_IN_2
#endif

      AUDIO_SAMPLE((508*2), (516*2))
      dec         ebx
      jnz         SYNTH_LOOP_OUT_2
   }

#else
# error "Don't know how to insert inline ASM code for this compiler"
#endif
}

#undef SUM_FWD
#undef SUM_BCK
#undef AUDIO_SAMPLE
#undef WIN_CONST
#undef UNROLL_LOOPS 

