/***************************************************************************
                       P_MACROBLOCK.c  -  description
                             -------------------
   Stuff needed for handling P picture macroblocks.
                             -------------------
    begin                : Mon Oct 30 2006
    copyright            : (C) 2006 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/*#define _BITSTREAM_TRACE_
#define BIT_TRACE_BEG_BLK  VC1_TRACE_BEG_BLK
#define BIT_TRACE_END_BLK  VC1_TRACE_END_BLK*/

/*#define VC1_VIDEO_TRACE 2*/
#include <string.h>
#include <stdlib.h>

#include "P_MACROBLOCK.h"
#include "vlc_tables.h"
#include "tables.h"
#include "inv_trans.h"
#include "bitstream.h"
#include "VC1_motion_vectors.h"
#include "mpeg2_video/motion_vectors.h"


#if defined(VC1_VIDEO_TRACE)||defined(OUTPUT_BLOCK_TRACE)
# include <stdio.h>
#endif

#define DATA_PTR        ptr
#define END_DATA_PTR    end_pos
#define PARSER_STATE    seq->state
#define DATA_BYTE       data
#define BIT_SHIFT_BYTE  bit_shift


/* Read the P picture macroblock data.
      INPUT: seq - Pointer to a SEQ_LAYER structure.
             pic - Pointer to a PIC_LAYER structure. */
void VC1_VIDEO_read_P_MACROBLOCK(SEQ_LAYER *seq, PIC_LAYER *pic)
{
   const u8 *ptr       = seq->payload,
            *end_pos   = seq->payload + seq->payload_len,
            *ACdelta;
   s16      *b, *c, *a, *d;
   u8        bit_shift = seq->bit_shift,
             data      = seq->data,
             bit_shift_inc,
             tmp;

#undef  STOP_PARSING
#define STOP_PARSING    goto VC1_VIDEO_read_P_MACROBLOCK_STOP

#define NEXT            (VC1_VIDEO_read_P_MACROBLOCK_STATE+ 1)
#define START           (VC1_VIDEO_read_P_MACROBLOCK_STATE+ 2)
#define MV4_CBPCY       (VC1_VIDEO_read_P_MACROBLOCK_STATE+ 3)
#define MV4_BLK_MVDATA  (VC1_VIDEO_read_P_MACROBLOCK_STATE+ 4)
#define MV4_INDEX       (VC1_VIDEO_read_P_MACROBLOCK_STATE+ 5)
#define MV4_DMV0        (VC1_VIDEO_read_P_MACROBLOCK_STATE+ 6)
#define MV4_DMV1        (VC1_VIDEO_read_P_MACROBLOCK_STATE+ 7)
#define MV4_DMV2        (VC1_VIDEO_read_P_MACROBLOCK_STATE+ 8)
#define MV4_DMV3        (VC1_VIDEO_read_P_MACROBLOCK_STATE+ 9)
#define MV1_INDEX       (VC1_VIDEO_read_P_MACROBLOCK_STATE+10)
#define MV1_DMV0        (VC1_VIDEO_read_P_MACROBLOCK_STATE+11)
#define MV1_DMV1        (VC1_VIDEO_read_P_MACROBLOCK_STATE+12)
#define MV1_DMV2        (VC1_VIDEO_read_P_MACROBLOCK_STATE+13)
#define MV1_DMV3        (VC1_VIDEO_read_P_MACROBLOCK_STATE+14)
#define MV1_ACPRED      (VC1_VIDEO_read_P_MACROBLOCK_STATE+15)
#define MV1_CBPCY       (VC1_VIDEO_read_P_MACROBLOCK_STATE+16)
#define MV1_MQUANT0     (VC1_VIDEO_read_P_MACROBLOCK_STATE+17)
#define TT_MB0          (VC1_VIDEO_read_P_MACROBLOCK_STATE+18)
#define TT_MB1          (VC1_VIDEO_read_P_MACROBLOCK_STATE+19)
#define START_BLOCKS    (VC1_VIDEO_read_P_MACROBLOCK_STATE+20)
#define DC_DIFF1        (VC1_VIDEO_read_P_MACROBLOCK_STATE+21)
#define DC_DIFF2        (VC1_VIDEO_read_P_MACROBLOCK_STATE+22)
#define DC_DIFF_ESC     (VC1_VIDEO_read_P_MACROBLOCK_STATE+23)
#define DC_PRED         (VC1_VIDEO_read_P_MACROBLOCK_STATE+24)
#define TT_BLK0         (VC1_VIDEO_read_P_MACROBLOCK_STATE+25)
#define TT_BLK1         (VC1_VIDEO_read_P_MACROBLOCK_STATE+26)
#define TT_BLK2         (VC1_VIDEO_read_P_MACROBLOCK_STATE+27)
#define TT_BLK3         (VC1_VIDEO_read_P_MACROBLOCK_STATE+28)
#define AC_INDEX        (VC1_VIDEO_read_P_MACROBLOCK_STATE+29)
#define AC_ESC_MODE     (VC1_VIDEO_read_P_MACROBLOCK_STATE+30)
#define AC_ESC_MODE3_1  (VC1_VIDEO_read_P_MACROBLOCK_STATE+31)
#define AC_ESC_MODE3_2  (VC1_VIDEO_read_P_MACROBLOCK_STATE+32)
#define AC_ESC_MODE3_3  (VC1_VIDEO_read_P_MACROBLOCK_STATE+33)
#define AC_ESC_MODE3_4  (VC1_VIDEO_read_P_MACROBLOCK_STATE+34)
#define AC_ESC_MODE3_5  (VC1_VIDEO_read_P_MACROBLOCK_STATE+35)
#define AC_ESC_MODE3_6  (VC1_VIDEO_read_P_MACROBLOCK_STATE+36)
#define AC_INDEX2       (VC1_VIDEO_read_P_MACROBLOCK_STATE+37)
#define AC_SIGN         (VC1_VIDEO_read_P_MACROBLOCK_STATE+38)
#define AC_NEXT         (VC1_VIDEO_read_P_MACROBLOCK_STATE+39)
   if (PARSER_STATE > VC1_VIDEO_read_P_MACROBLOCK_STATE) {
      switch (PARSER_STATE) {
      case NEXT:
         goto VC1_VIDEO_read_P_MACROBLOCK_NEXT;
      case START:
         goto VC1_VIDEO_read_P_MACROBLOCK_START;
      case MV4_CBPCY:
         goto VC1_VIDEO_read_P_MACROBLOCK_MV4_CBPCY;
      case MV4_BLK_MVDATA:
         goto VC1_VIDEO_read_P_MACROBLOCK_MV4_BLK_MVDATA;
      case MV4_INDEX:
         goto VC1_VIDEO_read_P_MACROBLOCK_MV4_INDEX;
      case MV4_DMV0:
         goto VC1_VIDEO_read_P_MACROBLOCK_MV4_DMV0;
      case MV4_DMV1:
         a = pic->coeff;
         if (pic->b < 0x0010)
            a++;
         goto VC1_VIDEO_read_P_MACROBLOCK_MV4_DMV1;
      case MV4_DMV2:
         a = pic->coeff;
         if (pic->b < 0x0010)
            a++;
         goto VC1_VIDEO_read_P_MACROBLOCK_MV4_DMV2;
      case MV4_DMV3:
         d = pic->MV + (pic->x >> 1) + pic->c;
         goto VC1_VIDEO_read_P_MACROBLOCK_MV4_DMV3;
      case MV1_INDEX:
         goto VC1_VIDEO_read_P_MACROBLOCK_MV1_INDEX;
      case MV1_DMV0:
         goto VC1_VIDEO_read_P_MACROBLOCK_MV1_DMV0;
      case MV1_DMV1:
         a = pic->coeff;
         if (pic->b < 0x0010)
            a++;
         goto VC1_VIDEO_read_P_MACROBLOCK_MV1_DMV1;
      case MV1_DMV2:
         a = pic->coeff;
         if (pic->b < 0x0010)
            a++;
         goto VC1_VIDEO_read_P_MACROBLOCK_MV1_DMV2;
      case MV1_DMV3:
         d = pic->MV + (pic->x >> 1);
         goto VC1_VIDEO_read_P_MACROBLOCK_MV1_DMV3;
      case MV1_ACPRED:
         goto VC1_VIDEO_read_P_MACROBLOCK_MV1_ACPRED;
      case MV1_CBPCY:
         goto VC1_VIDEO_read_P_MACROBLOCK_MV1_CBPCY;
      case MV1_MQUANT0:
         goto VC1_VIDEO_read_P_MACROBLOCK_MV1_MQUANT0;
      case TT_MB0:
         goto VC1_VIDEO_read_P_MACROBLOCK_TT_MB0;
      case TT_MB1:
         goto VC1_VIDEO_read_P_MACROBLOCK_TT_MB1;
      case START_BLOCKS:
         goto VC1_VIDEO_read_P_MACROBLOCK_START_BLOCKS;
      case DC_DIFF1:
         goto VC1_VIDEO_read_P_MACROBLOCK_DC_DIFF1;
      case DC_DIFF2:
         goto VC1_VIDEO_read_P_MACROBLOCK_DC_DIFF2;
      case DC_DIFF_ESC:
         goto VC1_VIDEO_read_P_MACROBLOCK_DC_DIFF_ESC;
      case DC_PRED:
         goto VC1_VIDEO_read_P_MACROBLOCK_DC_PRED;
      case TT_BLK0:
         goto VC1_VIDEO_read_P_MACROBLOCK_TT_BLK0;
      case TT_BLK1:
         goto VC1_VIDEO_read_P_MACROBLOCK_TT_BLK1;
      case TT_BLK2:
         goto VC1_VIDEO_read_P_MACROBLOCK_TT_BLK2;
      case TT_BLK3:
         goto VC1_VIDEO_read_P_MACROBLOCK_TT_BLK3;
      case AC_INDEX:
         goto VC1_VIDEO_read_P_MACROBLOCK_AC_INDEX;
      case AC_ESC_MODE:
         goto VC1_VIDEO_read_P_MACROBLOCK_AC_ESC_MODE;
      case AC_ESC_MODE3_1:
         goto VC1_VIDEO_read_P_MACROBLOCK_AC_ESC_MODE3_1;
      case AC_ESC_MODE3_2:
         goto VC1_VIDEO_read_P_MACROBLOCK_AC_ESC_MODE3_2;
      case AC_ESC_MODE3_3:
         goto VC1_VIDEO_read_P_MACROBLOCK_AC_ESC_MODE3_3;
      case AC_ESC_MODE3_4:
         goto VC1_VIDEO_read_P_MACROBLOCK_AC_ESC_MODE3_4;
      case AC_ESC_MODE3_5:
         goto VC1_VIDEO_read_P_MACROBLOCK_AC_ESC_MODE3_5;
      case AC_ESC_MODE3_6:
         goto VC1_VIDEO_read_P_MACROBLOCK_AC_ESC_MODE3_6;
      case AC_INDEX2:
         goto VC1_VIDEO_read_P_MACROBLOCK_AC_INDEX2;
      case AC_SIGN:
         tmp = *(pic->tab+2);
         goto VC1_VIDEO_read_P_MACROBLOCK_AC_SIGN;
      case AC_NEXT:
         tmp = *(pic->tab+2);
         goto VC1_VIDEO_read_P_MACROBLOCK_AC_NEXT;
      }
   }

   /* initialize everything before we start decoding the first MB */
   pic->BPi     = pic->x   = pic-> y = 0;
   pic->overlap = pic->CBP = 0;
   pic->rnd    ^= 1;
   pic->BPbit   = 0x80;

   pic->Mquant = pic->Pquant;
   if (pic->Pquant <= 5)
      pic->DCstep = pic->Pquant & 0xFE;
   else if (pic->Pquant <= 9)
      pic->DCstep = 8;
   else
      pic->DCstep = (pic->Pquant >> 2) + 6;
   pic->DCstep_org = pic->DCstep;

   /* NOTE: CBP_l  = used to store the TT VLC table number.
            CBP_lt = used to store the sub-block pattern */
   if (pic->Pquant >= 26)
      pic->CBP_l = 0;
   else if (pic->Pquant >= 10)
      pic->CBP_l = 1;
   else
      pic->CBP_l = 2;

   memset(pic->ACpred_Y, 0x7F, (sizeof(s16)*((7*(VC1_MAX_FRAME_WID/8)) + 14)));
   memset(pic->ACpred_C, 0x7F, (sizeof(s16)*((7*(VC1_MAX_FRAME_WID/8)) + 14)));

   if (pic->MVmode & VC1_MVMODE_INT_COMP)
      VC1_MV_BUILD_LUT_TABLE(pic->LUT, (s32)pic->Yscale, (s32)pic->Yshift);

VC1_VIDEO_read_P_MACROBLOCK_NEXT:
#if defined(VC1_VIDEO_TRACE)||defined(OUTPUT_BLOCK_TRACE)
   VC1_TRACE_BEG_BLK
   printf("  <P_MACROBLOCK %hu,%hu>"
          "  (BPidx=%hu, BPbit=0x%02hX)\n",
          (pic->x >> 4), (pic->y >> 4),
          pic->BPi, (u16)pic->BPbit);
   VC1_TRACE_END_BLK
#endif
   GET_LSB;
   if ( (pic->MVmode & VC1_MVMODE_MIXED_MV)&&
        (pic->BPraw & VC1_BP_RAW_MVTYPE) ) {
      if (DATA_BYTE >= 0x80)
         pic->MVtype[pic->BPi] |=   pic->BPbit;
      else
         pic->MVtype[pic->BPi] &= ~(pic->BPbit);
      BIT_SHIFT_BYTE++;
      DATA_BYTE <<= 1;
   }

   if (pic->BPraw & VC1_BP_RAW_MBSKIP) {
      if (DATA_BYTE >= 0x80)
         pic->MBskip[pic->BPi] |=   pic->BPbit;
      else
         pic->MBskip[pic->BPi] &= ~(pic->BPbit);
      BIT_SHIFT_BYTE++;
      DATA_BYTE <<= 1;
   }

#ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   printf("    MVtype = %cMV\n"
          "    MBskip = %c\n",
          ((pic->MVtype[pic->BPi] & pic->BPbit) ? '4' : '1'),
          ((pic->MBskip[pic->BPi] & pic->BPbit) ? 'y' : 'n'));
   VC1_TRACE_END_BLK
#endif
   ALIGN0_DATA_BYTE(START)

VC1_VIDEO_read_P_MACROBLOCK_START:
   pic->coeff[0] =
   pic->coeff[1] = 0;
   pic->MVmode  &= ~(VC1_MVMODE_INTRA | VC1_MVMODE_CODED);
   pic->MV4mode  = 0;
   pic->CBP_lt   = 0;
   if (pic->MBskip[pic->BPi] & pic->BPbit) {
/**** START Skipped Macroblock ****/
      if (pic->MVtype[pic->BPi] & pic->BPbit) {
/******* 4MV skip *******/
         pic->MVtype[pic->BPi] &= ~(pic->BPbit);
         pic->CBP_t = 0;
         pic->c     = 0;
         goto VC1_VIDEO_read_P_MACROBLOCK_MV4_PREDICTION;
      }
      else
         goto VC1_VIDEO_read_P_MACROBLOCK_MV1_PREDICTION;
/**** END   Skipped Macroblock ****/
   }
   else if (pic->MVtype[pic->BPi] & pic->BPbit) {
/**** START 4MV Macroblock ****/
      pic->MVtype[pic->BPi] &= ~(pic->BPbit);
      pic->tab = VC1_PB_CBP_TAB[VC1_PIC_CBP_TABLEp(pic)];
VC1_VIDEO_read_P_MACROBLOCK_MV4_CBPCY:
      VC1_PARSE_VLC(pic->tab, MV4_CBPCY)
      bit_shift_inc = *(pic->tab+1);
      pic->CBP      = *(pic->tab+2);
#ifdef VC1_VIDEO_TRACE
      VC1_TRACE_BEG_BLK
      printf("    CBP = 0x%02hX\n", (u16)(pic->CBP >> 2));
      VC1_TRACE_END_BLK
#endif
      ALIGN_DATA_BYTE(MV4_BLK_MVDATA, bit_shift_inc);

VC1_VIDEO_read_P_MACROBLOCK_MV4_BLK_MVDATA:
      pic->CBP_t = pic->CBP & 0xF0;
      pic->c     = 0;
      do {
         if (pic->CBP_t < 0x80) {
#ifdef VC1_VIDEO_TRACE
            VC1_TRACE_BEG_BLK
            printf("    MV4index[%hu] = NONE\n", (pic->c >> 1));
            VC1_TRACE_END_BLK
#endif
            goto VC1_VIDEO_read_P_MACROBLOCK_MV4_PREDICTION;
         }

         pic->tab = VC1_MV_DIFF_TAB[VC1_PIC_MV_TABLEp(pic)];
VC1_VIDEO_read_P_MACROBLOCK_MV4_INDEX:
         VC1_PARSE_VLC(pic->tab, MV4_INDEX)
         bit_shift_inc = *(pic->tab+1);
         ALIGN_DATA_BYTE(MV4_DMV0, bit_shift_inc);

VC1_VIDEO_read_P_MACROBLOCK_MV4_DMV0:
         tmp = *(pic->tab+2);
#ifdef VC1_VIDEO_TRACE
         VC1_TRACE_BEG_BLK
         printf("    MV4index[%hu] = %hhu\n", (pic->c >> 1), tmp);
         VC1_TRACE_END_BLK
#endif
         if (tmp >= 37) {
            tmp -= 37;
            pic->CBP_t |= 0x08;  /* VC1_MVMODE_CODED */
            if (tmp != 36)
               pic->CBP_lt = pic->VST;
         }

         if (tmp == 35) {
            pic->b = (u16)pic->MVsize + 0x0011;
            a      = pic->coeff;
VC1_VIDEO_read_P_MACROBLOCK_MV4_DMV1:
            while (pic->b) {
               tmp = (u8)pic->b & 0x0F;
               if (tmp > 8) {
                  tmp    -= 8;
                  pic->b -= 8;
                  GET_LSB;
                  *a = (s16)DATA_BYTE << tmp;
                  ALIGN8_DATA_BYTE(MV4_DMV1)
               }
               pic->b >>= 4;
               GET_LSB;
               *a |= (s16)(DATA_BYTE >> (8 - tmp));
               ALIGN_DATA_BYTE(MV4_DMV1, tmp);
               a++;
            }
         }
         else if (tmp == 36)
            pic->MV4mode |= 0x08;   /* VC1_MVMODE_INTRA */
         else if (tmp) {
            pic->b = (((u16)tmp % 6) | (((u16)tmp / 6) << 4)) + 0x0011;
            a      = pic->coeff;
VC1_VIDEO_read_P_MACROBLOCK_MV4_DMV2:
            while (pic->b) {
               tmp      = (u8)pic->b & 0x0F;
               pic->b >>= 4;
               if (tmp > 1) {
                  bit_shift_inc =      VC1_MV_SIZE_TAB[tmp-2];
                  *a            = (s16)VC1_MV_OFFSET_TAB[tmp-2];
                  GET_LSB;
                  *a += (s16)(DATA_BYTE >> (8 - bit_shift_inc));
                  DATA_BYTE     <<= bit_shift_inc;
                  BIT_SHIFT_BYTE += bit_shift_inc + 1;
                  if (DATA_BYTE >= 0x80)
                     *a = -(*a);
                  DATA_BYTE     <<= 1;
                  ALIGN0_DATA_BYTE(MV4_DMV2)
               }
               a++;
            }
         }

VC1_VIDEO_read_P_MACROBLOCK_MV4_PREDICTION:
         /* MV Prediction */
         d = pic->MV + (pic->x >> 1) + pic->c;
         if ( ( (pic->x == 0)&&(pic->y == 0)&&(pic->c == 0) )||
              (pic->MV4mode & 0x08) ) {
            if (pic->c == 6) {
               d[2] = d[0];   /* store B predictor for next block 0 */
               d[3] = d[1];
            }
            /* A and C are out of bounds */
            d[0] =
            d[1] = 0;
         }
         else if ( (pic->y == 0)&&(pic->c < 4) ) {
            /* A is out of bounds */
            if (pic->c) {  /* block 1 */
               d[0] = d[-2];
               d[1] = d[-1];
            }
            else {         /* block 0 */
               d[0] = d[-6];
               d[1] = d[-5];
            }
            if ( (d[0])||(d[1]) ) {
               a    = NULL;
               goto VC1_VIDEO_read_P_MACROBLOCK_MV4_PULLBACK;
            }
         }
         else if ( (pic->x == 0)&&(!(pic->c & 2)) ) {
            /* C is out of bounds */
            if (pic->c) {  /* block 2 */
               a = d - 4;
               b = d - 2;
            }
            else {         /* block 0 */
               a = d + 4;
               b = d + 6;
            }
            d[0] = VC1_median3(a[0], b[0], 0);
            d[1] = VC1_median3(a[1], b[1], 0);
#if (VC1_VIDEO_TRACE > 2)
            VC1_TRACE_BEG_BLK
            printf("        A=(%hd,%hd) B=(%hd,%hd)\n",
                   a[0], a[1], b[0], b[1]);
            VC1_TRACE_END_BLK
#endif
            if ( (d[0])||(d[1]) ) {
               a    = NULL;
               goto VC1_VIDEO_read_P_MACROBLOCK_MV4_PULLBACK;
            }
         }
         else {
            /* A, B and C are in bounds */
            a = b = c = d;
            if (pic->c == 0) {      /* block 0 */
               a += 4;
               c -= 6;
            }
            else if (pic->c == 2) { /* block 1 */
               a += 4;
               b += (seq->wid <= (pic->x+16)) ? 2 : 10;
               c -= 2;
            }
            else if (pic->c == 4) { /* block 2 */
               a -= 4;
               b -= 2;
               c -= 6;
            }
            else {                  /* block 3 */
               a -= 4;
               b -= 6;
               c -= 2;
               d[2] = d[0];   /* store B predictor for next block 0 */
               d[3] = d[1];
            }
            d[0] = VC1_median3(a[0], b[0], c[0]);
            d[1] = VC1_median3(a[1], b[1], c[1]);
#if (VC1_VIDEO_TRACE > 2)
            VC1_TRACE_BEG_BLK
            printf("        A=(%hd,%hd) B=(%hd,%hd) C=(%hd,%hd)\n",
                   a[0], a[1], b[0], b[1], c[0], c[1]);
            VC1_TRACE_END_BLK
#endif

VC1_VIDEO_read_P_MACROBLOCK_MV4_PULLBACK:
#if (VC1_VIDEO_TRACE > 2)
            VC1_TRACE_BEG_BLK
            printf("        pre-pullback(%hd,%hd)\n", d[0], d[1]);
            VC1_TRACE_END_BLK
#endif
            if (d[0]) {
               pic->coeff[2] = (s16)(pic->x   << 2);
               pic->coeff[3] = (s16)(seq->wid << 2) - 4 - pic->coeff[2];
               if (d[0] > pic->coeff[3])
                  d[0] = pic->coeff[3];
               else {
                  pic->coeff[2] = -60 - pic->coeff[2];
                  if (d[0] < pic->coeff[2])
                     d[0] = pic->coeff[2];
               }
            }
            if (d[1]) {
               pic->coeff[2] = (s16)(pic->y   << 2);
               pic->coeff[3] = (s16)(seq->hgt << 2) - 4 - pic->coeff[2];
               if (d[1] > pic->coeff[3])
                  d[1] = pic->coeff[3];
               else {
                  pic->coeff[2] = -60 - pic->coeff[2];
                  if (d[1] < pic->coeff[2])
                     d[1] = pic->coeff[2];
               }
            }

            if (!a)
               goto VC1_VIDEO_read_P_MACROBLOCK_MV4_DMV3;

#if (VC1_VIDEO_TRACE > 2)
            VC1_TRACE_BEG_BLK
            printf("        pre-hybrid(%hd,%hd)\n", d[0], d[1]);
            VC1_TRACE_END_BLK
#endif
            /* HYBRIDPRED */
            GET_LSB;
            pic->coeff[2] = ((d[0] >= a[0]) ? (d[0]-a[0]) : (a[0]-d[0])) +
                            ((d[1] >= a[1]) ? (d[1]-a[1]) : (a[1]-d[1]));
            if (pic->coeff[2] > 32) {
               b = (DATA_BYTE >= 0x80) ? a : c;
            }
            else {
               pic->coeff[2] = ((d[0] >= c[0]) ? (d[0]-c[0]) : (c[0]-d[0])) +
                               ((d[1] >= c[1]) ? (d[1]-c[1]) : (c[1]-d[1]));
               if (pic->coeff[2] > 32) {
                  b = (DATA_BYTE >= 0x80) ? a : c;
               }
               else
                  goto VC1_VIDEO_read_P_MACROBLOCK_MV4_DMV3;
            }
            d[0] = b[0];
            d[1] = b[1];
            ALIGN_DATA_BYTE(MV4_DMV3, 1);
         }
VC1_VIDEO_read_P_MACROBLOCK_MV4_DMV3:
#ifdef VC1_VIDEO_TRACE
         VC1_TRACE_BEG_BLK
         printf("      Intra = %c\n"
                "      Coded = %c\n"
                "      pred  = %hd,%hd\n"
                "      DMV   = %hd,%hd\n",
                ((pic->MV4mode & 0x08) ? 'y' : 'n'),
                ((pic->CBP_t   & 0x08) ? 'y' : 'n'),
                d[0], d[1],
                pic->coeff[0], pic->coeff[1]);
         VC1_TRACE_END_BLK
#endif
         if ( (pic->coeff[0])||(pic->coeff[1]) ) {
            d[0]         += pic->coeff[0];
            d[1]         += pic->coeff[1];
            pic->coeff[0] = (s16)1 << (pic->MVsize & 0x0F);
            pic->coeff[1] = (s16)1 << (pic->MVsize >>   4);
            d[0]         += pic->coeff[0];
            d[1]         += pic->coeff[1];
            d[0]         &= (pic->coeff[0] << 1) - 1;
            d[1]         &= (pic->coeff[1] << 1) - 1;
            d[0]         -= pic->coeff[0];
            d[1]         -= pic->coeff[1];
            pic->coeff[0] =
            pic->coeff[1] = 0;
         }
#ifdef VC1_VIDEO_TRACE
         VC1_TRACE_BEG_BLK
         printf("      MV    = %hd,%hd\n", d[0], d[1]);
         VC1_TRACE_END_BLK
#endif
         pic->MV4mode <<= 1;
         pic->CBP_t   <<= 1;
         pic->c        += 2;
      } while (pic->c < 8);
      pic->CBP &= pic->CBP_t | 0x0F;

      if (pic->MV4mode != 0xF0) {
         VC1_4MVinc(pic->MVinc, pic->MVout_qp, (d-6),
                    pic->x, pic->y, seq->wid, seq->hgt, pic->frame_wid,
                    VC1_EPT_FASTUV_MCp(seq), &(pic->MV4mode));
         if ( (!(pic->CBP_lt))&&(!(pic->MV4mode & 0x0C))&&(pic->CBP & 0x0C) )
            pic->CBP_lt = pic->VST;
      }
      else
         pic->MV4mode = 0xFC;
#ifdef VC1_VIDEO_TRACE
      VC1_TRACE_BEG_BLK
      printf("    CBP     = 0x%02hX\n"
             "    MV4mode = 0x%02hX\n",
             (u16)(pic->CBP >> 2), (u16)(pic->MV4mode >> 2));
      VC1_TRACE_END_BLK
#endif
      if ( (pic->MV4mode)||(pic->CBP) ) {
         if (pic->MV4mode) {
            tmp = pic->MV4mode;
            if (pic->MV4mode == 0xFC) {
               /* entire MB is intra coded */
               pic->MVmode |= VC1_MVMODE_INTRA;
               pic->MV4mode = 0;
            }
            else
               pic->MV4mode++;
            pic->tab = NULL;

            /* 4MV AC prediction */
            a = pic->ACpred_Y + (7*(pic->x >> 3));
            b = pic->ACpred_Y + (7*(VC1_MAX_FRAME_WID/8));
            if (tmp & 0x80) { /* block 0 */
               if ( ( (pic->y)&&(a[0] != 0x7F7F) )||
                    ( (pic->x)&&(b[0] != 0x7F7F) ) )
                  goto VC1_VIDEO_read_P_MACROBLOCK_MV1_MQUANT0;
            }
            if (tmp & 0x40) { /* block 1 */
               if ( (tmp & 0x80)||( (pic->y)&&(a[7] != 0x7F7F) ) )
                  goto VC1_VIDEO_read_P_MACROBLOCK_MV1_MQUANT0;
            }
            if (tmp & 0x20) { /* block 2 */
               if ( (tmp & 0x80)||( (pic->x)&&(b[7] != 0x7F7F) ) )
                  goto VC1_VIDEO_read_P_MACROBLOCK_MV1_MQUANT0;
            }
            if (tmp & 0x10) { /* block 3 */
               if (tmp & 0x60)
                  goto VC1_VIDEO_read_P_MACROBLOCK_MV1_MQUANT0;
            }
            a = pic->ACpred_C + (7*(pic->x >> 3));
            b = pic->ACpred_C + (7*(VC1_MAX_FRAME_WID/8));
            if (tmp & 0x08) { /* block 4 */
               if ( ( (pic->y)&&(a[0] != 0x7F7F) )||
                    ( (pic->x)&&(b[0] != 0x7F7F) ) )
                  goto VC1_VIDEO_read_P_MACROBLOCK_MV1_MQUANT0;
            }
            if (tmp & 0x04) { /* block 5 */
               if ( ( (pic->y)&&(a[7] != 0x7F7F) )||
                    ( (pic->x)&&(b[7] != 0x7F7F) ) )
                  goto VC1_VIDEO_read_P_MACROBLOCK_MV1_MQUANT0;
            }
            pic->mv_cbp_tab &= 0xFC;
         }
         pic->tab = (const u8 *)1;
         goto VC1_VIDEO_read_P_MACROBLOCK_MV1_MQUANT0;
      }
/**** END   4MV Macroblock ****/
   }
   else {
/**** START 1MV Macroblock ****/
         pic->tab = VC1_MV_DIFF_TAB[VC1_PIC_MV_TABLEp(pic)];
VC1_VIDEO_read_P_MACROBLOCK_MV1_INDEX:
         VC1_PARSE_VLC(pic->tab, MV1_INDEX)
         bit_shift_inc = *(pic->tab+1);
         ALIGN_DATA_BYTE(MV1_DMV0, bit_shift_inc);

VC1_VIDEO_read_P_MACROBLOCK_MV1_DMV0:
         tmp = *(pic->tab+2);
#ifdef VC1_VIDEO_TRACE
         VC1_TRACE_BEG_BLK
         printf("    MV1index = %hhu\n", tmp);
         VC1_TRACE_END_BLK
#endif
         if (tmp >= 37) {
            tmp -= 37;
            pic->MVmode |= VC1_MVMODE_CODED;
            if (tmp != 36)
               pic->CBP_lt = pic->VST;
         }

         if (tmp == 35) {
            pic->b = (u16)pic->MVsize;
            if (!(pic->MVmode & VC1_MVMODE_HALFPIX))
               pic->b += 0x0011;
            a = pic->coeff;
VC1_VIDEO_read_P_MACROBLOCK_MV1_DMV1:
            while (pic->b) {
               tmp = (u8)pic->b & 0x0F;
               if (tmp > 8) {
                  tmp    -= 8;
                  pic->b -= 8;
                  GET_LSB;
                  *a = (s16)DATA_BYTE << tmp;
                  ALIGN8_DATA_BYTE(MV1_DMV1)
               }
               pic->b >>= 4;
               GET_LSB;
               *a |= (s16)(DATA_BYTE >> (8 - tmp));
               ALIGN_DATA_BYTE(MV1_DMV1, tmp);
               a++;
            }
            if (pic->MVmode & VC1_MVMODE_HALFPIX) {
               pic->coeff[0] <<= 1;
               pic->coeff[1] <<= 1;
            }
         }
         else if (tmp == 36)
            pic->MVmode |= VC1_MVMODE_INTRA;
         else if (tmp) {
            pic->b = (((u16)tmp % 6) | (((u16)tmp / 6) << 4)) + 0x0011;
            a      = pic->coeff;
VC1_VIDEO_read_P_MACROBLOCK_MV1_DMV2:
            while (pic->b) {
               tmp      = (u8)pic->b & 0x0F;
               pic->b >>= 4;
               if (tmp > 1) {
                  bit_shift_inc =      VC1_MV_SIZE_TAB[tmp-2];
                  *a            = (s16)VC1_MV_OFFSET_TAB[tmp-2];
                  if ( (tmp == 6)&&(pic->MVmode & VC1_MVMODE_HALFPIX) )
                     bit_shift_inc--;
                  GET_LSB;
                  *a += (s16)(DATA_BYTE >> (8 - bit_shift_inc));
                  DATA_BYTE     <<= bit_shift_inc;
                  BIT_SHIFT_BYTE += bit_shift_inc + 1;
                  if (DATA_BYTE >= 0x80)
                     *a = -(*a);
                  DATA_BYTE     <<= 1;
                  ALIGN0_DATA_BYTE(MV1_DMV2)
               }
               a++;
            }
            if (pic->MVmode & VC1_MVMODE_HALFPIX) {
               pic->coeff[0] <<= 1;
               pic->coeff[1] <<= 1;
            }
         }

VC1_VIDEO_read_P_MACROBLOCK_MV1_PREDICTION:
         /* MV Prediction */
         d = pic->MV + (pic->x >> 1);
         if ( ( (pic->x == 0)&&(pic->y == 0) )||
              (pic->MVmode & VC1_MVMODE_INTRA) ) {
            /* A and C are out of bounds */
            d[0] =
            d[1] = 0;
         }
         else if (pic->y == 0) {
            /* A is out of bounds */
            d[0] = d[-6];
            d[1] = d[-5];
            if ( (d[0])||(d[1]) ) {
               a    = NULL;
               goto VC1_VIDEO_read_P_MACROBLOCK_MV1_PULLBACK;
            }
         }
         else if (pic->x == 0) {
            /* C is out of bounds */
            if (seq->wid <= 16) {   /* last MB */
               /* B is out of bounds */
               d[0] = d[4];
               d[1] = d[5];
            }
            else {
               /* B is in bounds */
               a = d + 4;
               b = d + 4 + 8;
               d[0] = VC1_median3(a[0], b[0], 0);
               d[1] = VC1_median3(a[1], b[1], 0);
#if (VC1_VIDEO_TRACE > 2)
               VC1_TRACE_BEG_BLK
               printf("      A=(%hd,%hd) B=(%hd,%hd)\n",
                      a[0], a[1], b[0], b[1]);
               VC1_TRACE_END_BLK
#endif
            }
            if ( (d[0])||(d[1]) ) {
               a    = NULL;
               goto VC1_VIDEO_read_P_MACROBLOCK_MV1_PULLBACK;
            }
         }
         else {
            /* A, B and C are in bounds */
            a = d + 4;
            c = d - 6;
            if (seq->wid <= (pic->x+16))
               b = d;   /* last MB */
            else
               b = a + 8;
            d[0] = VC1_median3(a[0], b[0], c[0]);
            d[1] = VC1_median3(a[1], b[1], c[1]);
#if (VC1_VIDEO_TRACE > 2)
            VC1_TRACE_BEG_BLK
            printf("      A=(%hd,%hd) B=(%hd,%hd) C=(%hd,%hd)\n",
                   a[0], a[1], b[0], b[1], c[0], c[1]);
            VC1_TRACE_END_BLK
#endif

VC1_VIDEO_read_P_MACROBLOCK_MV1_PULLBACK:
#if (VC1_VIDEO_TRACE > 2)
            VC1_TRACE_BEG_BLK
            printf("      pre-pullback(%hd,%hd)\n", d[0], d[1]);
            VC1_TRACE_END_BLK
#endif
            if (d[0]) {
               pic->coeff[2] = (s16)(pic->x   << 2);
               pic->coeff[3] = (s16)(seq->wid << 2) - 4 - pic->coeff[2];
               if (d[0] > pic->coeff[3])
                  d[0] = pic->coeff[3];
               else {
                  pic->coeff[2] = -60 - pic->coeff[2];
                  if (d[0] < pic->coeff[2])
                     d[0] = pic->coeff[2];
               }
            }
            if (d[1]) {
               pic->coeff[2] = (s16)(pic->y   << 2);
               pic->coeff[3] = (s16)(seq->hgt << 2) - 4 - pic->coeff[2];
               if (d[1] > pic->coeff[3])
                  d[1] = pic->coeff[3];
               else {
                  pic->coeff[2] = -60 - pic->coeff[2];
                  if (d[1] < pic->coeff[2])
                     d[1] = pic->coeff[2];
               }
            }

            if (!a)
               goto VC1_VIDEO_read_P_MACROBLOCK_MV1_DMV3;

#if (VC1_VIDEO_TRACE > 2)
            VC1_TRACE_BEG_BLK
            printf("      pre-hybrid(%hd,%hd)\n", d[0], d[1]);
            VC1_TRACE_END_BLK
#endif
            /* HYBRIDPRED */
            GET_LSB;
            pic->coeff[2] = ((d[0] >= a[0]) ? (d[0]-a[0]) : (a[0]-d[0])) +
                            ((d[1] >= a[1]) ? (d[1]-a[1]) : (a[1]-d[1]));
            if (pic->coeff[2] > 32) {
               b = (DATA_BYTE >= 0x80) ? a : c;
            }
            else {
               pic->coeff[2] = ((d[0] >= c[0]) ? (d[0]-c[0]) : (c[0]-d[0])) +
                               ((d[1] >= c[1]) ? (d[1]-c[1]) : (c[1]-d[1]));
               if (pic->coeff[2] > 32) {
                  b = (DATA_BYTE >= 0x80) ? a : c;
               }
               else
                  goto VC1_VIDEO_read_P_MACROBLOCK_MV1_DMV3;
            }
            d[0] = b[0];
            d[1] = b[1];
            ALIGN_DATA_BYTE(MV1_DMV3, 1);
         }
VC1_VIDEO_read_P_MACROBLOCK_MV1_DMV3:
#ifdef VC1_VIDEO_TRACE
         VC1_TRACE_BEG_BLK
         printf("    Intra = %c\n"
                "    Coded = %c\n"
                "    pred  = %hd,%hd\n"
                "    DMV   = %hd,%hd\n",
                ((pic->MVmode & VC1_MVMODE_INTRA) ? 'y' : 'n'),
                ((pic->MVmode & VC1_MVMODE_CODED) ? 'y' : 'n'),
                d[0], d[1],
                pic->coeff[0], pic->coeff[1]);
         VC1_TRACE_END_BLK
#endif
         if ( (pic->coeff[0])||(pic->coeff[1]) ) {
            d[0]         += pic->coeff[0];
            d[1]         += pic->coeff[1];
            pic->coeff[0] = (s16)1 << (pic->MVsize & 0x0F);
            pic->coeff[1] = (s16)1 << (pic->MVsize >>   4);
            d[0]         += pic->coeff[0];
            d[1]         += pic->coeff[1];
            d[0]         &= (pic->coeff[0] << 1) - 1;
            d[1]         &= (pic->coeff[1] << 1) - 1;
            d[0]         -= pic->coeff[0];
            d[1]         -= pic->coeff[1];
         }
         d[8] = d[6];   /* store B predictor for next macroblock */
         d[9] = d[7];
         d[2] = d[4] = d[6] = d[0];
         d[3] = d[5] = d[7] = d[1];
#ifdef VC1_VIDEO_TRACE
         VC1_TRACE_BEG_BLK
         printf("    MV    = %hd,%hd\n", d[0], d[1]);
         VC1_TRACE_END_BLK
#endif
         if (!(pic->MVmode & VC1_MVMODE_INTRA)) {
            VC1_1MVinc(pic->MVinc, pic->MVout_qp, d,
                       pic->x, pic->y, seq->wid, seq->hgt, pic->frame_wid,
                       VC1_EPT_FASTUV_MCp(seq));
         }

         if (pic->MVmode & VC1_MVMODE_CODED) {
            pic->tab = VC1_PB_CBP_TAB[VC1_PIC_CBP_TABLEp(pic)];
            if (pic->MVmode & VC1_MVMODE_INTRA) {
VC1_VIDEO_read_P_MACROBLOCK_MV1_ACPRED:
               GET_LSB;
               if (DATA_BYTE >= 0x80)
                  pic->mv_cbp_tab |= 0x01;
               else
                  pic->mv_cbp_tab &= 0xFC;
#ifdef VC1_VIDEO_TRACE
               VC1_TRACE_BEG_BLK
               printf("    ACpred = %c\n", (VC1_AC_PREDp(pic) ? 'y' : 'n'));
               VC1_TRACE_END_BLK
#endif
               ALIGN_DATA_BYTE(((pic->tab)?MV1_CBPCY:TT_MB0), 1);
               if (pic->tab == NULL)
                  goto VC1_VIDEO_read_P_MACROBLOCK_TT_MB0;
            }
VC1_VIDEO_read_P_MACROBLOCK_MV1_CBPCY:
            VC1_PARSE_VLC(pic->tab, MV1_CBPCY)
            bit_shift_inc = *(pic->tab+1);
            pic->CBP      = *(pic->tab+2);
#ifdef VC1_VIDEO_TRACE
            VC1_TRACE_BEG_BLK
            printf("    CBP = 0x%02hX\n", (u16)(pic->CBP >> 2));
            VC1_TRACE_END_BLK
#endif
            ALIGN_DATA_BYTE(MV1_MQUANT0, bit_shift_inc);
            goto VC1_VIDEO_read_P_MACROBLOCK_MV1_MQUANT0;
         }
         else if (pic->MVmode & VC1_MVMODE_INTRA) {
            pic->tab = NULL;
VC1_VIDEO_read_P_MACROBLOCK_MV1_MQUANT0:
            if (VC1_PIC_DQUANT_FRMp(pic)) {
               bit_shift_inc = tmp = 0;
               switch (VC1_PIC_DQUANT_PROFILEp(pic)) {
               case VC1_DQUANT_PROFILE_4EDGES:
                  if ( (pic->x)&&(pic->x < (seq->wid-16))&&
                       (pic->y)&&(pic->y < (seq->hgt-16)) )
                     tmp = 1;
                  break;
               case VC1_DQUANT_PROFILE_2EDGES:
                  switch (VC1_PIC_DQUANT_EDGEp(pic)) {
                  case VC1_DQUANT_EDGE_LEFT:
                     if ( (pic->x)&&(pic->y) )
                        tmp = 1;
                     break;
                  case VC1_DQUANT_EDGE_TOP:
                     if ( (pic->y)&&(pic->x < (seq->wid-16)) )
                        tmp = 1;
                     break;
                  case VC1_DQUANT_EDGE_RIGHT:
                     if ( (pic->x < (seq->wid-16))&&(pic->y < (seq->hgt-16)) )
                        tmp = 1;
                     break;
                  case VC1_DQUANT_EDGE_BOTTOM:
                     if ( (pic->x)&&(pic->y < (seq->hgt-16)) )
                        tmp = 1;
                  }
                  break;
               case VC1_DQUANT_PROFILE_1EDGE:
                  switch (VC1_PIC_DQUANT_EDGEp(pic)) {
                  case VC1_DQUANT_EDGE_LEFT:
                     if (pic->x)
                        tmp = 1;
                     break;
                  case VC1_DQUANT_EDGE_TOP:
                     if (pic->y)
                        tmp = 1;
                     break;
                  case VC1_DQUANT_EDGE_RIGHT:
                     if (pic->x < (seq->wid-16))
                        tmp = 1;
                     break;
                  case VC1_DQUANT_EDGE_BOTTOM:
                     if (pic->y < (seq->hgt-16))
                        tmp = 1;
                  }
                  break;
               case VC1_DQUANT_PROFILE_ALL_MBS:
                  GET_LSB;
                  if (VC1_PIC_DQUANT_BILEVELp(pic)) {
                     tmp           = (DATA_BYTE & 0x80) ^ 0x80;
                     bit_shift_inc = 1;
                  }
                  else {
                     if (DATA_BYTE >= 0xE0) {
                        pic->Pquant_alt = (DATA_BYTE & 0x1F) << 1;
                        bit_shift_inc   = 8;
                     }
                     else {
                        pic->Pquant_alt = ((DATA_BYTE  & 0xE0) >> 4) +
                                          (pic->Pquant & 0xFE);
                        bit_shift_inc   = 3;
                        if (pic->Pquant_alt < 2)
                           pic->Pquant_alt =  2;
                        else if (pic->Pquant_alt > 62)
                           pic->Pquant_alt = 62;
                     }
                  }
               }

               if (tmp) {
                  pic->Mquant = pic->Pquant;
                  pic->DCstep = pic->DCstep_org;
               }
               else {
                  pic->Mquant = pic->Pquant_alt;
                  if (pic->Mquant <= 5)
                     pic->DCstep = pic->Mquant & 0xFE;
                  else if (pic->Mquant <= 9)
                     pic->DCstep = 8;
                  else
                     pic->DCstep = (pic->Mquant >> 2) + 6;
               }
#ifdef VC1_VIDEO_TRACE
               VC1_TRACE_BEG_BLK
               printf("    Mquant = %hd\n", (u16)(pic->Mquant >> 1));
               VC1_TRACE_END_BLK
#endif
               if (bit_shift_inc) {
                  ALIGN_DATA_BYTE(((pic->tab)?TT_MB0:MV1_ACPRED), bit_shift_inc);
               }
            }
            if (pic->tab == NULL)
               goto VC1_VIDEO_read_P_MACROBLOCK_MV1_ACPRED;
         }
/**** END   1MV Macroblock ****/
   }

VC1_VIDEO_read_P_MACROBLOCK_TT_MB0:
   if (pic->CBP_lt & VC1_VST_LEVEL_BLOCK) {
      pic->tab = VC1_TTMB_TAB[pic->CBP_l];
VC1_VIDEO_read_P_MACROBLOCK_TT_MB1:
      VC1_PARSE_VLC(pic->tab, TT_MB1)
      bit_shift_inc = *(pic->tab+1);
      pic->VST      = *(pic->tab+2);
#ifdef VC1_VIDEO_TRACE
      VC1_TRACE_BEG_BLK
      printf("    VST_Level = %s\n"
             "    VST_Type  = %hd\n"
             "    VST_Coded_MB = 0x%02hX\n",
             ((pic->VST & VC1_VST_LEVEL_FRAME) ? "Macroblock" : "Block"),
             (u16)VC1_VST_TYPEp(pic), (u16)VC1_VST_CODEDp(pic));
      VC1_TRACE_END_BLK
#endif
     ALIGN_DATA_BYTE(START_BLOCKS, bit_shift_inc);
   }

VC1_VIDEO_read_P_MACROBLOCK_START_BLOCKS:
#ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   printf("    DCstep = %hd\n", (u16)pic->DCstep);
   VC1_TRACE_END_BLK
#endif

   pic->ACidx      = VC1_PIC_TRANS_AC1p(pic);
   if (!(pic->MVmode & VC1_MVMODE_INTRA))
      pic->ACidx  |= 4;
   pic->ACrunlevel = VC1_RUN_LEVEL_TAB[pic->ACidx];
   pic->b = 0;
   do {
      if (pic->MV4mode) {
         if (pic->MV4mode >= 0x80) {
            pic->MVmode |=  VC1_MVMODE_INTRA;
            pic->CBP_lt  =  0;
            pic->ACidx  &= ~4;
         }
         else {
            pic->MVmode &= ~VC1_MVMODE_INTRA;
            pic->CBP_lt  =  1;
            pic->ACidx  |=  4;
         }
         if (pic->b == 4) {
            pic->ACidx  |= 4;
            pic->MV4mode = 0;
         }
         else
            pic->MV4mode <<= 1;
         pic->ACrunlevel = VC1_RUN_LEVEL_TAB[pic->ACidx];
      }
      memset(pic->coeff, 0, (sizeof(s16)*64));
#if defined(VC1_VIDEO_TRACE)||defined(OUTPUT_BLOCK_TRACE)
      VC1_TRACE_BEG_BLK
      printf("    <BLOCK %hu>\n", pic->b);
      VC1_TRACE_END_BLK
#endif
      if (pic->MVmode & VC1_MVMODE_INTRA) {
         /* DC Prediction */
         pic->tab = VC1_DC_DIFF_TAB[VC1_PIC_TRANS_DCp(pic) | (pic->b & 4)];

VC1_VIDEO_read_P_MACROBLOCK_DC_DIFF1:
         VC1_PARSE_VLC(pic->tab, DC_DIFF1)
         bit_shift_inc = *(pic->tab+1);
         pic->coeff[0] = (s16)(*(pic->tab+2));
#ifdef VC1_VIDEO_TRACE
         VC1_TRACE_BEG_BLK
         printf("      DCdiff = %hd -> ", pic->coeff[0]);
         VC1_TRACE_END_BLK
#endif

         if (pic->coeff[0]) {
            ALIGN_DATA_BYTE(DC_DIFF2, bit_shift_inc);
VC1_VIDEO_read_P_MACROBLOCK_DC_DIFF2:
            GET_LSB;
            if (pic->coeff[0] == 255) {   /* ESCAPE */
               pic->coeff[0] = (s16)DATA_BYTE;
               ALIGN8_DATA_BYTE(DC_DIFF_ESC)
VC1_VIDEO_read_P_MACROBLOCK_DC_DIFF_ESC:
               GET_LSB;
               if (pic->Mquant > 5)
                  goto VC1_VIDEO_read_P_MACROBLOCK_DC_DIFF_SIGN;
               bit_shift_inc   = (pic->Mquant <= 3) ? 2 : 1;
               pic->coeff[0] <<= bit_shift_inc;
            }
            else if (pic->Mquant > 5)
               goto VC1_VIDEO_read_P_MACROBLOCK_DC_DIFF_SIGN;
            else {
               bit_shift_inc   = (pic->Mquant <= 3) ? 2 : 1;
               pic->coeff[0] <<= bit_shift_inc;
               pic->coeff[0]  -= (s16)(1 | bit_shift_inc);
            }
            pic->coeff[0]  += (s16)(DATA_BYTE >> (8-bit_shift_inc));
            DATA_BYTE     <<= bit_shift_inc;
            BIT_SHIFT_BYTE += bit_shift_inc;

VC1_VIDEO_read_P_MACROBLOCK_DC_DIFF_SIGN:
            bit_shift_inc = 1;
            if (DATA_BYTE >= 0x80)
               pic->coeff[0] = -(pic->coeff[0]);
         }
         ALIGN_DATA_BYTE(DC_PRED, bit_shift_inc);

VC1_VIDEO_read_P_MACROBLOCK_DC_PRED:
#ifdef VC1_VIDEO_TRACE
         VC1_TRACE_BEG_BLK
         printf("0x%04hX\n", (u16)pic->coeff[0]);
         VC1_TRACE_END_BLK
#endif
         /* DC predictor */
         tmp = pic->mv_cbp_tab & 0xFD;
         if (pic->b < 4) {
            b = pic->DCpred_Y + (pic->x >> 2);
            if (pic->b >= 2) {
               c  = b + ((pic->b - 2) << 1);
               b  = c + 1; d  = c + 2; a  = c + 3;
            }
            else if (pic->b) {
               b += 2;
               c  = b + 1; a  = b + 2; d  = b + 3;
            }
            else {
               b += 3;
               c  = b - 2; a  = b - 1; d  = b;
               b[4] = b[1];
            }
            if ( (*a != 0x7F7F)&&(*c != 0x7F7F)&&
                 ( (a > d)||(pic->y) )&&( (pic->x)||(pic->b & 1) ) ) {
               if (*b != 0x7F7F) {
                  if (pic->DCstep != pic->DCstep_org) {
                     pic->c = *a;
                     a      = (s16 *)pic->blk;
                     *a     = (s16)VC1_SCALE_COEFF((s32)pic->c,
                                                   (s32)pic->DCstep_org,
                                                   (s32)pic->DCstep);
                     pic->c = *b;
                     b      = a + 1;
                     *b     = (s16)VC1_SCALE_COEFF((s32)pic->c,
                                                   (s32)pic->DCstep_org,
                                                   (s32)pic->DCstep);
                     pic->c = *c;
                     c      = b + 1;
                     *c     = (s16)VC1_SCALE_COEFF((s32)pic->c,
                                                   (s32)pic->DCstep_org,
                                                   (s32)pic->DCstep);
                  }
#if (VC1_VIDEO_TRACE == 1)
                  VC1_TRACE_BEG_BLK
                  printf("B=%04hX A=%04hX\nC=%04hX d=", *b, *a, *c);
                  VC1_TRACE_END_BLK
#endif
                  if ( (*c != *a)&&(abs(*b - *a) > abs(*b - *c)) ) {
                     c = a;
                     tmp |= 0x02;   /* pred_dir = top */
                  }
               }
               else {
                  if (pic->DCstep != pic->DCstep_org) {
                     pic->c = *a;
                     a      = (s16 *)pic->blk;
                     *a     = (s16)VC1_SCALE_COEFF((s32)pic->c,
                                                   (s32)pic->DCstep_org,
                                                   (s32)pic->DCstep);
                     pic->c = *c;
                     c      = a + 1;
                     *c     = (s16)VC1_SCALE_COEFF((s32)pic->c,
                                                   (s32)pic->DCstep_org,
                                                   (s32)pic->DCstep);
                  }
#if (VC1_VIDEO_TRACE == 1)
                  VC1_TRACE_BEG_BLK
                  printf("B=null A=%04hX\nC=%04hX d=", *a, *c);
                  VC1_TRACE_END_BLK
#endif
                  if (abs(*a) > abs(*c)) {
                     c = a;
                     tmp |= 0x02;   /* pred_dir = top */
                  }
               }
            }
            else if ( (*a != 0x7F7F)&&( (a > d)||(pic->y) ) ) {
               if (pic->DCstep != pic->DCstep_org) {
                  pic->c = *a;
                  a      = (s16 *)pic->blk;
                  *a     = (s16)VC1_SCALE_COEFF((s32)pic->c,
                                                (s32)pic->DCstep_org,
                                                (s32)pic->DCstep);
               }
#if (VC1_VIDEO_TRACE == 1)
               VC1_TRACE_BEG_BLK
               printf("B=null A=%04hX\nC=null d=", *a);
               VC1_TRACE_END_BLK
#endif
               c = a;
               tmp |= 0x02;   /* pred_dir = top */
            }
            else if ( (pic->x)||(pic->b & 1) ) {
               if ( (*c != 0x7F7F)&&(pic->DCstep != pic->DCstep_org) ) {
                  pic->c = *c;
                  c      = (s16 *)pic->blk;
                  *c     = (s16)VC1_SCALE_COEFF((s32)pic->c,
                                                (s32)pic->DCstep_org,
                                                (s32)pic->DCstep);
               }
#if (VC1_VIDEO_TRACE == 1)
               VC1_TRACE_BEG_BLK
               printf("B=null A=null\nC=%04hX d=", ((*c == 0x7F7F) ? 0 : *c));
               VC1_TRACE_END_BLK
#endif
            }
            else {
#if (VC1_VIDEO_TRACE == 1)
               VC1_TRACE_BEG_BLK
               printf("B=null A=null\nC=null d=");
               VC1_TRACE_END_BLK
#endif
               *c = 0;
            }
         }
         else {
            b = pic->DCpred_C + (pic->b - 4);
            c = b + (pic->x >> 3);
            d = c + 2;
            if ( (*d != 0x7F7F)&&(*c != 0x7F7F)&&(pic->y)&&(pic->x) ) {
               if (*b != 0x7F7F) {
                  if (pic->DCstep != pic->DCstep_org) {
                     pic->c = *b;
                     *b     = *d;
                     *d     = (s16)VC1_SCALE_COEFF((s32)(*d),
                                                   (s32)pic->DCstep_org,
                                                   (s32)pic->DCstep);
                     b      = (s16 *)pic->blk;
                     *b     = (s16)VC1_SCALE_COEFF((s32)pic->c,
                                                   (s32)pic->DCstep_org,
                                                   (s32)pic->DCstep);
                     pic->c = *c;
                     c      = b + 1;
                     *c     = (s16)VC1_SCALE_COEFF((s32)pic->c,
                                                   (s32)pic->DCstep_org,
                                                   (s32)pic->DCstep);
                  }
#if (VC1_VIDEO_TRACE == 1)
                  VC1_TRACE_BEG_BLK
                  printf("B=%04hX A=%04hX\nC=%04hX d=", *b, *d, *c);
                  VC1_TRACE_END_BLK
#endif
                  if ( (*c != *d)&&(abs(*b - *d) > abs(*b - *c)) ) {
                     c = d;
                     tmp |= 0x02;   /* pred_dir = top */
                  }
               }
               else {
                  if (pic->DCstep != pic->DCstep_org) {
                     *b     = *d;
                     *d     = (s16)VC1_SCALE_COEFF((s32)(*d),
                                                   (s32)pic->DCstep_org,
                                                   (s32)pic->DCstep);
                     pic->c = *c;
                     c      = (s16 *)pic->blk;
                     *c     = (s16)VC1_SCALE_COEFF((s32)pic->c,
                                                   (s32)pic->DCstep_org,
                                                   (s32)pic->DCstep);
                     b      = c + 1;
                  }
#if (VC1_VIDEO_TRACE == 1)
                  VC1_TRACE_BEG_BLK
                  printf("B=null A=%04hX\nC=%04hX d=", *d, *c);
                  VC1_TRACE_END_BLK
#endif
                  if (abs(*d) > abs(*c)) {
                     c = d;
                     tmp |= 0x02;   /* pred_dir = top */
                  }
               }
            }
            else if ( (*d != 0x7F7F)&&(pic->y) ) {
               if (pic->DCstep != pic->DCstep_org) {
                  *b = *d;
                  *d = (s16)VC1_SCALE_COEFF((s32)(*d),
                                            (s32)pic->DCstep_org,
                                            (s32)pic->DCstep);
                  b  = (s16 *)pic->blk;
               }
#if (VC1_VIDEO_TRACE == 1)
               VC1_TRACE_BEG_BLK
               printf("B=null A=%04hX\nC=null d=", *d);
               VC1_TRACE_END_BLK
#endif
               c = d;
               tmp |= 0x02;   /* pred_dir = top */
            }
            else if (pic->x) {
               if ( (*c != 0x7F7F)&&(pic->DCstep != pic->DCstep_org) ) {
                  pic->c = *c;
                  c      = (s16 *)pic->blk;
                  *c     = (s16)VC1_SCALE_COEFF((s32)pic->c,
                                                (s32)pic->DCstep_org,
                                                (s32)pic->DCstep);
               }
#if (VC1_VIDEO_TRACE == 1)
               VC1_TRACE_BEG_BLK
               printf("B=null A=null\nC=%04hX d=", ((*c == 0x7F7F) ? 0 : *c));
               VC1_TRACE_END_BLK
#endif
            }
            else {
#if (VC1_VIDEO_TRACE == 1)
               VC1_TRACE_BEG_BLK
               printf("B=null A=null\nC=null d=");
               VC1_TRACE_END_BLK
#endif
               *d = 0;
            }
            *b = *d;
         }

#if (VC1_VIDEO_TRACE != 2)
         pic->c = *c;
#endif
         if (*c != 0x7F7F)
            pic->coeff[0] += *c;
#if (VC1_VIDEO_TRACE == 1)
         VC1_TRACE_BEG_BLK
         printf("%04hX\n", pic->coeff[0]);
         VC1_TRACE_END_BLK
#endif
         if (pic->DCstep != pic->DCstep_org) {
            *d = (s16)VC1_SCALE_COEFF((s32)pic->coeff[0],
                                      (s32)pic->DCstep,
                                      (s32)pic->DCstep_org);
         }
         else
            *d = pic->coeff[0];
         pic->mv_cbp_tab = tmp;
#ifdef VC1_VIDEO_TRACE
         VC1_TRACE_BEG_BLK
         if (tmp & 0x01) {
            printf("      pred_dir = %s\n", ((tmp & 0x02) ? "TOP" : "LEFT"));
         }
# if (VC1_VIDEO_TRACE != 2)
         printf("      DCpred = 0x%04hX\n"
                "      coeff[0] = 0x%04hX\n",
                (u16)pic->c, (u16)pic->coeff[0]);
# endif
         VC1_TRACE_END_BLK
#endif
         if (pic->b == 4) {
            pic->ACidx     |= 4;
            pic->ACrunlevel = VC1_RUN_LEVEL_TAB[pic->ACidx];
         }
         tmp = 1;
      }
      else
         tmp = 0;

      if (pic->CBP >= 0x80) {
         if (pic->CBP_lt) {
            if (!(pic->VST & (VC1_VST_LEVEL_FRAME|VC1_VST_CODED_4x4_ALL))) {
               pic->tab = VC1_TTBLK_TAB[pic->CBP_l];
VC1_VIDEO_read_P_MACROBLOCK_TT_BLK0:
               VC1_PARSE_VLC(pic->tab, TT_BLK0)
               bit_shift_inc = *(pic->tab+1);
               pic->VST      = *(pic->tab+2);
#ifdef VC1_VIDEO_TRACE
               VC1_TRACE_BEG_BLK
               printf("    VST_Type  = %hd\n"
                      "    VST_Coded_BLK = 0x%02hX\n",
                      (u16)VC1_VST_TYPEp(pic), (u16)VC1_VST_CODEDp(pic));
               VC1_TRACE_END_BLK
#endif
               ALIGN_DATA_BYTE(TT_BLK1, bit_shift_inc);
            }
            else
               pic->VST &= ~VC1_VST_CODED_BLK_INIT;

VC1_VIDEO_read_P_MACROBLOCK_TT_BLK1:
            tmp = pic->VST & VC1_VST_CODED_8x4_ALL;
            if (!tmp) {
               pic->tab = VC1_TTSUBBLK_TAB[(VC1_VST_TYPEp(pic)==VC1_VST_TYPE_4x4)
                                           ? pic->CBP_l : 3];
VC1_VIDEO_read_P_MACROBLOCK_TT_BLK2:
               VC1_PARSE_VLC(pic->tab, TT_BLK2)
               bit_shift_inc = *(pic->tab+1);
               pic->CBP_lt   = *(pic->tab+2);
               ALIGN_DATA_BYTE(TT_BLK3, bit_shift_inc);
            }
            else {
               pic->CBP_lt = tmp;
               if (pic->VST & VC1_VST_CODED_SUB_INIT)
                  pic->VST &= ~VC1_VST_CODED_4x4_ALL;
            }
VC1_VIDEO_read_P_MACROBLOCK_TT_BLK3:
#ifdef VC1_VIDEO_TRACE
            VC1_TRACE_BEG_BLK
            printf("    VST_Coded_SUB = 0x%02hX\n", (u16)pic->CBP_lt);
            VC1_TRACE_END_BLK
#endif
            pic->zz = VC1_ZIGZAG_TAB[4 + VC1_VST_TYPEp(pic)];
            if (pic->CBP_lt < 0x80)
               goto VC1_VIDEO_read_P_MACROBLOCK_NEXT_SUB_BLK;
         }
         else
            pic->zz = VC1_ZIGZAG_TAB[4] + tmp;

         /* AC Coeffs */
         do {
            pic->tab = VC1_AC_INDEX_TAB[pic->ACidx];
            pic->c   = 0;
VC1_VIDEO_read_P_MACROBLOCK_AC_INDEX:
            VC1_PARSE_VLC(pic->tab, AC_INDEX)
            bit_shift_inc = *(pic->tab+1);
            tmp           = *(pic->tab+2);
#if (VC1_VIDEO_TRACE > 2)
            VC1_TRACE_BEG_BLK
            printf("      AC(tbl,index,run,level,sign) = %hu, ",
                   (u16)pic->ACidx);
            VC1_TRACE_END_BLK
#endif
            if (tmp == 255) {    /* ESCAPE */
               ALIGN_DATA_BYTE(AC_ESC_MODE, bit_shift_inc);
VC1_VIDEO_read_P_MACROBLOCK_AC_ESC_MODE:
               GET_LSB;
               if (DATA_BYTE < 0x40) {       /* Mode3 */
                  if (DATA_BYTE >= 0x20)
                     pic->trans_tab |= VC1_AC_MODE3_LAST;
#if (VC1_VIDEO_TRACE > 3)
                  VC1_TRACE_BEG_BLK
                  printf("\nMODE3: last=%c\n",
                         (VC1_AC_MODE3_LASTp(pic) ? 'y' : 'n'));
                  VC1_TRACE_END_BLK
#endif
                  ALIGN_DATA_BYTE(AC_ESC_MODE3_1, 3);

VC1_VIDEO_read_P_MACROBLOCK_AC_ESC_MODE3_1:
                  if (!VC1_AC_MODE3p(pic)) {
                     pic->trans_tab |= VC1_AC_MODE3;

                     pic->tab = VC1_AC_MODE3_LEVEL_SIZE_TAB[VC1_AC_MODE3_TABp(pic)];
VC1_VIDEO_read_P_MACROBLOCK_AC_ESC_MODE3_2:
                     VC1_PARSE_VLC(pic->tab, AC_ESC_MODE3_2)
                     bit_shift_inc   = *(pic->tab+1);
                     pic->level_size = *(pic->tab+2);
                     DATA_BYTE     <<= bit_shift_inc;
                     BIT_SHIFT_BYTE += bit_shift_inc;
                     pic->run_size   = 5 - (DATA_BYTE >> 6);
#if (VC1_VIDEO_TRACE > 3)
                     VC1_TRACE_BEG_BLK
                     printf("       size(level,run)=%hu,%hu\n",
                            (u16)pic->level_size, (u16)(8 - pic->run_size));
                     VC1_TRACE_END_BLK
#endif
                     ALIGN_DATA_BYTE(AC_ESC_MODE3_3, 2);
                  }

VC1_VIDEO_read_P_MACROBLOCK_AC_ESC_MODE3_3:
                  GET_LSB;
                  bit_shift_inc = 8 - pic->run_size;
                  pic->zz      += DATA_BYTE >> pic->run_size;
#if (VC1_VIDEO_TRACE > 2)
                  VC1_TRACE_BEG_BLK
                  printf("255,%2hu,", (u16)(DATA_BYTE >> pic->run_size));
                  VC1_TRACE_END_BLK
#endif
                  DATA_BYTE     <<= bit_shift_inc;
                  BIT_SHIFT_BYTE += bit_shift_inc;

                  if (DATA_BYTE < 0x80)
                     pic->c = 0x4000;
                  ALIGN_DATA_BYTE(AC_ESC_MODE3_4, 1);

VC1_VIDEO_read_P_MACROBLOCK_AC_ESC_MODE3_4:
                  if (pic->level_size > 8) {
                     GET_LSB;
                     pic->c |= (s16)DATA_BYTE << (pic->level_size - 8);
                     ALIGN8_DATA_BYTE(AC_ESC_MODE3_5)
VC1_VIDEO_read_P_MACROBLOCK_AC_ESC_MODE3_5:
                     bit_shift_inc = pic->level_size - 8;
                  }
                  else
                     bit_shift_inc = pic->level_size;
                  GET_LSB;
                  pic->c |= (s16)DATA_BYTE >> (8 - bit_shift_inc);
#if (VC1_VIDEO_TRACE > 2)
                  VC1_TRACE_BEG_BLK
                  printf("%2hd,%hd\n", (pic->c & 0x0FFF), ((pic->c >> 14)^1));
                  VC1_TRACE_END_BLK
#endif
                  if (pic->c >= 0x4000)
                     pic->c -= 0x4000;
                  else
                     pic->c  = -(pic->c);
                  ALIGN_DATA_BYTE(AC_ESC_MODE3_6, bit_shift_inc);
VC1_VIDEO_read_P_MACROBLOCK_AC_ESC_MODE3_6:
                  if (VC1_AC_MODE3_LASTp(pic)) {
                     pic->trans_tab &= ~VC1_AC_MODE3_LAST;
                     tmp = 255;
                  }
                  else
                     tmp =   0;
                  goto VC1_VIDEO_read_P_MACROBLOCK_AC_NEXT;
               }
               else if (DATA_BYTE < 0x80) {  /* Mode2 */
                  bit_shift_inc   = 2;
                  pic->trans_tab |= VC1_AC_MODE2;
               }
               else                          /* Mode1 */
                  bit_shift_inc   = 1;

               pic->tab = VC1_AC_INDEX_TAB[pic->ACidx];
               ALIGN_DATA_BYTE(AC_INDEX2, bit_shift_inc);
VC1_VIDEO_read_P_MACROBLOCK_AC_INDEX2:
               VC1_PARSE_VLC(pic->tab, AC_INDEX2)
               tmp     = *(pic->tab+2);
               ACdelta = pic->ACrunlevel;
#if (VC1_VIDEO_TRACE > 2)
               VC1_TRACE_BEG_BLK
               printf("%3hu,", (u16)tmp);
               VC1_TRACE_END_BLK
#endif
               if (VC1_AC_MODE2p(pic)) {
                  bit_shift_inc = ACdelta[((u32)tmp << 1)+7];
                  ACdelta      += 3;
               }
               else {
                  bit_shift_inc = ACdelta[((u32)tmp << 1)+6];
                  ACdelta      += 1;
               }
               if (tmp >= *(pic->ACrunlevel))
                  ACdelta++;
               ACdelta      += *ACdelta;
               ACdelta      += bit_shift_inc;
               bit_shift_inc = *(pic->tab+1);
               ACdelta      += *(pic->ACrunlevel) + *(pic->ACrunlevel);
               if (VC1_AC_MODE2p(pic)) {
                  pic->trans_tab &= ~VC1_AC_MODE2;
                  pic->zz        += *(ACdelta);
#if (VC1_VIDEO_TRACE > 2)
                  VC1_TRACE_BEG_BLK
                  printf("%hu+%hu,%2hu,",
                         (u16)pic->ACrunlevel[((u32)tmp << 1)+6],
                         (u16)ACdelta[0],
                         (u16)pic->ACrunlevel[((u32)tmp << 1)+7]);
                  VC1_TRACE_END_BLK
#endif
               }
               else {
                  pic->c = (s16)(*(ACdelta));
#if (VC1_VIDEO_TRACE > 2)
                  VC1_TRACE_BEG_BLK
                  printf("%2hu,%hu+%hu,",
                         (u16)pic->ACrunlevel[((u32)tmp << 1)+6],
                         (u16)pic->ACrunlevel[((u32)tmp << 1)+7],
                         (u16)ACdelta[0]);
                  VC1_TRACE_END_BLK
#endif
               }
            }
#if (VC1_VIDEO_TRACE > 2)
            else {
               VC1_TRACE_BEG_BLK
               printf("%3hu,%2hu,%2hu,",
                      (u16)tmp,
                      (u16)pic->ACrunlevel[((u32)tmp << 1)+6],
                      (u16)pic->ACrunlevel[((u32)tmp << 1)+7]);
               VC1_TRACE_END_BLK
            }
#endif
            pic->zz += (u16)pic->ACrunlevel[((u32)tmp << 1)+6];
            pic->c  += (s16)pic->ACrunlevel[((u32)tmp << 1)+7];
            ALIGN_DATA_BYTE(AC_SIGN, bit_shift_inc);

VC1_VIDEO_read_P_MACROBLOCK_AC_SIGN:
            GET_LSB;
            if (DATA_BYTE >= 0x80)
               pic->c = -(pic->c);
#if (VC1_VIDEO_TRACE > 2)
            VC1_TRACE_BEG_BLK
            printf("%hu\n", (u16)(DATA_BYTE >> 7));
            VC1_TRACE_END_BLK
#endif
            ALIGN_DATA_BYTE(AC_NEXT, 1);

VC1_VIDEO_read_P_MACROBLOCK_AC_NEXT:
            pic->coeff[*(pic->zz++)] = pic->c;
            if ( (pic->CBP_lt)&&(tmp >= *(pic->ACrunlevel)) ) {
VC1_VIDEO_read_P_MACROBLOCK_NEXT_SUB_BLK:
               pic->CBP_lt <<= 1;
               if (!pic->CBP_lt) {
                  pic->CBP_lt++;
#if (VC1_VIDEO_TRACE > 2)
                  VC1_TRACE_BEG_BLK
                  printf("    <NO MORE SUB_BLOCKS>\n");
                  VC1_TRACE_END_BLK
#endif
                  break;
               }
               while (*(pic->zz) < 128) {
                  pic->zz++;
               }
               if (*(pic->zz) > 128) {
#if (VC1_VIDEO_TRACE > 2)
                  VC1_TRACE_BEG_BLK
                  printf("    <END OF ZZ TABLE>\n");
                  VC1_TRACE_END_BLK
#endif
                  break;
               }
               pic->zz++;
               if (pic->CBP_lt < 0x80)
                  goto VC1_VIDEO_read_P_MACROBLOCK_NEXT_SUB_BLK;
#if (VC1_VIDEO_TRACE > 2)
               VC1_TRACE_BEG_BLK
               printf("    <END SUB_BLOCK>\n");
               VC1_TRACE_END_BLK
#endif
               tmp = 0;
            }
         } while (tmp < *(pic->ACrunlevel));
      }

      VC1_VIDEO_end_BLOCK(seq, pic);
      pic->CBP <<= 1;

#ifdef VC1_VIDEO_TRACE
      VC1_TRACE_BEG_BLK
      for (tmp = 0; tmp < 64; tmp++) {
         if ((tmp & 0x07) == 0x00)
            printf("      ");
         printf("%04hX ", pic->coeff[tmp]);
         if ((tmp & 0x07) == 0x07)
            printf("\n");
      }
      printf("    <END_BLOCK>\n");
      VC1_TRACE_END_BLK
#endif
   } while (++(pic->b) < 6);

   pic->x += 16;
   if (pic->x >= seq->wid) {
      pic->x  =  0;  /* finished a row */
      pic->y += 16;
      if (pic->y >= seq->hgt) {
         pic->y = 0; /* finished the frame */
#ifdef VC1_VIDEO_TRACE
         VC1_TRACE_BEG_BLK
         printf("<END_PIC>\n");
         VC1_TRACE_END_BLK
#endif
         goto VC1_VIDEO_read_P_MACROBLOCK_END_PIC;
      }

      memset((pic->ACpred_Y + (7*(VC1_MAX_FRAME_WID/8))), 0x7F, (sizeof(s16)*14));
      memset((pic->ACpred_C + (7*(VC1_MAX_FRAME_WID/8))), 0x7F, (sizeof(s16)*14));
   }

   if (pic->BPbit == 0x01) {
      pic->BPbit   = 0x80;
      pic->BPi++;
   }
   else
      pic->BPbit >>= 1;
   goto VC1_VIDEO_read_P_MACROBLOCK_NEXT;

VC1_VIDEO_read_P_MACROBLOCK_END_PIC:
   PARSER_STATE = VC1_VIDEO_frame_completed_STATE;

VC1_VIDEO_read_P_MACROBLOCK_STOP:
   seq->bit_shift   = BIT_SHIFT_BYTE;
   seq->data        = DATA_BYTE;
   seq->payload_len = 0;

#undef NEXT
#undef START
#undef MV4_CBPCY
#undef MV4_BLK_MVDATA
#undef MV4_INDEX
#undef MV4_DMV0
#undef MV4_DMV1
#undef MV4_DMV2
#undef MV4_DMV3
#undef MV1_INDEX
#undef MV1_DMV0
#undef MV1_DMV1
#undef MV1_DMV2
#undef MV1_DMV3
#undef MV1_ACPRED
#undef MV1_CBPCY
#undef MV1_MQUANT0
#undef TT_MB0
#undef TT_MB1
#undef START_BLOCKS
#undef DC_DIFF1
#undef DC_DIFF2
#undef DC_DIFF_ESC
#undef DC_PRED
#undef TT_BLK0
#undef TT_BLK1
#undef TT_BLK2
#undef TT_BLK3
#undef AC_INDEX
#undef AC_ESC_MODE
#undef AC_ESC_MODE3_1
#undef AC_ESC_MODE3_2
#undef AC_ESC_MODE3_3
#undef AC_ESC_MODE3_4
#undef AC_ESC_MODE3_5
#undef AC_ESC_MODE3_6
#undef AC_INDEX2
#undef AC_SIGN
#undef AC_NEXT
}


/* Process an inter block.
      INPUT: seq - Pointer to a SEQ_LAYER structure.
             pic - Pointer to a PIC_LAYER structure. */
void VC1_VIDEO_end_BLOCK(SEQ_LAYER *seq, PIC_LAYER *pic)
{
   s16 *pred_top, *pred_lft, *coeff;
   int  i, tf;

   if (pic->b < 4) {
      pred_top = pred_lft = pic->ACpred_Y;
      if (pic->b & 1)
         pred_top += 7;
      if (pic->b & 2)
         pred_lft += 7;
   }
   else {
      pred_lft = pic->ACpred_C;
      if (pic->b & 1)
         pred_lft += 7;
      pred_top = pred_lft;
   }
   pred_top += 7*(pic->x >> 3);
   pred_lft += 7*(VC1_MAX_FRAME_WID/8);

   if (pic->MVmode & VC1_MVMODE_INTRA) {
      if (*pred_lft != 0x7F7F)
         pic->overlap |= 0x04;   /* vertical overlap */
      if (*pred_top != 0x7F7F)
         pic->overlap |= 0x08;   /* horizontal overlap */

      /* AC Prediction */
      coeff = pic->coeff + 1;
      i     = 14;
      if (!(pic->mv_cbp_tab & 0x01))
         goto VC1_VIDEO_end_BLOCK_NO_AC_PRED;

      if ( (pic->overlap & 0x04)&&
           ( (!(pic->overlap & 0x08))||(VC1_AC_PREDp(pic) == VC1_AC_PRED_LEFT) ) ) {
         if (pic->Mquant != pic->Pquant) {
            do {
               *(pred_top++) = (s16)VC1_SCALE_COEFF((s32)(*(coeff++)),
                                                    (s32)pic->Mquant-1,
                                                    (s32)pic->Pquant-1);
            } while (--i > 7);
            do {
               *coeff       += (s16)VC1_SCALE_COEFF((s32)(*pred_lft),
                                                    (s32)pic->Pquant-1,
                                                    (s32)pic->Mquant-1);
               *(pred_lft++) = (s16)VC1_SCALE_COEFF((s32)(*coeff),
                                                    (s32)pic->Mquant-1,
                                                    (s32)pic->Pquant-1);
               coeff        += 8;
            } while (--i);
         }
         else {
            do {
               *(pred_top++) = *(coeff++);
            } while (--i > 7);
            do {
               *coeff       += *pred_lft;
               *(pred_lft++) = *coeff;
               coeff        += 8;
            } while (--i);
         }
      }
      else if ( (pic->overlap & 0x08)&&
                ( (!(pic->overlap & 0x04))||(VC1_AC_PREDp(pic) == VC1_AC_PRED_TOP) ) ) {
         if (pic->Mquant != pic->Pquant) {
            do {
               *coeff       += (s16)VC1_SCALE_COEFF((s32)(*pred_top),
                                                    (s32)pic->Pquant-1,
                                                    (s32)pic->Mquant-1);
               *(pred_top++) = (s16)VC1_SCALE_COEFF((s32)(*(coeff++)),
                                                    (s32)pic->Mquant-1,
                                                    (s32)pic->Pquant-1);
            } while (--i > 7);
            do {
               *(pred_lft++) = (s16)VC1_SCALE_COEFF((s32)(*coeff),
                                                    (s32)pic->Mquant-1,
                                                    (s32)pic->Pquant-1);
               coeff        += 8;
            } while (--i);
         }
         else {
            do {
               *coeff       += *pred_top;
               *(pred_top++) = *(coeff++);
            } while (--i > 7);
            do {
               *(pred_lft++) = *coeff;
               coeff        += 8;
            } while (--i);
         }
      }
      else {
VC1_VIDEO_end_BLOCK_NO_AC_PRED:
         if (pic->Mquant != pic->Pquant) {
            do {
               *(pred_top++) = (s16)VC1_SCALE_COEFF((s32)(*(coeff++)),
                                                    (s32)pic->Mquant-1,
                                                    (s32)pic->Pquant-1);
            } while (--i > 7);
            do {
               *(pred_lft++) = (s16)VC1_SCALE_COEFF((s32)(*coeff),
                                                    (s32)pic->Mquant-1,
                                                    (s32)pic->Pquant-1);
               coeff        += 8;
            } while (--i);
         }
         else {
            do {
               *(pred_top++) = *(coeff++);
            } while (--i > 7);
            do {
               *(pred_lft++) = *coeff;
               coeff        += 8;
            } while (--i);
         }
      }

      coeff       = pic->coeff;
      i           = 63;
      tf          =  0;
      *(coeff++) *= (s16)pic->DCstep;
   }
   else {
      *pred_top = *pred_lft = 0x7F7F;  /* inter-block value */
      if (pic->b < 4) {
         pred_top = pic->DCpred_Y + (pic->x >> 2);
         if (pic->b >= 2)
            pred_top += ((pic->b - 2) << 1) + 2;
         else if (pic->b)
            pred_top += 3 + 2;
         else
            pred_top += 3;
      }
      else {
         pred_lft  = pic->DCpred_C + (pic->b - 4);
         pred_top  = pred_lft + (pic->x >> 3) + 2;
         *pred_lft = *pred_top;
      }
      *pred_top = 0x7F7F;  /* inter-block value */

      if (pic->CBP < 0x80)
         goto VC1_VIDEO_end_BLOCK_SKIP_TRANSFORM;
      coeff = pic->coeff;
      i     = 64;
      tf    = (int)VC1_VST_TYPEp(pic);
   }

   /* Inverse Quantization */
   if (VC1_PIC_UNIFORM_QUANTp(pic)) {
      do {
         if (*coeff)
            *coeff *= (s16)pic->Mquant;
         coeff++;
      } while (--i);
   }
   else {
      s16 q = (s16)(pic->Mquant >> 1);
      do {
         if (*coeff) {
            *coeff *= (s16)pic->Mquant;
            if (*coeff > 0)
               *coeff += q;
            else
               *coeff -= q;
         }
         coeff++;
      } while (--i);
   }

   /* Inverse Transformation */
   VC1_inv_trans[tf](pic->coeff);

VC1_VIDEO_end_BLOCK_SKIP_TRANSFORM:
   /* put the final block into the frame */
   i = (int)pic->frame_wid;
   if (pic->b >= 4) {
      i >>= 1;
      coeff = (s16 *)(pic->LUT + 256);
   }
   else
      coeff = (s16 *)pic->LUT;

#ifdef OUTPUT_BLOCK_TRACE
   VC1_TRACE_BEG_BLK
   printf("BLK: pred_blk = %u\n", (u32)pic->pred_blk);
   if (pic->pred_blk < pic->pred_frame)
      printf("SAT_ERROR: pred_blk < pred_frame  wid=%d\n", i);
   else if ((pic->pred_blk+(i*7)+8) > (pic->pred_frame+pic->frame_size))
      printf("SAT_ERROR: pred_blk > pred_frame+%u  wid=%d\n", pic->frame_size, i);
   VC1_TRACE_END_BLK
#else
# ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   printf("BLK: pred_blk = %8u\n", (u32)(pic->pred_blk - pic->pred_frame));
   VC1_TRACE_END_BLK
# endif


   if ( (VC1_SEQ_OVERLAPp(seq))&&(pic->Pquant >= 18) ) {
      /* Overlap Smoothing */
# ifndef USE_PRECISE_OVERLAP
      if ( (pic->x)||(pic->b == 1)||(pic->b == 3) ) {
         if (pic->overlap & 0x04)
            VC1_overlap_Y(pic->coeff, pic->pred_blk, i);
         if ( (pic->y)||(pic->b & 2) ) {
            if (pic->overlap >= 0x80)
               VC1_overlap_X_last(pic->pred_blk, i);
            if ( (pic->overlap & 0x08)&&(pic->b & 0x05)&&
                 ((pic->x+16) >= seq->wid) )
               VC1_overlap_X(pic->coeff, pic->pred_blk, i);
         }
      }
# else
      if ( (pic->x)||(pic->b == 1)||(pic->b == 3) ) {
         if ( (pic->y)||(pic->b & 2) ) {
            if (pic->overlap >= 0x80)
               pic->overlap |= 0x02;
            if ( (pic->overlap & 0x08)&&(pic->b & 0x05)&&
                 ((pic->x+16) >= seq->wid) )
               pic->overlap |= 0x01;
         }
      }
      else
         pic->overlap &= 0xF8;

      pred_top = pred_lft = pic->Ocoeff;
      if (pic->b <= 3) {
         pred_lft += ((pic->b & 1) << 4) + (pic->x << 1);
         pred_top += ((pic->b & 2) << 4) + (VC1_MAX_FRAME_WID*2);
         pic->c    = 64 - ((pic->b & 2) << 3);
      }
      else if (pic->b == 4) {
         pred_lft += (VC1_MAX_FRAME_WID*2) + 96 + pic->x;
         pred_top += (VC1_MAX_FRAME_WID*3) + 96;
         pic->c    = 32;
      }
      else {
         pred_lft += (VC1_MAX_FRAME_WID*3) +144 + pic->x;
         pred_top += (VC1_MAX_FRAME_WID*4) +144;
         pic->c    = 32;
      }
      VC1_overlap_Y(pic->coeff, (pred_top + pic->c),
                    ((pic->b) ? pred_lft : NULL), pred_top,
                    ((pic->overlap & 0x04) ? pic->pred_blk : NULL), i);
      VC1_overlap_X_last(pic->coeff, pred_top,
                         ((pic->overlap & 0x02) ? pic->pred_blk : NULL), i);
      VC1_overlap_X(pic->coeff, pred_lft, pred_top,
                    ((pic->overlap & 0x01) ? pic->pred_blk : NULL), i);
# endif
      if (pic->b & 0x05)
         pic->overlap <<= 1;
      else if (pic->overlap & 0x08)
         pic->overlap  |= 0x80;
      else
         pic->overlap  &= 0x70;
   }
   pic->overlap &= 0xF0;
#endif
   if (pic->MVmode & VC1_MVMODE_INTRA) {
      MV_SATURATION(pic->coeff, pic->pred_blk, i);
   }
   else {
      VC1_INT_FUNC int_func;

# ifdef VC1_VIDEO_TRACE
      VC1_TRACE_BEG_BLK
      printf("     ref_blk  = %8u\n",
             (u32)(pic->ref_blkF + pic->MVinc[pic->b] - pic->ref_frameF));
      VC1_TRACE_END_BLK
# endif
      int_func = VC1_get_blk(pic->blk, (pic->ref_blkF + pic->MVinc[pic->b]),
                             pic->MVout_qp[pic->b], (u32)i,
                             (pic->MVmode | (u8)(pic->b & 4)), (u8 *)coeff);
      if (int_func)
         int_func(pic->blk, (pic->MVout_qp[pic->b] & 0x000F), pic->rnd);
      VC1_MV_ADD_PRED(pic->coeff, pic->blk, pic->pred_blk, i);
   }
# ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   VC1_MV_COPY_OUTPUT(pic->coeff, pic->pred_blk, i);
   VC1_TRACE_END_BLK
# endif

   /* move frame pointers into correct position */
   switch (pic->b) {
   case 0:
      i   =  8;
      break;
   case 1:
      i <<=  3;
      i  -=  8;
      break;
   case 2:
      i   =  8;
      break;
   case 3:
      i  *= ((3 * (int)pic->y) >> 2) + 8;
      i  += (16 + (int)pic->x) >> 1;
      i   = (int)((pic->luma_size * 5) >> 2) - i;
      break;
   case 4:
      i   = (int)pic->luma_size / -4;
      break;
   case 5:
      if ((pic->x+16) == seq->wid) {
         i *= ((3 * (int)pic->y) >> 1) + 30;
         i += (int)pic->frame_wid - (int)seq->wid;
      }
      else
         i *=  (3 * (int)pic->y) >> 1;

      i  += (32 + (int)pic->x) >> 1;
      i  -= (int)pic->luma_size;
   }

#if defined(OUTPUT_BLOCK_TRACE)
   VC1_TRACE_BEG_BLK
   printf("MV_REF: blkF=%u  blkB=%u\n", (u32)pic->ref_blkF, (u32)pic->ref_blkB);
   VC1_TRACE_END_BLK
#elif defined(VC1_VIDEO_TRACE)
   VC1_TRACE_BEG_BLK
   printf("MV_REF: blkF  = %8u  blkB  = %8u\n",
          (u32)(pic->ref_blkF - pic->ref_frameF),
          (u32)(pic->ref_blkB - pic->ref_frameB));
   VC1_TRACE_END_BLK
#endif

   pic->pred_blk += i;
   pic->ref_blkF += i;

#if defined(OUTPUT_BLOCK_TRACE)
   VC1_TRACE_BEG_BLK
   printf("              = %u\n", (u32)pic->pred_blk);
   printf("            =%u      =%u\n", (u32)pic->ref_blkF, (u32)pic->ref_blkB);
   VC1_TRACE_END_BLK
#elif defined(VC1_VIDEO_TRACE)
   VC1_TRACE_BEG_BLK
   printf("              = %8u\n", (u32)(pic->pred_blk - pic->pred_frame));
   printf("              = %8u        = %8u\n",
          (u32)(pic->ref_blkF - pic->ref_frameF),
          (u32)(pic->ref_blkB - pic->ref_frameB));
   VC1_TRACE_END_BLK
#endif
}


/* Scale the given 'coeff' to the new step size */
s32 VC1_SCALE_COEFF(s32 coeff, s32 step_old, s32 step_new)
{
   if (coeff) {
/*
      step_new = (1 << 18) / step_new;
      coeff   *= step_old;
      if (coeff > 2047)
         coeff =  2047;
      else if (coeff < -2048)
         coeff = -2048;
*/
#if defined(VC1_VIDEO_TRACE) && (VC1_VIDEO_TRACE != 2)

      s32 tmp;

      tmp = coeff * step_old * ((1 << 18) / step_new);
      tmp = ((tmp + (1 << 17)) >> 18);
      VC1_TRACE_BEG_BLK
      printf("  SCALE_COEFF(%2d,%2d) = 0x%04hX -> 0x%04hX\n",
             step_old, step_new, (u16)coeff, (u16)tmp);
      VC1_TRACE_END_BLK
      return tmp;

#else

      coeff *= step_old * ((1 << 18) / step_new);
      return ((coeff + (1 << 17)) >> 18);

#endif
   }
   return 0;
}

